
#include "Application.h"
#include "AbstractDocument.h"
#include "ScriptableDBDocument.h"
#include "ResolveObjectSpecifier.h"

#include "Exceptions.h"

TPropertyDescription TApplication::fPropertiesOfClass[] = {
	// "color" is temporary for resolution mode
	{	pColor,					0,		typeLongInteger,			typeLongInteger			},
	{	pSelection,				0,		typeAEList,					typeObjectSpecifier		},
	{	pIsFrontProcess,		0,		typeBoolean,				typeBoolean				},
	{	pClipboard,				0,		typeObjectSpecifier,		typeObjectSpecifier		},
	{	pInsertionLoc,			0,		typeObjectSpecifier,		typeObjectSpecifier		}
};

#pragma segment ObjectResident
ImplementClassData(TApplication, clApplication);

//--------------------------------------------------------------------------------
// TApplication::ObjectClass
//--------------------------------------------------------------------------------
DescType TApplication::ObjectClass(TTransaction*, Boolean /*recordedClass*/)
{
	return cApplication;
}

//--------------------------------------------------------------------------------
// TApplication::DerivedFromOSLClass
//--------------------------------------------------------------------------------
Boolean TApplication::DerivedFromOSLClass(TTransaction* t, DescType objectClass)
{
	return ((objectClass == cApplication) || Inherited::DerivedFromOSLClass(t, objectClass));
}

//--------------------------------------------------------------------------------
// TApplication::ElementIterator
//--------------------------------------------------------------------------------
TAbstractObjectIterator* TApplication::ElementIterator(TTransaction*)
{
	return new TDocumentLoop();
}

//--------------------------------------------------------------------------------
// TApplication::Idle
//
// Called once per waitnextevent
//--------------------------------------------------------------------------------
void TApplication::Idle()
{
	if(fCloseRequestsPending)
		this->ProcessCloseRequest();
}

//--------------------------------------------------------------------------------
// TApplication::TearDownApplication
//--------------------------------------------------------------------------------
void TApplication::TearDownApplication()
{
	//
	// Close all open documents (documents should save themselves before closing)
	//
	for(TDocumentLoop iter; iter.More(nil); iter.Next(nil))
	{
		TAbstractDocument* doc = iter.RemoveCurrentDocument(nil);
		delete doc;
	}
} // TApplication::TearDownApplication

//--------------------------------------------------------------------------------
// TApplication::MarkForClose
//--------------------------------------------------------------------------------
void TApplication::MarkForClose(TAbstractDocument* doc)
{
	fCloseRequestsPending = true;
	doc->MarkForClose();
} // TApplication::MarkForClose

//--------------------------------------------------------------------------------
// TApplication::ProcessCloseRequest
//--------------------------------------------------------------------------------
void TApplication::ProcessCloseRequest()
{
	Boolean requestsPending = false;
	
	if(fCloseRequestsPending)
	{
		for(TDocumentLoop iter; iter.More(nil); iter.Next(nil))
		{
			if(iter.CurrentDocument()->CloseRequested())
			{
				//
				// Also test to see if doc can be closed
				// (wouldn't want to close a document that
				// some thread was working with)
				// 
				if(true)
				{
					TAbstractDocument* doc = iter.RemoveCurrentDocument(nil);
					delete doc;
				}
				else
					requestsPending = true;
			}
		}
		fCloseRequestsPending = requestsPending;
	}
} // TApplication::ProcessCloseRequest

//--------------------------------------------------------------------------------
// TApplication::FindDocument
//--------------------------------------------------------------------------------
TAbstractDocument* TApplication::FindDocument(Int64 documentID, Boolean searchForAndOpen /* = false */)
{
	TAbstractDocument* doc = nil;
	
	for(TDocumentLoop iter; iter.More(nil); iter.Next(nil))
	{
		if(iter.CurrentDocument()->DocumentIdentifier() == documentID)
		{
			doc = iter.CurrentDocument();
			break;
		}
	}
	
	if(searchForAndOpen == true)
	{
		//
		// Look for a document somewhere in our search path
		// that has the given unique ID
		//
	}
	
	return doc;
}


//--------------------------------------------------------------------------------
// TApplication::DocumentIndex
//--------------------------------------------------------------------------------
long TApplication::DocumentIndex(Int64 documentID)
{
	long docIndex = 0;
	Boolean found = false;
	
	for(TDocumentLoop iter; iter.More(nil); iter.Next(nil))
	{
		++docIndex;
		
		if(iter.CurrentDocument()->DocumentIdentifier() == documentID)
		{
			found = true;
			break;
		}
	}
	
	if(found == false)
		FailErr(errAENoSuchObject);
	
	return docIndex;
}

//--------------------------------------------------------------------------------
// TApplication::AccessByUniqueID
//--------------------------------------------------------------------------------
TAbstractScriptableObject* TApplication::AccessByUniqueID(TTransaction* t, DescType desiredClass, TDescriptor uniqueID)
{
	TAbstractScriptableObject* result = nil;
	
	if(desiredClass == cDocument)
	{
		result = this->FindDocument(uniqueID.GetInt64(), true);
	}
	else
		result = Inherited::AccessByUniqueID(t, desiredClass, uniqueID);

	return result;
} // TApplication::AccessByUniqueID

//--------------------------------------------------------------------------------
// TApplication::GetProperty
//--------------------------------------------------------------------------------
TDescriptor TApplication::GetProperty(TTransaction* t, DescType propertyName, DescType desiredType, unsigned long additionalInfo)
{
	TDescriptor result;
	
	switch(propertyName)
	{
		case pName:
		{
			//
			// Need to ask the process manager what our name is
			//
			result.MakeString("\pdb+");
			break;
		}
		
		//
		// This is the resolution mode:  use "color" temporarily
		//
		case pColor:
		{
			result.MakeLong(fResolutionMode);
			break;
		}
		
		default:
		{
			result = Inherited::GetProperty(t, propertyName, desiredType, additionalInfo);
		}
	}
	
	return result;
}

//--------------------------------------------------------------------------------
// TApplication::SetProperty
//--------------------------------------------------------------------------------
void TApplication::SetProperty(TTransaction* t, DescType propertyName, TDescriptor& data, unsigned long additionalInfo)
{
	switch(propertyName)
	{
		//
		// This is the resolution mode:  use "color" temporarily
		//
		case pColor:
		{
			this->SetResolutionMode(data.GetLong());			
			break;
		}
		
		default:
		{
			Inherited::SetProperty(t, propertyName, data, additionalInfo);
		}
	}
}

//--------------------------------------------------------------------------------
// TApplication::AECommand
//--------------------------------------------------------------------------------
TDescriptor TApplication::AECommand(TTransaction* t, TAEvent ae, TAEvent reply, long aeCommandID, TAbstractScriptableObject* auxObjects, long auxInfo)
{
	TDescriptor result;
	TDescriptor intermediate;
	TDescriptor descriptor;
	
	switch(aeCommandID)
	{
		case kAEOpen:
		{
			TDescriptor openSpecDesc = ae.GetDescriptor(keyDirectObject);
			FOREACHDESCRIPTOR(&openSpecDesc, descriptor)
			{
				FSSpec openFSSpec;
				descriptor.GetFSS(openFSSpec);
				
				TScriptableDBDocument* document = TScriptableDBDocument::OpenDatabase(openFSSpec);
				intermediate = document->BuildObjectSpecifier(t);
				result.AdoptList(&intermediate);
			}
			openSpecDesc.Dispose();
			break;
		}
		
		default:
		{
			result = Inherited::AECommand(t, ae, reply, aeCommandID, auxObjects, auxInfo);
			break;
		}
	}
	
	return result;
}

//--------------------------------------------------------------------------------
// TApplication::CreateNewElement
//--------------------------------------------------------------------------------
TAbstractScriptableObject* TApplication::CreateNewElement(TTransaction* t, TAEvent ae, TAEvent reply, DescType newObjectClass, TDescriptor initialData, TDescriptor initialProperties, Boolean& usedInitialData, Boolean& usedInitialProperties)
{
	TAbstractScriptableObject* result = nil;
	
	//
	// The application can only create documents
	//
	if(newObjectClass == cDocument)
	{
		//
		// If we create a new document, it will automatically
		// be added to the global documents list
		//
		result = TScriptableDBDocument::NewDatabase();
	}
	else
		result = Inherited::CreateNewElement(t, ae, reply, newObjectClass, initialData, initialProperties, usedInitialData, usedInitialProperties);

	return result;
}

//--------------------------------------------------------------------------------
// TApplication::SetResolutionMode
//--------------------------------------------------------------------------------
void TApplication::SetResolutionMode(long newValue)
{
	switch(newValue)
	{
		//
		// Don't do anything special, not even marking
		// (NOTE:  In this mode, object specifier resolutions
		// that have more than one result will not work, as
		// we do not have code that deals with AE-lists-of-tokens)
		//
		case 0:
			TDescriptor::SetCallbackFlags(kAEIDoMinimum);
			InstallPreResolveProc(nil);
			break;
		
		//
		// Let the OSL resolve all specifiers, including whose clauses
		//
		case 1:
			TDescriptor::SetCallbackFlags(kAEIDoMarking);
			InstallPreResolveProc(nil);
			break;
		
		//
		// OSL does most resolution, but we do whose clauses
		//
		case 2:
			TDescriptor::SetCallbackFlags(kAEIDoMarking + kAEIDoWhose);
			InstallPreResolveProc(nil);
			break;
		
		//
		// We do all object specifier resolution without ever calling OSL
		// (but set the callback flags anyway, just in case).
		//
		case 3:
			TDescriptor::SetCallbackFlags(kAEIDoMarking + kAEIDoWhose);
			InstallPreResolveProc(ResolveObjectSpecifier);
			break;
		
		default:
			newValue = fResolutionMode;
			break;
	}
	fResolutionMode = newValue;
}

