
#include "ScriptableDBDocument.h"

#include "HFSBackingStore.h"
#include "DBElementToken.h"
#include "Application.h"

#include "Exceptions.h"

TScriptableDBDocument* gScriptableDocumentList = nil;

TPropertyDescription TScriptableDBDocument::fPropertiesOfClass[] = {
	{	pIsModified,			0,		typeBoolean,				typeBoolean				}
};

#pragma segment ObjectResident
ImplementClassData(TScriptableDBDocument, clScriptableDBDocument);

//--------------------------------------------------------------------------------
// TScriptableDBDocument::~TScriptableDBDocument
//--------------------------------------------------------------------------------
TScriptableDBDocument::~TScriptableDBDocument()
{
	//
	// Save changes before going away.  Generally speaking, it's bad
	// to do things like this in the destructor.  It's far better to
	// have a Dispose() method that calls all of the methods such as
	// 'save' that may need to be called, and then does a 'delete this'.
	// That way, derived classes can override methods called from the
	// destructor and they will still be called.
	//
	this->Save();
}

//--------------------------------------------------------------------------------
// TScriptableDBDocument::NewDatabase
//--------------------------------------------------------------------------------
TScriptableDBDocument* TScriptableDBDocument::NewDatabase()
{
	TScriptableDBDocument* newDatabase = new TScriptableDBDocument();
	
	return newDatabase;
}

//--------------------------------------------------------------------------------
// TScriptableDBDocument::OpenDatabase
//--------------------------------------------------------------------------------
TScriptableDBDocument* TScriptableDBDocument::OpenDatabase(FSSpec& fileSpec)
{
	THFSBackingStore* backingStore = new THFSBackingStore(fileSpec);
	TScriptableDBDocument* newDatabase = new TScriptableDBDocument(backingStore);
	
	return newDatabase;
}

//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// Overrides of class TDatabaseDocument
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

//--------------------------------------------------------------------------------
// TScriptableDBDocument::DocumentIdentifier
//--------------------------------------------------------------------------------
Int64 TScriptableDBDocument::DocumentIdentifier() const
{
	return TDatabaseDocument::ObjectsKeySpace();
}

//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// Overrides of class TAbstractScriptableObject
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

//--------------------------------------------------------------------------------
// TScriptableDBDocument::GetProperty
//--------------------------------------------------------------------------------
TDescriptor TScriptableDBDocument::GetProperty(TTransaction* t, DescType propertyName, DescType desiredType, unsigned long additionalInfo /*= 0*/)
{
	TDescriptor result;
	
	switch(propertyName)
	{
		case pName:
		{
			//
			// This fixed-length storage isn't too desirable.
			// We should call DocumentName with a tiny TUpdateDataReference
			// just to find out how long the name is, then preallocate
			// a handle to copy the whole data into.
			//
			// This method is good enough, though, because we know
			// that filenames are never anywhere close to 255 characters long.
			//
			char nameStorage[256];
			TUpdataDataReference name(typeChar, nameStorage, 0, 256);
			this->DocumentName(name);
			result.Make(name);
			break;
		}

		case pID:
		{
			result.MakeInt64(this->DocumentIdentifier());
			break;
		}
		
		case pIndex:
		{
			result.MakeLong(gApplication->DocumentIndex(this->DocumentIdentifier()));
			break;
		}
				
		case pIsModified:
		{
			result.MakeBoolean(this->DocumentNeedsSave());
			break;
		}
		
		default:
		{
			result = Inherited::GetProperty(t, propertyName, desiredType, additionalInfo);
			break;
		}
	}
		
	return result;
}

//--------------------------------------------------------------------------------
// TScriptableDBDocument::SetProperty
//--------------------------------------------------------------------------------
void TScriptableDBDocument::SetProperty(TTransaction* transaction, DescType propertyName, TDescriptor& data, unsigned long additionalInfo /*= 0*/)
{
	Inherited::SetProperty(transaction, propertyName, data, additionalInfo);
}

//--------------------------------------------------------------------------------
// TScriptableDBDocument::MakeKeyDataForSelf
//--------------------------------------------------------------------------------
void TScriptableDBDocument::MakeKeyDataForSelf(TTransaction* t, DescType& keyForm, TDescriptor& keyData)
{
	//
	// If this document can be saved, then it has a name.
	//
	if(this->CanSaveDocument())
	{
		Inherited::MakeKeyDataForSelf(t, keyForm, keyData);
	}
	//
	// If we don't have a name, then use our ID
	//
	else
	{
		keyForm = formUniqueID;
		keyData = this->GetProperty(t, pID, typeWildCard, 0); // 
	}
}

//--------------------------------------------------------------------------------
// TScriptableDBDocument::AECommand
//--------------------------------------------------------------------------------
TDescriptor TScriptableDBDocument::AECommand(TTransaction* t, TAEvent ae, TAEvent reply, long aeCommandID, TAbstractScriptableObject* auxObjects /*= nil*/, long auxInfo /*= 0*/)
{
	TDescriptor result;
	
	switch(aeCommandID)
	{
		case kAESave:
		{
			//
			// If the event specified a "save in" parameter, then
			// do a save-as.  Otherwise, just save.
			//
			TDescriptor saveInFile = ae.GetOptionalParameter(keyAEFile);
			if(saveInFile.IsNullDescriptor() == true)
			{
				//
				// n.b. Currently, 'save' will silently fail
				// if no file has been specified.  We can test for
				// this situation by calling this->CanSaveDocument.
				// Should we fail, or make an arbitrary file in 
				// the temp items folder, or...
				//
				this->Save();
			}
			else
			{
				//
				//  n.b. Currently, 'save as' will silently
				// fail if there is already a file with the given
				// fsspec on the disk.  Should we allow save-over?
				// 
				FSSpec saveInFSSpec;
				if(saveInFile.DescriptorType() == typeChar)
				{
					saveInFSSpec.vRefNum = 0;
					saveInFSSpec.parID = 0;
					saveInFile.GetString(saveInFSSpec.name, 63);
				}
				else
					saveInFile.GetFSS(saveInFSSpec);
				this->SaveAs(saveInFSSpec);
			}

			result = this->BuildObjectSpecifier(t);
			break;
		}
		
		case kAEClose:
		{
			//
			// Close can be pretty dangerous if we're multithreaded
			// and some other thread is working with this document.
			// Therefore, we will close this document later.
			//
			this->Save();
			gApplication->MarkForClose(this);
			result = this->BuildObjectSpecifier(t);
			break;
		}
		
		default:
		{
			result = Inherited::AECommand(t, ae, reply, aeCommandID, auxObjects, auxInfo);
			break;
		}
	}
	
	return result;
}

//--------------------------------------------------------------------------------
// TScriptableDBDocument::CreateNewElement
//--------------------------------------------------------------------------------
TAbstractScriptableObject* TScriptableDBDocument::CreateNewElement(TTransaction* transaction, TAEvent ae, TAEvent reply, DescType newObjectClass, TDescriptor initialData, TDescriptor initialProperties, Boolean& usedInitialData, Boolean& usedInitialProperties)
{
	return Inherited::CreateNewElement(transaction, ae, reply, newObjectClass, initialData, initialProperties, usedInitialData, usedInitialProperties);
}

//----------------------------------------------------------------------------------------
// TScriptableDBDocument::RepresentativeScriptingObject
//
// The representative can provide properties for the document.
// All commands identified by SendCommandToRepresentative are also sent to the
// representative
//----------------------------------------------------------------------------------------
TAbstractScriptableObject* TScriptableDBDocument::RepresentativeScriptingObject(TTransaction*)
	{
	return new TDBElementToken(this->GetMetaRoot());
	}

//----------------------------------------------------------------------------------------
// TScriptableDBDocument::SendCommandToRepresentative
//----------------------------------------------------------------------------------------
Boolean TScriptableDBDocument::SendCommandToRepresentative(TTransaction* t, long aeCommandID)
	{
	if((aeCommandID == kAESave) || (aeCommandID == kAEClose))
		return false;
	
	return Inherited::SendCommandToRepresentative(t, aeCommandID);
	} // TScriptableDBDocument::SendCommandToRepresentative
	
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// Methods of TScriptableDBDocument
//::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

//--------------------------------------------------------------------------------
// TScriptableDBDocument::SaveAs
//--------------------------------------------------------------------------------
void TScriptableDBDocument::SaveAs(FSSpec& fileSpec)
{
	//
	// For now, we do not allow scripts to save one database over
	// the top of another.
	//
	if(FSpCreate(&fileSpec, kDBCreator, kDBType, 0) == noErr)
	{
		//
		// 'fileSpec' may contain a relative pathname; call 'FSMakeFSSpec'
		// to make an FSSpec whose name field contains only the name
		// of the file, and none of the path to that file.
		//
		FSSpec actualFSSpec;
		FSMakeFSSpec(fileSpec.vRefNum, fileSpec.parID, fileSpec.name, &actualFSSpec);

		//
		// Create a backing store object to save the database into
		// and call SetBackingStore to set up and save the file.
		//
		THFSBackingStore* backingStore = new THFSBackingStore(actualFSSpec);
		this->SetBackingStore(backingStore);
	}
}
