/*
	File:		List.c

	Contains:	a general unsequenced container for objects.

	Written by:	Bruce Horn, Steve Capps, Larry Kenyon, John Meier, scott douglass, Darin Adler,
				Paul Mercer, Bryan Stearns, Dave Owens
				
				Andy Nicholas, Tom Conrad, Chris Bingham

	Copyright:	 1988-1990, 1993-1995 by Apple Computer, Inc., all rights reserved.

		 <6>	 6/13/95	andy	
*/

#ifdef MWTRACEBACKTABLES
#pragma traceback on
#endif

#include "List.h"
#include "Exceptions.h"
#include "Debug.h"

//
// For BlockMoveData
//
#include <Memory.h>


//========================================================================================
//
// CLASS TList
//
//	A TList is an array of void*s.  The array is stored in a Handle.  The handle
//	can have extra space at the end.  The handle size (in units of pointers, not
//	bytes) is stored in fSize so that we dont have to call GetHandleSize all the
//	time, and increases sizeof(TList) from 6 to 8 bytes (while debugging).  This
//	can be rationalized away by beleving that the memory manager will allocate the
//	same amount of memory whether you ask for 6 bytes or 8.
//
//	While debugging a list contains a count of how many TLoops are active on it so
//	that we can prevent Adds and Removes, etc.
//
//========================================================================================

#define INHERITED

#pragma segment List

//----------------------------------------------------------------------------------------
// TList::~TList: 
//----------------------------------------------------------------------------------------
TList::~TList()
	{
	ASSERT(fNumLoops == 0);

	//
	// Inform the removed items that we no longer care about their delete notification.
	// NOTE: Only use delete notification for TObjectLists!
	// (Duplicate of code in RemoveAllNoTrim)
	//
	if (fUseDeleteNotification)
		{
		FOREACHITEM(this)
			((TObject*)item)->RegisterForDeleteNotification(this, kUnregisterForDeleteNotification);
		}
	} 

//----------------------------------------------------------------------------------------
// TList::Make: 
//----------------------------------------------------------------------------------------
TList* TList::Make(unsigned long guessCount, Boolean /* forceSystemHeap */)
	{
	TList* newList = new TList;

	newList->fObjectBlockCache.Create(guessCount << 2);

	newList->fByteCount = 0;
	newList->fLocked = false;
	newList->fUseDeleteNotification = false;

#ifdef DEBUG
	newList->fShouldNeverSeeDeletion = false;
	newList->fNumLoops = 0;
	newList->fSignature = 'List';
#endif

	return newList;
	}
	


//----------------------------------------------------------------------------------------
// TList::Clone: 
// Create a new list the same size as us and copy our data into it.  If forceCreate is
// false, dont create a new list if the original is empty.  If forceCreate is true,
// always create a new list no matter what.  Note that if a new list is created, it will
// be exactly the same size as the original (for example, if the original list contains
// 4 items but has a Cache big enough for 12 items, so will the clone).
//----------------------------------------------------------------------------------------
TList* TList::Clone(Boolean forceCreate /* = true */)
	{
	unsigned long count = this->Count();										// How big should the new list be?
	TList* newList;

	if (forceCreate || (count > 0))												// Only make a new list if we have to
		{
		newList = TList::Make(count);											// Make a new list of that size
		FailNil(newList);														// Fail out if it didnt work
		newList->SetByteCount(this->fByteCount, false);							// Make sure the sizes are the same

		if (this->fByteCount > 0)												// If theres any data,
			{
			Ptr sourcePtr 	= Start();
			Ptr destPtr		= newList->Start();
			BlockMoveData(sourcePtr, destPtr, this->fByteCount);				// 		copy it
			}
		}
	else																		// If were empty and no forceCreate,
		newList = this;															// 		avoid creating a new object

	return newList;
	} // TList::Clone 



#ifdef AVOIDINLINES

//----------------------------------------------------------------------------------------
// TList::ByteIth: 
//----------------------------------------------------------------------------------------
void* TList::ByteIth(unsigned long byteIndex)
	{
	ASSERT(byteIndex < fByteCount);

	return (void*) fObjectBlockCache.GetUInt32(byteIndex);
	} // TList::ByteIth 

#endif // AVOIDINLINES

//----------------------------------------------------------------------------------------
// TList::ForEachDo: 
//----------------------------------------------------------------------------------------
void TList::ForEachDo(void (*func)(void*))
	{
	FOREACHITEM(this)
		(*func)(item);
	} // TList::ForEachDo 


//----------------------------------------------------------------------------------------
// TList::FindFirstSuchThat: 
//----------------------------------------------------------------------------------------
void* TList::FindFirstSuchThat(Boolean (*func)(const void*))
	{
	FOREACHITEM(this)
		if ((*func)(item)) return item;
	return nil;
	} // TList::FindFirstSuchThat 


//----------------------------------------------------------------------------------------
// TList::ByteIndex: 
//----------------------------------------------------------------------------------------
long TList::ByteIndex(const void* anitem)
	{
	//
	// Shortcut the short case: if we only have one (or no) entries, return 0 (as the index
	// to the only entry) if it matches, or kByteItemNotFound if it doesn't. (We know that if we
	// have no entries, fDatum will be nil).
	//
	if (!Empty())
		{
		FOREACHITEM(this)
			if (anitem == item)
				return loop.CurrentByteIndex();
		}

	return kByteItemNotFound;									// item not found
	} // TList::ByteIndex 

//----------------------------------------------------------------------------------------
// TList::Insert: 
//----------------------------------------------------------------------------------------
void TList::Insert(void* newitem, unsigned long beforeIndex /*= 0*/, unsigned long numItems /*= 1*/)
	{
	ASSERT(newitem);
	ASSERT(fNumLoops == 0);
	ASSERT(beforeIndex <= fByteCount);

	size_t newCount = fByteCount + kItemSize;
	if (ByteSize() < newCount)
		SetSize(newCount);

	void** add = (void**) fObjectBlockCache.GetPtr(beforeIndex);

	if (beforeIndex < fByteCount)								// short-circuit for Append case
		BlockMoveData(add, ((char *) add) + kItemSize * numItems, fByteCount - beforeIndex);
	
	if (numItems == 1)
		*add = newitem;
	else
		//
		// In the case of an add, numItems > 1, newItem is a ptr to list data, 
		// NOT A PTR to an element!
		//
		BlockMoveData( newitem, add, kItemSize*numItems );

	fByteCount = fByteCount + kItemSize * numItems;
	
	//
	// Ask the new item to notify us if its ever deleted.
	// NOTE: Only use delete notification for TObjectLists!
	//
	if (fUseDeleteNotification)
		((TObject*)newitem)->RegisterForDeleteNotification(this, kRegisterForDeleteNotification);
		
	} // TList::Insert 


//----------------------------------------------------------------------------------------
// TList::Concat: 
//----------------------------------------------------------------------------------------
void TList::Concat(TList* append)
	{
	append->Trim();
	long numElements = append->Count();
	Ptr data = (Ptr) append->Start();
	
	this->Insert(data,fByteCount,numElements);
	}


#ifdef AVOIDINLINES
//----------------------------------------------------------------------------------------
// TList::SetSize: 
//----------------------------------------------------------------------------------------
void TList::SetSize(size_t used)
	{
	if (used != ByteSize())
		fObjectBlockCache.SetUsed(used);
	} // TList::SetSize 

	
//----------------------------------------------------------------------------------------
// TList::Append: 
//----------------------------------------------------------------------------------------
void TList::Append(void* add)
	{
	Insert(add, fByteCount);
	} // TList::Append 

//----------------------------------------------------------------------------------------
// TList::Add: 
//----------------------------------------------------------------------------------------
void TList::Add(void* add)
	{
	Append(add); 
	} // TList::Add 
#endif // AVOIDINLINES

//----------------------------------------------------------------------------------------
// TList::AddUnique: 
//----------------------------------------------------------------------------------------
Boolean TList::AddUnique(void* add)
	{
	ASSERT(add);
	ASSERT(fNumLoops == 0);

	Boolean notHere = (ByteIndex(add) == kByteItemNotFound);	// is it present yet?
	if (notHere) Add(add);										// it's not here yet, so add it.
	return notHere;												// return true if we added it, else false.
	} // TList::AddUnique 





//----------------------------------------------------------------------------------------
// TList::RemoveNoTrimNoUnregister: 
//----------------------------------------------------------------------------------------
void TList::RemoveNoTrimNoUnregister(unsigned long index)
	{
	ASSERT(fNumLoops == 0);
	ASSERT(index < fByteCount);
	ASSERT(index == ((index>>2)<<2));

	fByteCount = fByteCount - kItemSize;
	
	//
	// Short-circuit for remove last case
	//
	if (index < fByteCount)		
		{	
		Ptr sourcePtr 	= fObjectBlockCache.GetPtr(index + kItemSize);
		Ptr destPtr		= fObjectBlockCache.GetPtr(index);
		
		BlockMoveData(sourcePtr, destPtr, fByteCount-index);
		}
		
#ifdef DEBUG
	fObjectBlockCache.Set( (UInt32) 0xDEADBEEF, fByteCount);			// clear out at end
#endif
	}

//----------------------------------------------------------------------------------------
// TList::RemoveNoTrim: 
//----------------------------------------------------------------------------------------
void TList::RemoveNoTrim(unsigned long index)
	{
	ASSERT(fNumLoops == 0);
	ASSERT(index < fByteCount);
	ASSERT(index == ((index>>2)<<2));

	//
	// Inform the removed item that we no longer care about its delete notification.
	// NOTE: Only use delete notification for TObjectLists!
	//
	if (fUseDeleteNotification)
		{
		void* item = ByteIth(index);
		((TObject*)item)->RegisterForDeleteNotification(this, kUnregisterForDeleteNotification);
		}	
	
	this->RemoveNoTrimNoUnregister(index);
	}

//----------------------------------------------------------------------------------------
// TList::Remove: 
//----------------------------------------------------------------------------------------
void TList::Remove(unsigned long index)
	{
	RemoveNoTrim(index);
	Trim();
	} 

//----------------------------------------------------------------------------------------
// TList::RemoveNoTrim: 
//----------------------------------------------------------------------------------------
void TList::RemoveNoTrim(const void* item)
	{
	long index = ByteIndex(item);
	if(index != kByteItemNotFound)
		RemoveNoTrim(index);
	} 

//----------------------------------------------------------------------------------------
// TList::Remove: 
//----------------------------------------------------------------------------------------
void TList::Remove(const void* item)
	{
	RemoveNoTrim(item);
	Trim();
	}
	

//----------------------------------------------------------------------------------------
// TList::RemoveAllNoTrim: 
//----------------------------------------------------------------------------------------
void TList::RemoveAllNoTrim()
	{
	ASSERT(fNumLoops == 0);

	//
	// Inform the removed items that we no longer care about their delete notification.
	// NOTE: Only use delete notification for TObjectLists!
	//
	if (fUseDeleteNotification)
		{
		FOREACHITEM(this)
			((TObject*)item)->RegisterForDeleteNotification(this, kUnregisterForDeleteNotification);
		}

#ifdef DEBUG
	// *** should fill with kBadAddress
#endif
	fByteCount = 0;
	}

//----------------------------------------------------------------------------------------
// TList::RemoveAll: 
//----------------------------------------------------------------------------------------
void TList::RemoveAll()
	{
	RemoveAllNoTrim();
	Trim();
	} 


//----------------------------------------------------------------------------------------
// TList::Next: 
//----------------------------------------------------------------------------------------
void* TList::Next(const void* item)
	{
	if (!item) 
		return First();

	long index = ByteIndex(item);

	if (index == kByteItemNotFound)		// *** shouldn't happen because last item was supposed to be in list
		{
		ASSERT(false);
		return nil;
		}

	index = index + kItemSize;			// go to next item

	if (index == fByteCount)			// last item?
		return nil;

	return ByteIth(index);
	} // TList::Next 


//----------------------------------------------------------------------------------------
// TList::Previous: 
//----------------------------------------------------------------------------------------
void* TList::Previous(const void* item)
	{
	if (!item) return Last();

	long index = ByteIndex(item);

	//
	// Shouldn't happen because last item was supposed to be in list
	//
	if ((index == kByteItemNotFound) || (index == 0))
		return nil;

	index = index - kItemSize;				// go to previous item

	return ByteIth(index);
	} // TList::Previous 

//----------------------------------------------------------------------------------------
// TList::Guess: 
//----------------------------------------------------------------------------------------
void TList::Guess(unsigned short newSize)
	{
	newSize <<= 2;

	if (newSize > fByteCount && newSize > ByteSize())
		SetSize(newSize);			// *** should this failure be ignored?
	} // TList::Guess 

//----------------------------------------------------------------------------------------
// TList::Trim: 
//----------------------------------------------------------------------------------------
void TList::Trim(unsigned long count /* = 0 */)
	{
	ASSERT(fByteCount <= ByteSize());

	count  <<= 2;

	if (fByteCount < ByteSize() && fByteCount <= count)
		SetSize(count);
	} // TList::Trim 
	
	
//----------------------------------------------------------------------------------------
// TList::UseDeleteNotification: 
//
// IMPORTANT:  Only call this method on TObjectLists!  (To do:  move this method
// down to TObjectList!)
//----------------------------------------------------------------------------------------
void TList::UseDeleteNotification()
	{
	if (fUseDeleteNotification == false)
		{
		FOREACHITEM(this)
			((TObject*)item)->RegisterForDeleteNotification(this, kRegisterForDeleteNotification);
		fUseDeleteNotification = true;
		}
	} 

//----------------------------------------------------------------------------------------
// TList::ItemBeingDeleted: 
//----------------------------------------------------------------------------------------
void TList::ItemBeingDeleted(const TObject* item)
	{
	ASSERT(fUseDeleteNotification);
	ASSERT(fShouldNeverSeeDeletion == false);
	long index = ByteIndex((void*)item);
	ASSERT(index != kByteItemNotFound);
	
	//
	// If the item being deleted was found inside this list
	// (we expect it will--otherwise we would not have registered
	// for delete notification), then call 'RemoveNoTrimNoUnregister.'
	// We don't want to unregister an item for delete notification
	// if it is being deleted!
	//
	if (index != kByteItemNotFound)	
		RemoveNoTrimNoUnregister(index);
	} 

//----------------------------------------------------------------------------------------
// TList::Replace: 
//
// Replace the item with the given index with newItem
//----------------------------------------------------------------------------------------
void* TList::Replace(unsigned long index, void* newItem)
	{
	INLINEASSERT(index < fByteCount);

	void* oldItem = (void*) fObjectBlockCache.GetUInt32(index);
	fObjectBlockCache.Set( (UInt32) newItem, (UInt32) index);

	return oldItem;
	} // TList::Replace 

//----------------------------------------------------------------------------------------
// TList::AssertNoLoops
//----------------------------------------------------------------------------------------
#ifdef DEBUG
void TList::AssertNoLoops()
	{
	ASSERTPRINT(fNumLoops == 0, ("List has loops"));
	}
#endif

//========================================================================================
// CLASS TDeleteNotificationRegistreeList
//========================================================================================

IMPLEMENTLIST(TDeleteNotificationRegistreeList, TCanReceiveDeleteNotification);


//========================================================================================
// CLASS TObjectList
//========================================================================================


//----------------------------------------------------------------------------------------
// TObjectList::AddUniqueByKey: class TObjectList
//
// *** could use Index and ReplaceByKey
//----------------------------------------------------------------------------------------
Boolean TObjectList::AddUniqueByKey(TObject* add)
	{
	ASSERT(add);
	ASSERT(fNumLoops == 0);
	TObject* existing = Lookup(add->Key());

	if (existing) // found an object with the right key
		{
		// if it's the same object, skip out.
		if (existing == add)
			return false;	// return indication that they are same


		Remove(existing); // not the same object; replace the existing object.
		delete existing;
		}

	Add(add);			// add new object
	return (existing != nil);
	} // TObjectList::AddUniqueByKey 


//----------------------------------------------------------------------------------------
// TObjectList::Lookup: 
//----------------------------------------------------------------------------------------
TObject* TObjectList::Lookup(long key)
	{
	long index = ByteIndex(key);

	return (index == kByteItemNotFound) ? nil : (TObject*) ByteIth(index);
	} // TObjectList::Lookup 


//----------------------------------------------------------------------------------------
// TObjectList::ForEachDo: 
//----------------------------------------------------------------------------------------
void TObjectList::ForEachDo(void (TObject::*func)())
	{
	FOREACHOBJECT(this)
		(object->*func)();
	} // TObjectList::ForEachDo 


//----------------------------------------------------------------------------------------
// TObjectList::FindFirstSuchThat: 
//----------------------------------------------------------------------------------------
TObject* TObjectList::FindFirstSuchThat(Boolean (TObject::*func)(const void*), const void* refCon)
	{
	FOREACHOBJECT(this)
		if ((object->*func)(refCon)) return object;
	return nil;
	} // TObjectList::FindFirstSuchThat


//----------------------------------------------------------------------------------------
// TObjectList::BinarySearch: 
//
// Pass in a function of TObject that returns negative for 'less than', 0 for 'equal', and
// 1 for 'greater than'.  Unfortunately, the item being compared must be stored in a
// global (this should be changed).
//
// Locates the FIRST object in a sorted list for which func returns 0 or greater.
// If your list is not actually sorted by the result of func, then this routine
// obviously won't work correctly.
//
// If the list is empty, we return -1 (no match).  If every item in the list returns
// a number that is less than zero, then we return the index of the last item in the
// list.  If you're looking for an exact match, you'll need to make one more compare
// to find out.
//----------------------------------------------------------------------------------------
short TObjectList::BinarySearch(short (TObject::*func)(const void*), const void* refCon)
	{
	TObject* obj = nil;

#if 0	
	//
	// Special checking:  if the first item in the list
	// returns 0 or greater (thus satisfying the criteria
	// for the search), then return the index of the first
	// item (0) without doing a binary search.
	//
	// I don't think that this test is necessary, nor do I
	// think that it improves the average case.  It was masking
	// a common symptom of a bug in this routine (now fixed),
	// so I took it out.
	//
	//						-ga 11 April 95
	//
	obj = First();
	if ((obj->*func)(refCon) >= 0)
		return 0;
#endif

	//
	// Begin with 'smaller' pointing at the first item in the list,
	// 'bigger' pointing at the last item in the list, and 'middle'
	// pointing at some item in between them (or 'smaller,' if there
	// are only two items in the list).
	//
	unsigned short smaller = 0;
	unsigned short bigger = (unsigned short) Count()-1;	// arbitrary limit
	unsigned short middle = -1; //  = bigger/2;
	
	//
	// Because of the exit conditions of this search (that is, we
	// return the FIRST item that returns zero or greater), we
	// must continue iterating until we converge on a single item.
	// For this part of the loop, though, we will stop when we are
	// down to two items to choose between.
	//
	while (bigger > smaller + 1)
		{
		//
		// 'middle' is the index of the item to compare;
		// always look at the item halfway between the two
		// items.
		//
		middle = (bigger + smaller)/2;
		
		//
		// If the middle object is too small (doesn't match
		// the search criteria), then move the smaller index
		// up to the middle.  If the middle object does
		// match the search criteria, then move 'bigger'
		// down to the middle object.
		//
		// For example, suppose that on the first iteration
		// of the search we have the following situation:
		//
		//		+---+---+---+---+---+
		//		| 0 | 1 | 2 | 3 | 4 |		<-- Index of item
		//		|=??|=??|=??|=??|=??|		<-- value returned by item->func
		//		+---+---+---+---+---+
		//        ^       ^       ^
		//        S       M       B			(smaller, middle and bigger, respectively)
		//
		// If item 2 returns a number less than zero, then we will have:
		//
		//		+---+---+---+---+---+
		//		| 0 | 1 | 2 | 3 | 4 |		<-- Index of item
		//		|=-9|=-7|=-2|=0 |=1 |		<-- value returned by item->func
		//		+---+---+---+---+---+
		//                ^       ^
		//                S       B
		//
		// If item 2 returns a number greater than or equal to zero, then
		// we will have:
		//
		//		+---+---+---+---+---+
		//		| 0 | 1 | 2 | 3 | 4 |		<-- Index of item
		//		|=-1|=0 |=0 |=3 |=4 |		<-- value returned by item->func
		//		+---+---+---+---+---+
		//        ^       ^
		//        S       B
		//
		// On the next iteration, 'middle' will be moved to the index halfway
		// between smaller and bigger.  This will continue until bigger = smaller + 1,
		// at which time middle will always == smaller.
		//
		obj = Ith(middle);
		if ((obj->*func)(refCon) < 0)
			smaller = middle;
		else
			bigger = middle;
		}

	//
	// At this point there are a number of possiblities:
	//
	//	1.	bigger == -1:	This will only happen if the list is empty.
	//		In this case, we return 'bigger' (-1).
	//
	//	2.	bigger == smaller: This will only happen if there is
	//		only one item in the list.  In this case, we return 'bigger' (0).
	//
	//	3.	bigger == smaller + 1:  This will happen for any list
	//		of two items or more.  In that case, middle will be one of:
	//
	//		b.	smaller:	If the smaller item was the middle item on the
	//						last iteration of the loop, then we've already
	//						called 'func', and we know that it returned a
	//						number that was < 0; therefore, we should return
	//						'bigger', because 'smaller' doesn't meet the
	//						criteria of the search.  Note that bigger may or
	//						may not meet the search criteria, but if it does
	//						not, then it must be the last item in the list,
	//						which is what we return when no items match.
	//
	//		c.	bigger:		In this case, we know that the bigger item was the
	//						middle item on the last iteration of the loop,
	//						OR there are only two items in the list.  If there
	//						are more than two items in the list, then we've already
	//						called 'func' on bigger, and we know that it returned
	//						a number that was >= 0 (if bigger->func < 0, then
	//						middle will always == smaller).  If there are two items
	//						in the list, then 'bigger' is always at least an
	//						honorary member of the search set (since we return
	//						'bigger' if there are no matches).
	//
	//						HOWEVER, 'func' was never called for the smaller
	//						item, so 'smaller' may or may not meet the search
	//						criteria.  The only way to find out is to call
	//						'func' on 'smaller', and return 'smaller' iff func
	//						returns zero or greater (otherwise we return 'bigger').
	//
	if(middle == bigger)
		{
		obj = Ith(smaller);
		if ((obj->*func)(refCon) >= 0)
			bigger = smaller;
		} 

	return bigger;
	} // TObjectList::BinarySearch 

//----------------------------------------------------------------------------------------
// TObjectList::ReplaceByKey: 
//----------------------------------------------------------------------------------------
TObject* TObjectList::ReplaceByKey(long key, TObject* newItem)
	{
	ASSERT(newItem);

	long index = ByteIndex(key);
	// *** what if not found?
	return (TObject*) Replace(index, newItem);
	} // TObjectList::ReplaceByKey 


//----------------------------------------------------------------------------------------
// TObjectList::DeleteAllNoTrim: 
//----------------------------------------------------------------------------------------
void TObjectList::DeleteAllNoTrim()
	{
	ASSERT(fNumLoops == 0);

	// ForEachDo(TObject::operator delete);
		{
		FOREACHOBJECT(this)
			delete object;
		}
	RemoveAllNoTrim();
	} // TObjectList::DeleteAllNoTrim 


//----------------------------------------------------------------------------------------
// TObjectList::DeleteAll: 
//----------------------------------------------------------------------------------------
void TObjectList::DeleteAll()
	{
	DeleteAllNoTrim();
	Trim();
	} // TObjectList::DeleteAll 


//----------------------------------------------------------------------------------------
// TObjectList::ByteIndex: 
//----------------------------------------------------------------------------------------
long TObjectList::ByteIndex(long key)
	{
	FOREACHOBJECT(this)
		if (object->Key() == key)
			return loop.CurrentByteIndex();

	return kByteItemNotFound;				// item not found
	} // TObjectList::ByteIndex 

#pragma segment ListSort

//========================================================================================
//	Non-recursive Stable Merge Sort
//
//	Horowitz and Shani (Fundamental Data Structures p.352)
//========================================================================================

//----------------------------------------------------------------------------------------
// Merge: 
//
// Merge two parts of the input lists into the output list
//----------------------------------------------------------------------------------------
void Merge(TObject** list1, TObject** list2, TObject** list2End, TObject** output, void* parm)
	{
	TObject** list1End = list2;

	while ((list1 < list1End) && (list2 < list2End))
		*output++ = (*list1)->SortOrder(*list2, parm) ? *list1++ : *list2++;

	if (list1 < list1End)						// finish up leftover
		BlockMoveData(list1, output, ((Ptr) list1End) - ((Ptr) list1));
	else
		BlockMoveData(list2, output, ((Ptr) list2End) - ((Ptr) list2));
	} // Merge 

//----------------------------------------------------------------------------------------
// SortSection: 
//
// Make one pass of a merge sort by merging adjacent subLists of a given size
//----------------------------------------------------------------------------------------
void SortSection(Ptr input, Ptr inputEnd, long sortSize, Ptr output, void* parm)
	{
	long doubleSize = sortSize * 2;

	TObject** passEnd = (TObject**) (input + sortSize);

	while (input + doubleSize <= inputEnd)
		{
		Merge((TObject**) input, passEnd, (TObject**) (input + doubleSize), (TObject**) output, parm);
		input += doubleSize;
		output += doubleSize;
		passEnd = (TObject**) (input + sortSize);
		}

	if ((Ptr) passEnd < inputEnd)
		Merge((TObject**) input, passEnd, (TObject**) inputEnd, (TObject**) output, parm);
	else
		BlockMoveData(input, output, inputEnd -input);
	} // SortSection 

//----------------------------------------------------------------------------------------
// TObjectList::Sort: 
//----------------------------------------------------------------------------------------
void TObjectList::Sort(void* parm)
	{
	ASSERT(fNumLoops == 0);

	long size = fByteCount;

	if (size <= kItemSize)			// short circuit trivial case because using fData assumes count > 1
		return;

	Ptr buffer = (Ptr) new char[size];
	FailNil(buffer);
	
	fObjectBlockCache.Lock();

	Ptr input = Start();

	Ptr inputEnd = input+size;
	Ptr bufferEnd = buffer+size;

	long sortSize = kItemSize;
	while (sortSize < size)
		{
		SortSection(input,inputEnd, sortSize, buffer, parm);		// move into temporary
		sortSize *= 2;												// double sort size

		SortSection(buffer, bufferEnd, sortSize, input, parm);		// move back into list
		sortSize *= 2;												// double sort size
		}

	fObjectBlockCache.Unlock();
	delete [] buffer;
	} // TObjectList::Sort 

//========================================================================================
// CLASS TLoop
//========================================================================================

//
// Resident segment because it is used from MountNotification
//
#pragma segment Resident

#if 0

//----------------------------------------------------------------------------------------
// TLoop::new: 
//----------------------------------------------------------------------------------------
void* TLoop::operator new(SIZE_T /* size */)
	{
	ASSERTPRINT(false, ("Don't use operator new on TLoops!")); 
	return nil; 	
	} // TLoop::new 

//----------------------------------------------------------------------------------------
// TLoop::delete: 
//----------------------------------------------------------------------------------------
void TLoop::operator delete(void* /* object */)
	{
	ASSERTPRINT(false, ("Don't use operator delete on TLoops!")); 			
	} // TLoop::delete 

#endif

//----------------------------------------------------------------------------------------
// TLoop::TLoop: 
//----------------------------------------------------------------------------------------
TLoop::TLoop(TList* list)
	{
	ASSERT(list);

	fList = list;
#ifdef DEBUG
	fSignature = 'Loop';
	fList->fNumLoops++;
#endif
	fListPreviouslyLocked = list->fLocked;						// remember whether the list was locked or not
	if (!fListPreviouslyLocked)									// if no one else has locked it
		fList->LockList();										// ...then lock it

#ifndef METROWERKS	
	ASSERT(gExceptionStack);
	if (gExceptionStack)
		gExceptionStack->AddExceptionAction(this);
#endif

	Reset();
	} // TLoop::TLoop 


//----------------------------------------------------------------------------------------
// TLoop Locking strategy:
//
//		Looping is sped up by making the loop a pointer walk.
//		This requires the list to be locked for the duration of the loop.
//		Normally, the list is locked in TLoop::TLoop and unlocked in TLoop::~TLoop.
//		To handle multiple simultaneous loops on the same list, a flag in TList
//		indicates if the list is currently locked.  When a loop begins with
//		TLoop::TLoop, it remembers the state of this flag, so it knows whether
//		or not it should unlock the list when its done (you dont want to unlock
//		it if some other loop is still running and thus needs it locked!).
//
//		HOWEVER, under certain circumstances (exceptions, etc.) TLoop::~TLoop is
//		skipped and thus the list will remain locked for all time (bad).  To
//		prevent this, whenever a list is created, its linked into a list pointed
// 		to by the current TException frame.  When someone calls Fail(), Fail()
//		will ensure that all loops in the list hanging off the current TException
//		are unlocked.
//
//		NOTE that compilers differ in their implementation of local destructors
//		vis-a-vis longjmp.  CFront and xlC DO NOT call destructors for local objects
//		in the scope out of which were longjmping.  Metrowerks DOES call destructors
//		for local objects and thus it is not necessary to call Add/RemoveLoop for
//		Metrowerks builds.
//
//		To minimize the compiler differences and generally play it safe, we always call
//		Cleanup() when we fail.  Cleanup() handles unlocking the list and sets fList to
// 		nil to prevent further references to it.  Further references are unsafe since
//		with Metrowerks the loop could conceivably be destroyed after the list has already
// 		been deleted.  Further references are also unnecessary since Cleanup() has done
//		the unlocking.
//
//		-bing 5/20/94
//
//----------------------------------------------------------------------------------------


//----------------------------------------------------------------------------------------
// TLoop::~TLoop: 
//----------------------------------------------------------------------------------------
TLoop::~TLoop()
	{
	if (fList != nil)										// Make we havent been cleaned up already
		{
#ifndef METROWERKS	
		ASSERT(gExceptionStack);
		if (gExceptionStack)
			gExceptionStack->RemoveExceptionAction(this);	// Remove this loop from the TException
#endif // METROWERKS

#ifdef DEBUG
		fList->fNumLoops--;
#endif
		if (!fListPreviouslyLocked)							// if we locked this list
			fList->UnlockList();							// ...then unlock it
		}
	} 


//----------------------------------------------------------------------------------------
// TLoop::CatchException:
//
// Walk down the list of loops, unlocking all of them
//
// CatchException is a substitute for the TLoop destructor and is called from the failure
// handler of TException
//----------------------------------------------------------------------------------------
void TLoop::CatchException()			
	{
	ASSERT(fList);
	
#ifdef DEBUG
	fList->fNumLoops--;
#endif

	if (!fListPreviouslyLocked)			// if we locked this list
		fList->UnlockList();			// then unlock it
		
	fList = nil;						// Prevent future references to our list (see long comment above)
	}



//----------------------------------------------------------------------------------------
// TLoop::Reset: 
//----------------------------------------------------------------------------------------
void TLoop::Reset()			
	{ 
	fFirstIndex = (long) fList->Start();						// the first element in the list
	fByteIndex 	= fFirstIndex - kItemSize;						// just before the first element
	fLastIndex 	= fByteIndex + fList->ByteCount();				// the last element in the list
	} // TLoop::Reset 


//----------------------------------------------------------------------------------------
// TLoop::Next: 
//----------------------------------------------------------------------------------------
void* TLoop::Next()
	{
	ASSERT(fByteIndex >= (fFirstIndex - kItemSize) && fByteIndex <= fLastIndex);
	if (fByteIndex >= fLastIndex)
		return nil;
	fByteIndex = fByteIndex + kItemSize;
	return Current();
	} // TLoop::Next 


//----------------------------------------------------------------------------------------
// TLoop::Previous: 
//----------------------------------------------------------------------------------------
void* TLoop::Previous()
	{
	ASSERT(fByteIndex >= fFirstIndex && fByteIndex <= (fLastIndex + kItemSize));
	if (fByteIndex <= fFirstIndex)
		return nil;
	fByteIndex = fByteIndex - kItemSize;
	return Current();
	} // TLoop::Previous 
	

//----------------------------------------------------------------------------------------
// TLoop::MovePointerForward: Same as Next(), but doesn't return the item
//----------------------------------------------------------------------------------------
Boolean TLoop::MovePointerForward()
	{
	ASSERT(fByteIndex >= (fFirstIndex - kItemSize) && fByteIndex < fLastIndex);
	if (fByteIndex >= fLastIndex)
		return false;
	fByteIndex = fByteIndex + kItemSize;
	return true;
	} // TLoop::MovePointerForward 


//----------------------------------------------------------------------------------------
// TLoop::MovePointerBackward: Same as Previous(), but doesn't return the item
//----------------------------------------------------------------------------------------
Boolean TLoop::MovePointerBackward()
	{
	ASSERT(fByteIndex > fFirstIndex && fByteIndex <= (fLastIndex + kItemSize));
	if (fByteIndex <= fFirstIndex)
		return false;
	fByteIndex = fByteIndex - kItemSize;
	return true;
	} // TLoop::MovePointerBackward 
	

//----------------------------------------------------------------------------------------
// TLoop::SetPointerToFirst: Set the current pointer to the first item in the list
//----------------------------------------------------------------------------------------
Boolean TLoop::SetPointerToFirst()
	{
	if (fLastIndex < fFirstIndex)								// Return false on empty list
		return false;
	fByteIndex = fFirstIndex;
	return true;
	} // TLoop::SetPointerToFirst 
	

//----------------------------------------------------------------------------------------
// TLoop::SetPointerToLast: Set the current pointer to the last item in the list
//----------------------------------------------------------------------------------------
Boolean TLoop::SetPointerToLast()			
	{
	if (fLastIndex < fFirstIndex)								// Return false on empty list
		return false;
	fByteIndex = fLastIndex;
	return true;
	} // TLoop::SetPointerToLast 
	

//----------------------------------------------------------------------------------------
// TLoop::ResetToEnd: 
// 
// Useful to set up for a backwards loop with the pointer past the last element in the list,
// just as Reset sets up for a forwards loop with the pointer before the first element.
//----------------------------------------------------------------------------------------
void TLoop::ResetToEnd()			
	{ 
	fFirstIndex = (long) fList->Start();							// the first element in the list
	fLastIndex 	= fFirstIndex + fList->ByteCount() - kItemSize;		// the last element in the list
	fByteIndex 	= fLastIndex + kItemSize;							// just after the last element
	} // TLoop::ResetToEnd 

//----------------------------------------------------------------------------------------
// TLoop::RemoveCurrent: 
//----------------------------------------------------------------------------------------
void TLoop::RemoveCurrent()
	{
	ASSERT(fList->fNumLoops == 1);

#ifdef DEBUG
	fList->fNumLoops--;		// allow TList::Remove()
#endif
	fList->RemoveNoTrim(fByteIndex - fFirstIndex);				// The byte index is really a pointer to the current item
#ifdef DEBUG
	fList->fNumLoops++;		// disallow again
#endif
	fByteIndex = fByteIndex - kItemSize;
	fLastIndex = fLastIndex - kItemSize;
	} // TLoop::RemoveCurrent 

IMPLEMENTLOOP(TObjectLoop, TObjectList, TObject)


//----------------------------------------------------------------------------------------
// TBackwardLoop::Current: 
//----------------------------------------------------------------------------------------
void* TBackwardLoop::Current()
	{
	INLINEASSERT(fByteIndex >= kByteItemNotFound && fByteIndex < fList->ByteCount());

	return (fByteIndex >= 0) ? fList->ByteIth(fByteIndex) : nil;
	} // TBackwardLoop::Current 


//----------------------------------------------------------------------------------------
// TBackwardLoop::Next: 
//----------------------------------------------------------------------------------------
void* TBackwardLoop::Next()
	{
	INLINEASSERT(fByteIndex >= 0 && fByteIndex <= fList->ByteCount());

	fByteIndex = fByteIndex - kItemSize;
	return Current();
	} // TBackwardLoop::Next 
	


