/*
	File:		List.h

	Contains:	Lists of void*s, lists of TObject*s, loops thereof, stacks and queues.
				Doesnt allow adding items that are nil.

				TLists are usually unordered, TObjectLists can be sorted.
				In general use TObjectList (or better, DECLAREOBJECTLIST) for items that
				are derived from TObject; use TList (or better, DECLARELIST) for items
				that are not derived from TObject.

	Written by:	Bruce Horn, Steve Capps, Larry Kenyon, John Meier, scott douglass, Darin Adler,
				Paul Mercer, Bryan Stearns, Dave Owens
				
				Andy Nicholas, Greg Anderson, Chris Bingham

	Copyright:	 1988-1995 by Apple Computer, Inc., all rights reserved.

		 <4>	  6/6/95	ga		
*/

#ifndef List_h
#define List_h

#include <stddef.h>

//
// We need exceptions.h so that TLoop can
// inherit from TExceptionAction
//
#include "Exceptions.h"

//
// For object lists
//
#include "Object.h"

//
// For object block cache
//
#include "ObjectCache.h"

enum {
	kItemSize 				= sizeof(unsigned long),
	kItemNotFound			= -1, 						// (to be used where you're expecting a index of bytes)
	kByteItemNotFound		= kItemNotFound * kItemSize // (to be used when you're expecting an index of entries)
	};

//========================================================================================
// CLASS List
//
// List of void*s
//
//	Add, Append, Insert, Remove and RemoveAll cant be called while the list has a TLoop
//	in scope but see TLoop::RemoveCurrent().
//
//========================================================================================

class TList : public TCanReceiveDeleteNotification
	{
	
	// --- Methods -------------------------------------------------------------------------------------
public:
	static TList* 		Make();
	static TList* 		MakeInSystemHeap();
	static TList* 		Make(unsigned long guessCount);
	static TList* 		Make(unsigned long guessCount, Boolean inSystemHeap);
						~TList();

	void				Constructor();

	TObjectBlockCache	fObjectBlockCache;

	TList*				Clone(Boolean forceCreate = true);		// Return a new list identical to ours
	
	long				Count();								//	return the number of items in the list
	long				ByteCount();							//	return the byte size of items in the list
	long				LastByteIndex();						//	return the byte index of the last item
	Boolean				Empty();								//	return true if list is empty
	void*				First();								//	return the first item in the list (or nil if none)
	void*				Last();									//	return the last item in the list (or nil if none)
	void*				Next(const void*);						//	return the item after the first occurance of the given item (or nil if none)
	void*				Previous(const void*);
	void*				Ith(unsigned short index);				//	return the ith item
	void*				ByteIth(unsigned long byteIndex);		//	return the ith item

	void				ForEachDo(void (*func)(void*));			// call func passing it each item in the list
	void*				FindFirstSuchThat(Boolean (*func)(const void*));// call func passing it each item in the list until func
																// returns true, returns the item or nil if none

	void				Add(void*);								//	add an item to the list (order unspecified)
	void				Append(void*);							//	append an item to the list
	void				Insert(void*, unsigned long beforeIndex = 0, unsigned long numItems = 1);	//	insert an item before the given index
	Boolean				AddUnique(void*);						// add an item if it's not already present (& return true); do nothing & return false if not
	void 				Concat(TList* append);					//concat the contents of append onto end of this

	void				RemoveNoTrim(unsigned long);			//	remove the item with the given index (does not Trim())
	void				Remove(unsigned long);					//	remove the item with the given index (does Trim())
	void				RemoveNoTrim(const void*);				//	remove the first instance of a given item (does not Trim())
	void				Remove(const void*);					//	remove the first instance of a given item (does Trim())
	void				RemoveAllNoTrim();						//	empty the list without deleting its objects (does not Trim())
	void				RemoveAll();							//	empty the list without deleting its objects (does Trim())

	void*				Replace(unsigned long, void* newItem);	//	replace the item with the given index with newItem; returns the item replaced
	void				Replace(const void* oldItem, void* newItem); //	replace the first instance of oldItem with newItem

	long				ByteIndex(const void*);					//	find the first index of an item (or kByteItemNotFound if not found)
	Boolean				Contains(const void*);					// is this item in the list at all?
	short				Index(const void*);

	void				LockList();								// Lock the list, set fLocked
	void				UnlockList();							// Unlock the list, clear fLocked, and remove from the list of locked lists
	Boolean				fLocked;								// Remember if were currently locked (saves calling memory manager just to find out)

	void				SetByteCount(size_t, Boolean trim);
	void				SetCount(unsigned short count);
	void				SetCountNoTrim(unsigned short count);
	void				Guess(unsigned short newSize);			// guess that there are going to be count items, will fail if there isnt enough memory
	void				Trim(unsigned long count = 0);			//	trim off any wasted guessing space

	Ptr					Start()									{ return fObjectBlockCache.Start();	}

	void				UseDeleteNotification();				// Tell this list to register its members for delete notification
	virtual void		ItemBeingDeleted(const TObject*);		// Call this when a member of this list is being deleted
#ifdef DEBUG
	void				ShouldNeverSeeDeletion();				// Tell this list that we don't expect to actually use delete notification
	void				AssertNoLoops();						// ASSERT if someone is looping on this list
#endif


private:
	void				RemoveNoTrimNoUnregister(unsigned long);
	void				SetSize(size_t newSize);				//	change the size of fData
	size_t				ByteSize();

	//
	// --- Fields -------------------------------------------------------------------------------------
	//
protected:

friend class TLoop;
	unsigned			fUseDeleteNotification : 1;				// Does this list receive delete notifications for its items?
#ifdef DEBUG
	unsigned			fShouldNeverSeeDeletion : 1;			// Expect that objects should never be deleted if in this list
	unsigned long		fSignature;								//	= 'List'
	unsigned long		fNumLoops;								// number of TLoops active on this list
#endif
	size_t				fByteCount;								// the byte count of items in the list
	};

//----------------------------------------------------------------------------------------
// TList::Constructor: 
//----------------------------------------------------------------------------------------
inline void TList::Constructor()
	{
#ifdef DEBUG
	fSignature = 'List';
#endif
	}

//----------------------------------------------------------------------------------------
// TList::ByteSize: 
//----------------------------------------------------------------------------------------
inline size_t TList::ByteSize() 						
	{ 
	return fObjectBlockCache.Used();
	}

#ifndef AVOIDINLINES
//----------------------------------------------------------------------------------------
// TList::ByteIth: 
//----------------------------------------------------------------------------------------	
inline void* TList::ByteIth(unsigned long byteIndex)	
	{ 
	return (void*) fObjectBlockCache.GetUInt32(byteIndex);
	}
#endif // AVOIDINLINES

//----------------------------------------------------------------------------------------
// TList::ByteCount: 
//----------------------------------------------------------------------------------------		
inline long TList::ByteCount()						
	{
	return fByteCount;
	}

//----------------------------------------------------------------------------------------
// TList::Count: 
//----------------------------------------------------------------------------------------			
inline long TList::Count()							
	{
	return fByteCount >> 2;
	}

//----------------------------------------------------------------------------------------
// TList::Empty: 
//----------------------------------------------------------------------------------------			
inline Boolean TList::Empty()							
	{
	return (fByteCount == 0);
	}

//----------------------------------------------------------------------------------------
// TList::LastByteIndex: 
//----------------------------------------------------------------------------------------			
inline long TList::LastByteIndex()					
	{
	return fByteCount-4;
	}

//----------------------------------------------------------------------------------------
// TList::Ith: 
//----------------------------------------------------------------------------------------			
inline void* TList::Ith(unsigned short index)		
	{
	return ByteIth(index << 2);
	}

//----------------------------------------------------------------------------------------
// TList::Index: 
//----------------------------------------------------------------------------------------			
inline short TList::Index(const void* item)			
	{
	return ByteIndex(item) >> 2;
	}

//----------------------------------------------------------------------------------------
// TList::Contains: 
//----------------------------------------------------------------------------------------			
inline Boolean TList::Contains(const void* item)
	{
	return (ByteIndex(item) != kByteItemNotFound);
	} 

#ifndef AVOIDINLINES
//----------------------------------------------------------------------------------------
// TList::SetSize: 
//----------------------------------------------------------------------------------------
inline void TList::SetSize(size_t used)
	{
	if (used != ByteSize())
		fObjectBlockCache.SetUsed(used);
	} 
#endif

//----------------------------------------------------------------------------------------
// TList::SetByteCount: 
//----------------------------------------------------------------------------------------
inline void TList::SetByteCount(size_t newByteCount, Boolean trim)
	{
	if (trim || (newByteCount > fByteCount))
		SetSize(newByteCount);
	fByteCount = newByteCount;
	}

//----------------------------------------------------------------------------------------
// TList::LockList: 
//----------------------------------------------------------------------------------------
inline void TList::LockList()
	{
	fLocked = true;								// remember that we locked it
	fObjectBlockCache.Lock();
	} // TList::LockList 

//----------------------------------------------------------------------------------------
// TList::UnlockList: 
//----------------------------------------------------------------------------------------
inline void TList::UnlockList()
	{
	fLocked = false; 							// and remember that its unlocked
	fObjectBlockCache.Unlock();
	} // TList::UnlockList 

//----------------------------------------------------------------------------------------
// TList::Make: 
//----------------------------------------------------------------------------------------
inline TList* TList::Make()
	{
	return Make(0, false);
	} // TList::Make 

//----------------------------------------------------------------------------------------
// TList::Make: 
//----------------------------------------------------------------------------------------
inline TList* TList::Make(unsigned long  guessCount)
	{
	return Make(guessCount, false);
	} // TList::Make 

//----------------------------------------------------------------------------------------
// TList::MakeInSystemHeap: 
//----------------------------------------------------------------------------------------
inline TList* TList::MakeInSystemHeap()
	{
	return Make(0, true);
	} // TList::MakeInSystemHeap 

#ifndef AVOIDINLINES
//----------------------------------------------------------------------------------------
// TList::Append: 
//----------------------------------------------------------------------------------------
inline void TList::Append(void* add)
	{
	Insert(add, fByteCount);
	} // TList::Append 

//----------------------------------------------------------------------------------------
// TList::Add: 
//----------------------------------------------------------------------------------------
inline void TList::Add(void* add)
	{
	Append(add); 
	} // TList::Add 
#endif

//----------------------------------------------------------------------------------------
// TList::Replace: 
//
// Replace the first instance of oldItem with newItem
//----------------------------------------------------------------------------------------
inline void TList::Replace(const void* oldItem, void* newItem)
	{
	long index = ByteIndex(oldItem);
	// *** what if not found?
	Replace(index, newItem);
	} // TList::Replace 

//----------------------------------------------------------------------------------------
// TList::First: 
//----------------------------------------------------------------------------------------
inline void* TList::First()
	{
	return Empty() ? nil : ByteIth(0);
	} // TList::First 

//----------------------------------------------------------------------------------------
// TList::Last: 
//----------------------------------------------------------------------------------------
inline void* TList::Last()
	{
	return Empty() ? nil : ByteIth(LastByteIndex());
	} // TList::Last 

//----------------------------------------------------------------------------------------
// TList::SetCount: 
//----------------------------------------------------------------------------------------
inline void TList::SetCount(unsigned short newCount)
	{
	SetByteCount(newCount << 2, true);
	} // TList::SetCount 

//----------------------------------------------------------------------------------------
// TList::SetCountNoTrim: 
//----------------------------------------------------------------------------------------
inline void TList::SetCountNoTrim(unsigned short newCount)
	{
	SetByteCount(newCount << 2, false);
	} // TList::SetCountNoTrim 

#ifdef DEBUG
//----------------------------------------------------------------------------------------
// TList::ShouldNeverSeeDeletion: 
//----------------------------------------------------------------------------------------
inline void TList::ShouldNeverSeeDeletion()
	{
	fShouldNeverSeeDeletion = true;
	if(fUseDeleteNotification == false)
		this->UseDeleteNotification();
	} // TList::ShouldNeverSeeDeletion 
#endif

//----------------------------------------------------------------------------------------
// The DECLARELIST macro can be used in a header file to declare a new list type.
//	Using it avoids having to cast return values from TList member functions.
//	e.g. DECLARELIST(TZoneList, TZone) declares TZoneList to be a list of pointers
//	to TZones (i.e. TZone*s).  Use IMPLEMENTLIST in the cooresponding .c file.
//	See also DECLAREOBJECTLIST, DECLARELOOP, DECLARESTACK and DECLAREQUEUE.
//----------------------------------------------------------------------------------------

#define DECLARELISTBASICOPERATIONS(listtype, itemtype)															\
	public:																										\
		static	listtype*	Make()								{ return (listtype*) TList::Make(); }			\
		static	listtype*	Make(unsigned short guessCount) 	{ return (listtype*) TList::Make(guessCount); }	\
				listtype*	Clone(Boolean forceCreate = true)	{ return (listtype*) TList::Clone(forceCreate); }	\
				itemtype*	First()								{ return (itemtype*) TList::First(); }			\
				itemtype*	Last()								{ return (itemtype*) TList::Last(); }			\
				itemtype*	Next(const itemtype* item)			{ return (itemtype*) TList::Next((void *)item); }		\
				itemtype*	Previous(const itemtype* item)		{ return (itemtype*) TList::Previous((void *)item); }	\
				itemtype*	Ith(unsigned short index)			{ return (itemtype*) TList::Ith(index); }		\
				itemtype*	ByteIth(unsigned long byteIndex)	{ return (itemtype*) TList::ByteIth(byteIndex); } \
				long		ByteIndex(const itemtype* item)		{ return TList::ByteIndex((void *)item); }					

#define DECLARELIST(listtype, itemtype)					\
	class listtype : public TList						\
		{												\
		DECLARELISTBASICOPERATIONS(listtype, itemtype);	\
		};

//----------------------------------------------------------------------------------------
//	The IMPLEMENTLIST macro can be used in a .c file to implement a new list type.
//	e.g. IMPLEMENTLIST(TZoneList, TZone)
//----------------------------------------------------------------------------------------

#define IMPLEMENTLIST(listtype, itemtype)


//========================================================================================
// CLASS TDeleteNotificationRegistreeList
//========================================================================================

DECLARELIST(TDeleteNotificationRegistreeList, TCanReceiveDeleteNotification);


typedef short (*ScrewyFnPtr)(void*);

//========================================================================================
// CLASS TObjectList
//
// List of TObject*s
//========================================================================================

class TObjectList : public TList
	{
	DECLARELISTBASICOPERATIONS(TObjectList, TObject)

	long		ByteIndex(long);						// find the first index of an item with a given key (or kByteItemNotFound if not found)
	TObject*	Lookup(long);							//	return the first item with a given key (or nil if not found)
	Boolean		AddUniqueByKey(TObject*);				//add an item if its key is unique in the list; returns true iff the item was added
	void		ForEachDo(void (TObject::*func)());
	TObject*	FindFirstSuchThat(Boolean (TObject::*func)(const void*), const void*);
	short		BinarySearch(short (TObject::*func)(const void*), const void*);
		
	TObject*	ReplaceByKey(long, TObject* newItem);	// replace the first item that has a given key with newItem; returns the item replaced or nil
		
	void		DeleteAllNoTrim();						//	operator delete & remove all of the items in the list (does not Trim())
	void		DeleteAll();							// operator delete & remove all of the items in the list (does Trim())
		
	void 		Sort(void*);							//stable sort using ptr to specify sort key
	};

//----------------------------------------------------------------------------------------
//	The DECLAREOBJECTLIST macro can be used in a header file to declare a new list type.
//	Using it avoids having to cast return values from TObjectList member functions.
//	e.g. DECLAREOBJECTLIST(TIconList, TIcon) declares TIconList to be a list of pointers
//	to TIcons (i.e. TIcon*s).  Use IMPLEMENTOBJECTLIST in the cooresponding .c file.
//	See also DECLARELIST, DECLARELOOP, DECLARESTACK and DECLAREQUEUE.
//----------------------------------------------------------------------------------------

#define DECLAREOBJECTLIST(listtype, itemtype)					\
	class listtype : public TObjectList							\
		{														\
		DECLARELISTBASICOPERATIONS(listtype, itemtype)			\
		itemtype*	Lookup(long key)							{ return (itemtype*) TObjectList::Lookup(key); }		\
		void		ForEachDo(void (itemtype::*func)())			{ TObjectList::ForEachDo((void (TObject::*)())func); }	\
		short		BinarySearch(short (itemtype::*func)(const void*), const void* refCon)	\
								{ return TObjectList::BinarySearch((short (TObject::*)(const void*))func, refCon); }	\
		};														


//		itemtype*	FindFirstSuchThat(Boolean (itemtype::*func)(const void*), const void* refCon)	\
//								{ return (itemtype*) TObjectList::FindFirstSuchThat((Boolean (TObject::*)(const void*))func, refCon); }	\


//----------------------------------------------------------------------------------------
//	The IMPLEMENTOBJECTLIST macro can be used in a .c file to implement a new list type.
//	e.g. IMPLEMENTOBJECTLIST(TIconList, TIcon)
//----------------------------------------------------------------------------------------

#define IMPLEMENTOBJECTLIST(listtype, itemtype)


//========================================================================================
// CLASS Loop
//
//  Iterate a TList
//
//	TList::Add, TList::Append, TList::Insert, TList::Remove and TList::RemoveAll
//	cant be called while the list has a TLoop in scope but TLoop::RemoveCurrent() can
//	be called if there is only one TLoop for the list in scope.
//
//	If TLoop had a primary base class, then it should mix in (muiltiply-inherit from)
//	TExceptionAction; however, since TLoop previously had no base class, it derives
//	directly from TExceptionAction.
//
//========================================================================================

class TLoop : public TExceptionAction
	{

	//
	// --- Methods -------------------------------------------------------------------------------------
	//
public:
				TLoop(TList* list);
				~TLoop();

#if 0
	void* 		operator new(SIZE_T /* size */);// These give warnings that operator new and operator delete are
	void 		operator delete(void* object);	// bad things to do to TLoops
#endif
	
	void		Reset();
	void*		Next();							// Moves pointer to next item and returns it (or nil if at end of list)
	void*		Previous();						// Moves pointer to previous item and returns it (or nil if at start of list)
	Boolean		MovePointerForward();			// Moves pointer to next item, returning false if at the end of list
	Boolean		MovePointerBackward();			// Moves pointer to previous item, returning false if at the start of list
	Boolean		SetPointerToFirst();			// Set the current pointer to the first
	Boolean		SetPointerToLast();				// ...or last item in the list (returns false if list is empty)
	void		ResetToEnd();					// Resets for a backwards loop
	void*		Current();						// Warning! Does no bounds checking! (This is handled by Next(), Previous(), etc.)
	void		RemoveCurrent();

	long		CurrentByteIndex();				// Returns offset (not actual pointer) to current item
	void		SetByteIndex(long newIndex);
	void		SetLastIndex(long newLast);		

	virtual void		CatchException();

	//
	// --- Fields -------------------------------------------------------------------------------------
	//	
private:
#ifdef DEBUG
	long		fSignature;						// = 'Loop'
#endif
	TList*		fList;
	long		fFirstIndex;					// A pointer to the first item
	long		fLastIndex;						// a pointer to the last item
	long		fByteIndex;						// a pointer to the current item
	TLoop*		fNextLoop;						// A pointer to the next loop, or nil
	Boolean		fListPreviouslyLocked;			// Was the list already locked before we got it?
	};


	
//----------------------------------------------------------------------------------------
// TLoop::Current: 
//----------------------------------------------------------------------------------------		
inline void* TLoop::Current()					
	{
	return *((void**)fByteIndex); 		
	}

//
// The next three methods maintain backwards compatibility by turning the internal ByteIndex
// pointer into an integer offset and back -- since in the old TLoop ByteIndex was an offset,
// not a pointer.
//

//----------------------------------------------------------------------------------------
// TLoop::CurrentByteIndex: 
//----------------------------------------------------------------------------------------		
inline long TLoop::CurrentByteIndex()			
	{
	return (fByteIndex - fFirstIndex); 	
	}
	
//----------------------------------------------------------------------------------------
// TLoop::SetByteIndex: 
//----------------------------------------------------------------------------------------		
inline void TLoop::SetByteIndex(long newIndex)
	{
	fByteIndex = newIndex + fFirstIndex; 	
	}
	
//----------------------------------------------------------------------------------------
// TLoop::SetLastIndex: 
//----------------------------------------------------------------------------------------		
inline void TLoop::SetLastIndex(long newLast) 	
	{
	fLastIndex = newLast + fFirstIndex; 	
	}

//----------------------------------------------------------------------------------------
//	The DECLARELOOP macro can be used in a header file to declare a new iterator type.
//	Using it avoids having to cast return values from TLoop member functions.
//	e.g. DECLARELOOP(TIconLoop, TIconList, TIcon) declares TIconLoop to be an iterator
//	for a TIconList.  Use IMPLEMENTLOOP in the cooresponding .c file.
//	See also DECLARELIST, DECLAREOBJECTLIST, DECLARESTACK and DECLAREQUEUE.
//----------------------------------------------------------------------------------------

#define DECLARELOOP(newlooptype, listtype, itemtype)	\
	class newlooptype : public TLoop					\
		{												\
	public:												\
		newlooptype(listtype* list) : TLoop(list) {}	\
														\
		itemtype*	Next();								\
		itemtype*	Previous();							\
		itemtype*	Current();							\
		};												\
														\
	inline itemtype* newlooptype::Next()				\
		{ return (itemtype*) TLoop::Next(); }			\
														\
	inline itemtype* newlooptype::Current()				\
		{ return (itemtype*) TLoop::Current(); }		\
														\
	inline itemtype* newlooptype::Previous()			\
		{ return (itemtype*) TLoop::Previous(); }

//----------------------------------------------------------------------------------------
//	The IMPLEMENTLOOP macro can be used in a .c file to implement a new iterator type.
//	e.g. IMPLEMENTLOOP(TIconLoop, TIconList, TIcon)
//----------------------------------------------------------------------------------------

#define IMPLEMENTLOOP(newlooptype, listtype, itemtype)

//----------------------------------------------------------------------------------------
//
//	The FOREACH macro sets up an iterator over a list.  It is followed by a statement
//	or block (which can contain break).  It defines an iterator variable loop in the
//	current scope.
//
//	e.g.  (warning this example is very contrived)
//
//		FOREACH(TIconLoop, Children(), TIcon, icon)
//			{
//			if (icon == myFavoriteIcon)
//				{
//				if (icon->Marked()) loop.RemoveCurrent();
//				break;
//				}
//			}
//
//	See also FOREACHITEM(), FOREACHOBJECT, TList::ForEachDo(), TList::FindFirstSuchThat(),
//	TObjectList::ForEachDo() and TObjectList::FindFirstSuchThat().
//----------------------------------------------------------------------------------------

#define FOREACH(looptype, listvar, itemtype, itemvar)	\
	looptype loop(listvar); itemtype* itemvar;			\
	while ((itemvar = loop.Next()) != nil) 


//----------------------------------------------------------------------------------------
//	The FOREACHITEM macro sets up an iterator over a TObjectList.  It is followed by
//	a statement or block (which can contain break).  It defines an iterator variable
//	loop an the item variable object in the current scope.
//
//	See also FOREACH(), FOREACHOBJECT(), TList::ForEachDo(), TList::FindFirstSuchThat(),
//	TObjectList::ForEachDo() and TObjectList::FindFirstSuchThat().
//----------------------------------------------------------------------------------------

#define FOREACHITEM(list) FOREACH(TLoop, list, void, item)


DECLARELOOP(TObjectLoop, TObjectList, TObject)

// Declare a type thats a list of lists
DECLARELIST(TListList, TList)
DECLARELOOP(TListLoop, TListList, TList)
#define FOREACHLIST(listlist)	FOREACH(TListLoop, listlist, TList, list)


//----------------------------------------------------------------------------------------
//	The FOREACHOBJECT macro sets up an iterator over a TObjectList.  It is followed by
//	a statement or block (which can contain break).  It defines an iterator variable
//	loop an the item variable object in the current scope.
//
//	See also FOREACH(), FOREACHITEM(), TList::ForEachDo(), TList::FindFirstSuchThat(),
//	TObjectList::ForEachDo() and TObjectList::FindFirstSuchThat().
//----------------------------------------------------------------------------------------

#define FOREACHOBJECT(listvar) FOREACH(TObjectLoop, listvar, TObject, object)


//========================================================================================
// CLASS TBackwardLoop
//
// Iterate a TList backwards
//========================================================================================

class TBackwardLoop
	{

	//
	// --- Methods -------------------------------------------------------------------------------------
	//
public:
				TBackwardLoop(TList*);
	void* 		Next();
	void* 		Current();

	//
	// --- Fields -------------------------------------------------------------------------------------
	//
private:
	TList*		fList;
	long		fByteIndex;
	};

//----------------------------------------------------------------------------------------
// TBackwardLoop::TBackwardLoop: 
//----------------------------------------------------------------------------------------
inline TBackwardLoop::TBackwardLoop(TList* list)
	{
	fByteIndex = list->ByteCount();			// start beyond last item
	fList = list;
	} // TBackwardLoop::TBackwardLoop 


//
// Prototypes
//
void Merge(TObject** list1, 
		   TObject** list2, 
		   TObject** list2End, 
		   TObject** output, 
		   void* parm);
void SortSection(Ptr input, 
				 Ptr inputEnd, 
				 long sortSize, 
				 Ptr output, 
				 void* parm);

#endif
