/*
	File:		ObjectCache.c

	Contains:	a cache to jam data into

	Written by:	Bruce Horn, Steve Capps, Larry Kenyon, John Meier, scott douglass, Darin Adler,
				Paul Mercer, Bryan Stearns, Dave Owens
				
				Andy Nicholas

	Copyright:	 1988-1995 by Apple Computer, Inc., all rights reserved.

		 <2>	 2/24/95	ga		
*/

#ifdef MWTRACEBACKTABLES
#pragma traceback on
#endif

#include "ObjectCache.h"
#include "Exceptions.h"

//
// For BlockMoveData
//
#include <Memory.h>

//
// For memFullErr
//
#include <Errors.h>

#include "Debug.h"

//
// WordAlign is from MoreTypes, but we don't want to use
// MoreTypes here.  This is a pretty simple inline, so I'm
// just going to duplicate it.
//
// Problem:		Precompiled header file gets in the way!
//				Solution:  Take simple stuff out of MoreTypes
//				and include it from this file and MoreTypes.
//				I'll do that later.
//
#if 1

void*					WordAlign(void*);
inline void*			WordAlign(void* p)	{ return (void*) (((long)p+1) & (~1));	}

#endif

//========================================================================================
// CLASS TObjectCache
//========================================================================================

//----------------------------------------------------------------------------------------
// TObjectCache::~TObjectCache: 
//
// This destructor gets called whenever a BlockCache or PropertyCache needs to be
// destroyed.
//----------------------------------------------------------------------------------------
TObjectCache::~TObjectCache()																		
	{
	TObjectCacheHeader* sourceBlock = (TObjectCacheHeader*) fCacheData;

	if (sourceBlock)
		{
		REQUIREVALIDPOINTER(sourceBlock);					// check for oddity, and 0xdeadbeef'ness

		//
		// We used operator new [] to allocate, so we'd better
		// user operator delete [] to deallocate.  In debug builds,
		// global operator delete will fill the block with 0xDEADBEEF.
		// On nodebug builds, the block is filled with 0 on V0 (bedrock
		// memory manager behavior) and left unchanged on V1.
		//
		delete [] sourceBlock;
		}
	}

//----------------------------------------------------------------------------------------
// TObjectCache::Allocate: 
//
// Returns a pointer to the cache data.  If it changed, then the cache data has
// been moved!
//----------------------------------------------------------------------------------------
Ptr	TObjectCache::Allocate(UInt32 cacheSize)
	{

	//
	// We should never be called to allocate zero bytes.
	//
	// The allocated size should always at least include the PropertyCache or BlockCache
	// header size.
	//
	ASSERTPRINT((cacheSize != 0), ("The object cache must be more than zero bytes!"));

	//
	// First, see if we have enough physical memory in this block to satisfy this
	// request.  Actually, see if we have -any- physical memory at all at this
	// point.
	//
	TObjectCacheHeader* sourceBlock = (TObjectCacheHeader*) fCacheData;

	if (sourceBlock)
		{
		if (sourceBlock->fPhysicalSize >= cacheSize)
			{
			sourceBlock->fUsed = cacheSize;
			return (Ptr) sourceBlock;
			}		
		}

	//
	// The cache grow size is currently a multiple of 32 
	// (ie, it rounds up to the nearest multiple of 32)
	//
	UInt32 newPhysicalSize = (cacheSize + kCacheGrowSize) & (~kCacheGrowSize);

	//
	// Try allocating our new block; use operator new [], which calls the
	// global operator new, which hopefully someone has written to use the
	// correct memory allocation strategy
	//
	void* ourBlock = (void*) new char[newPhysicalSize];															

	//
	// If whichever memory manager we are using returned nil for the result of the block
	//  allocation, that's real bad.  Fail and say we've run out of memory.  Hope someone
	//  catches this failure!
	//
	if (ourBlock == nil)
		FailErr(memFullErr);

	TObjectCacheHeader* newCacheData 	= (TObjectCacheHeader*) ourBlock;
	
	//
	// If a block already exists, try moving the "Used" amount of bytes from the old object
	// into the new object.
	//
	// "Used" is defined to be the size of the Block or Property cache header plus the
	// actual bytes used. 
	//
	if (sourceBlock)
		{
		UInt32 moveHowMuch = 0;
		
		//
		// If we were just asked to ratchet down the amount of space we're using,
		// make sure that we only copy the smaller number of bytes and don't overwrite
		// anything beyond the end of our block (gak).
		//
		if (sourceBlock->fUsed <= cacheSize)
			moveHowMuch = sourceBlock->fUsed;
		else
			moveHowMuch = cacheSize;

		BlockMoveData(sourceBlock, newCacheData, moveHowMuch);
		
		//
		// Free the source block
		//
		delete [] sourceBlock;
		}
	
	//
	// Update the physical size and used sizes of the block we just created
	//
#ifdef DEBUG
	newCacheData->fFirstCookie 	= 'andy';
#endif
	newCacheData->fPhysicalSize = newPhysicalSize;
	newCacheData->fUsed 		= cacheSize;

	//
	// Update the cache pointer which is (most likely) inside a TList, TCommand, or ???? object
	//	
	fCacheData = newCacheData;
	
	return (Ptr) newCacheData;
	} // TObjectCache::Allocate 

//========================================================================================
// CLASS TObjectBlockCache
//========================================================================================

//----------------------------------------------------------------------------------------
// TObjectBlockCache::Create: 
//
//  inline this
//----------------------------------------------------------------------------------------
void TObjectBlockCache::Create(UInt32 cacheSize)
	{
	TObjectBlockCacheHeader* result = (TObjectBlockCacheHeader*) Allocate(cacheSize + sizeof(TObjectBlockCacheHeader));	
	
	result->fLockCount = 0x0;		// nothing locked
	} // TObjectBlockCache::Create 

//----------------------------------------------------------------------------------------
// TObjectBlockCache::Lock: 
//
//  inline this
//----------------------------------------------------------------------------------------
void TObjectBlockCache::Lock() const
	{
	TObjectBlockCacheHeader* blockHeader = BlockCacheDataPtr();
	
	if (!blockHeader)
		return;
	
	blockHeader->fLockCount++;
	} // TObjectBlockCache::Lock 

//----------------------------------------------------------------------------------------
// TObjectBlockCache::Unlock: 
//
//  inline this
//----------------------------------------------------------------------------------------
void TObjectBlockCache::Unlock() const
	{
	TObjectBlockCacheHeader* blockHeader = BlockCacheDataPtr();
	
	if (!blockHeader)
		return;
	
	if (blockHeader->fLockCount)		// pin lock count at zero
		blockHeader->fLockCount--;
	} // TObjectBlockCache::Unlock 

//----------------------------------------------------------------------------------------
// TObjectBlockCache::Locked: 
//
//  inline this
//----------------------------------------------------------------------------------------
Boolean TObjectBlockCache::Locked() const
	{
	TObjectBlockCacheHeader* blockHeader = BlockCacheDataPtr();
	
	if (blockHeader && blockHeader->fLockCount)
		return true;
	
	return false;	
	} // TObjectBlockCache::Locked 

//----------------------------------------------------------------------------------------
// TObjectBlockCache::LockCount: 
//
//  inline this
//----------------------------------------------------------------------------------------
UInt16 TObjectBlockCache::LockCount() const
	{
	TObjectBlockCacheHeader* blockHeader = BlockCacheDataPtr();
	
	if (!blockHeader)
		return 0;

	return blockHeader->fLockCount;
	} // TObjectBlockCache::LockCount 

//----------------------------------------------------------------------------------------
// TObjectBlockCache::SetLockCount: 
//
//  inline this
//----------------------------------------------------------------------------------------
void TObjectBlockCache::SetLockCount(UInt16 count) const
	{
	TObjectBlockCacheHeader* blockHeader = BlockCacheDataPtr();
	
	if (!blockHeader)
		return;
	
	blockHeader->fLockCount = count;
	} // TObjectBlockCache::SetLockCount 

//----------------------------------------------------------------------------------------
// TObjectBlockCache::SetUsed: 
//
//  inline this
//----------------------------------------------------------------------------------------
void TObjectBlockCache::SetUsed(UInt32 cacheSize)
	{
	Ptr result = Allocate(cacheSize + sizeof(TObjectBlockCacheHeader));					
	} // TObjectBlockCache::SetUsed 

//----------------------------------------------------------------------------------------
// TObjectBlockCache::Start: 
//
//  inline this
//----------------------------------------------------------------------------------------
Ptr	TObjectBlockCache::Start() const
	{
	TObjectBlockCacheHeader* blockHeader = BlockCacheDataPtr();
	
	if (!blockHeader)
		return nil;
	else
		return ((Ptr) blockHeader) + sizeof(TObjectBlockCacheHeader);					
	} // TObjectBlockCache::Start 
	 
//----------------------------------------------------------------------------------------
// TObjectBlockCache::End: 
//
//  inline this
//----------------------------------------------------------------------------------------
Ptr	TObjectBlockCache::End() const
	{
	TObjectBlockCacheHeader* blockHeader = BlockCacheDataPtr();
	
	if (!blockHeader)
		return nil;
	else
		return ((Ptr) blockHeader) + blockHeader->fObjectCacheHeader.fUsed;					
	} // TObjectBlockCache::End 

//----------------------------------------------------------------------------------------
// TObjectBlockCache::Used: 
//
//  inline this
//----------------------------------------------------------------------------------------
UInt32 TObjectBlockCache::Used() const
	{
	TObjectBlockCacheHeader* blockHeader = BlockCacheDataPtr();
	
	if (!blockHeader)
		return 0;
	else
		return blockHeader->fObjectCacheHeader.fUsed - sizeof(TObjectBlockCacheHeader);					
	} // TObjectBlockCache::Used 
	
//----------------------------------------------------------------------------------------
// TObjectBlockCache::GetUInt16: 
//
//  inline this
//----------------------------------------------------------------------------------------
UInt16 TObjectBlockCache::GetUInt16(UInt32 offset) const				
	{
	TObjectBlockCacheHeader* blockHeader = BlockCacheDataPtr();
	
	if (!blockHeader)
		return 0;
	else
		return * (UInt16*) ( ((Ptr) blockHeader) + offset + sizeof(TObjectBlockCacheHeader));					
	} // TObjectBlockCache::GetUInt16 
	
//----------------------------------------------------------------------------------------
// TObjectBlockCache::GetUInt32: 
//
//  inline this
//----------------------------------------------------------------------------------------
UInt32 TObjectBlockCache::GetUInt32(UInt32 offset) const				
	{
	TObjectBlockCacheHeader* blockHeader = BlockCacheDataPtr();
	
	if (!blockHeader)
		return 0;
	else
		return * (UInt32*) ( ((Ptr) blockHeader) + offset + sizeof(TObjectBlockCacheHeader));	
	} // TObjectBlockCache::GetUInt32 
	
//----------------------------------------------------------------------------------------
// TObjectBlockCache::GetInt64: 
//
//  inline this
//----------------------------------------------------------------------------------------
Int64 TObjectBlockCache::GetInt64(UInt32 offset) const				
	{
	TObjectBlockCacheHeader* blockHeader = BlockCacheDataPtr();
	
	if (!blockHeader)
		return 0;
	else
		return * (Int64*)  ( ((Ptr) blockHeader) + offset + sizeof(TObjectBlockCacheHeader));	
	} // TObjectBlockCache::GetInt64 
	
//----------------------------------------------------------------------------------------
// TObjectBlockCache::GetPtr: 
//
//  inline this
//----------------------------------------------------------------------------------------
Ptr	TObjectBlockCache::GetPtr(UInt32 offset) const				
	{
	TObjectBlockCacheHeader* blockHeader = BlockCacheDataPtr();
	
	if (!blockHeader)
		return 0;
	else
		return (Ptr) ( ((Ptr) blockHeader) + offset + sizeof(TObjectBlockCacheHeader));	
	} // TObjectBlockCache::GetPtr 

//----------------------------------------------------------------------------------------
// TObjectBlockCache::Set:
//
// For UInt16's
//
//  inline this
//----------------------------------------------------------------------------------------
void TObjectBlockCache::Set(UInt16 value, UInt32 offset)		
	{
	TObjectBlockCacheHeader* blockHeader = BlockCacheDataPtr();
	
	//
	// If there wasn't a block before (ack!), or if the available space in the block doesn't
	// meet our needs, then call allocate and try to get some more.  TObjectCache::Allocate
	// will do chunky allocation first (ie, try to get more memory from the current block),
	// then it will allocate another block and move the contents from the old cache to the
	// new cache
	//
	if ( (!blockHeader) ||
	     ((blockHeader->fObjectCacheHeader.fUsed - sizeof(TObjectBlockCacheHeader) - sizeof(UInt16)) >= offset) )
		{
		blockHeader = (TObjectBlockCacheHeader*) Allocate( offset + sizeof(TObjectBlockCacheHeader) + sizeof(UInt16));
		}
		
 	* (UInt16*)  ( ((Ptr) blockHeader) + offset + sizeof(TObjectBlockCacheHeader)) = value;	
	} // TObjectBlockCache::Set 
	
//----------------------------------------------------------------------------------------
// TObjectBlockCache::Set:
//
// For UInt32's 
//
//  inline this
//----------------------------------------------------------------------------------------
void TObjectBlockCache::Set(UInt32 value, UInt32 offset)		
	{
	TObjectBlockCacheHeader* blockHeader = BlockCacheDataPtr();
	
	//
	// If there wasn't a block before (ack!), or if the available space in the block doesn't
	// meet our needs, then call allocate and try to get some more.  TObjectCache::Allocate
	// will do chunky allocation first (ie, try to get more memory from the current block),
	// then it will allocate another block and move the contents from the old cache to the
	// new cache
	//
	if ( (!blockHeader) ||
	     ((blockHeader->fObjectCacheHeader.fUsed - sizeof(TObjectBlockCacheHeader) - sizeof(UInt32)) >= offset) )
		{
		blockHeader = (TObjectBlockCacheHeader*) Allocate( offset + sizeof(TObjectBlockCacheHeader) + sizeof(UInt32));
		}

	* (UInt32*)  ( ((Ptr) blockHeader) + offset + sizeof(TObjectBlockCacheHeader)) = value;	
	} // TObjectBlockCache::Set 
	
//----------------------------------------------------------------------------------------
// TObjectBlockCache::Set:
//
// For Int64's 
//
//  inline this
//----------------------------------------------------------------------------------------
void TObjectBlockCache::Set(Int64 value,  UInt32 offset)		
	{
	TObjectBlockCacheHeader* blockHeader = BlockCacheDataPtr();
	
	//
	// If there wasn't a block before (ack!), or if the available space in the block doesn't
	// meet our needs, then call allocate and try to get some more.  TObjectCache::Allocate
	// will do chunky allocation first (ie, try to get more memory from the current block),
	// then it will allocate another block and move the contents from the old cache to the
	// new cache
	//
	if ( (!blockHeader) ||
	     ((blockHeader->fObjectCacheHeader.fUsed - sizeof(TObjectBlockCacheHeader) - sizeof(Int64)) >= offset) )
		{
		blockHeader = (TObjectBlockCacheHeader*) Allocate( offset + sizeof(TObjectBlockCacheHeader) + sizeof(Int64));
		}

	 * (Int64*) ( ((Ptr) blockHeader) + offset + sizeof(TObjectBlockCacheHeader)) = value;	
	} // TObjectBlockCache::Set 

//========================================================================================
// CLASS TObjectPropertyCache
//========================================================================================
	
//----------------------------------------------------------------------------------------
// TObjectPropertyCache::LookupProperty: 
//
// Look through the object cache to find the specified property
//
// NOTE: It would be useful to put some statistics gatherer into this routine which
//	     figures out on average how many properties we have to walk through before
//	     we find a match.  I believe that we don't usually have to walk very far,
//	     but it would be a useful piece of data to acquire. -andy
//----------------------------------------------------------------------------------------
TObjectPropertyCacheProperty* TObjectPropertyCache::LookupProperty(TProperty property) const
	{
	TObjectPropertyCacheProperty* prop 	= (TObjectPropertyCacheProperty*) PropertyCacheStart();
	TObjectPropertyCacheProperty* end	= (TObjectPropertyCacheProperty*) PropertyCacheEnd();

	while (prop < end)
		{
		if (prop->fProperty == property)
			{
			TObjectPropertyCacheHeader* propertyHeader = PropertyCacheDataPtr();
			
			propertyHeader->fLastProperty 	 = property;
			propertyHeader->fLastPropertyPtr = prop;
			 
			return prop;
			}
				
		prop = NextProperty(prop);
		}
		
	return nil;
	} // TObjectPropertyCache::LookupProperty 

//----------------------------------------------------------------------------------------
// TObjectPropertyCache::InvalAllWidths: 
//
// Make all the widths in this cache invalid
//----------------------------------------------------------------------------------------
void TObjectPropertyCache::InvalAllWidths() const
	{
	TObjectPropertyCacheProperty* prop 	= (TObjectPropertyCacheProperty*) PropertyCacheStart();
	TObjectPropertyCacheProperty* end	= (TObjectPropertyCacheProperty*) PropertyCacheEnd();
	
	if ( (prop == nil) || (end == nil) )
		return;

	while (prop < end)
		{
		prop->fWidth = kInvalidCacheWidth;
		prop = NextProperty(prop);
		}
		
	return;
	} // TObjectPropertyCache::InvalAllWidths 


//----------------------------------------------------------------------------------------
// TObjectPropertyCache::AddProperty:
//
// dataSize is allowed to be zero.  That kinda sucks, model-wise. 
//----------------------------------------------------------------------------------------
void TObjectPropertyCache::AddProperty(TProperty property, Ptr sourceDataPtr, UInt8 dataSize)
	{
	TObjectPropertyCacheHeader* propertyCacheHeader = PropertyCacheDataPtr();

	//
	// All properties are word-aligned 
	//  	(should they be long-aligned on PowerPC?)
	//	   	my thoughts on this are: probably yes since we should try really hard to keep
	//		everything long-aligned on PowerPC -andy	
	//
	UInt32 thisPropertySize = (UInt32) WordAlign( (void*) (dataSize + sizeof(TObjectPropertyCacheProperty)) );

	//
	// If there is no object cache, make one large enough to hold this
	// property and its data.
	//
	// Otherwise, a cache already existed, so allocate enough space to
	//  hold the data already being used in the cache and this property
	//  and its data.
	//
	if (!propertyCacheHeader)
		propertyCacheHeader = (TObjectPropertyCacheHeader*) Allocate(thisPropertySize + sizeof(TObjectPropertyCacheHeader) );
	else
		propertyCacheHeader = (TObjectPropertyCacheHeader*) Allocate(propertyCacheHeader->fObjectCacheHeader.fUsed + thisPropertySize);

	//
	// The pointer to the propertyCacheHeader should -never- be nil or you'll die soon afterwards.
	//
	ASSERT(propertyCacheHeader != nil);
		
	//
	// Point to the new property's spot
	//
	// At this point we have already claimed enough space within the object cache to
	// hold this property and its data, so we need to skip to the end of the used
	// data then backup the size of the property which we're adding (to the end)
	//
	TObjectPropertyCacheProperty* newProp = (TObjectPropertyCacheProperty*) 
		( ((Ptr) propertyCacheHeader) + propertyCacheHeader->fObjectCacheHeader.fUsed - thisPropertySize);

	//
	// Update the property fields within the cache
	//
	newProp->fProperty 		= property;
	newProp->fIsValid		= true;
	newProp->fWidth			= kInvalidCacheWidth;
	newProp->fStringSize	= dataSize;
	
	//
	// Move the data into place
	//
	Ptr destinationDataPtr 		= (Ptr) newProp + sizeof(TObjectPropertyCacheProperty);
	UInt32 wordAlignedDataSize 	= (UInt32) WordAlign( (void*) dataSize);
	
	switch(wordAlignedDataSize)
		{
		case 0:
			break;

		case 2:
			* (UInt16*) destinationDataPtr = * (UInt16*) sourceDataPtr;
			break;
			
		case 4:
			* (UInt32*) destinationDataPtr = * (UInt32*) sourceDataPtr;
			break;

		case 8:
			* (UInt32*)  destinationDataPtr 	 = * (UInt32*)  sourceDataPtr;
			* (UInt32*) (destinationDataPtr + 4) = * (UInt32*) (sourceDataPtr + 4);
			break;
		
		//
		// For the most part, calling BlockMoveData to copy this is dumb because
		// most of the strings we copy will be -really- short.
		//
		default:
			BlockMoveData(sourceDataPtr, destinationDataPtr, wordAlignedDataSize);
			break;
		}
	
	//
	// Update the cache's "Last Used" property cache for this cache
	//
	propertyCacheHeader->fLastProperty 		= property;
	propertyCacheHeader->fLastPropertyPtr 	= newProp;
	} // TObjectPropertyCache::AddProperty 

//----------------------------------------------------------------------------------------
// TObjectPropertyCache::RemoveProperty: 
//----------------------------------------------------------------------------------------
void TObjectPropertyCache::RemoveProperty(TProperty property)
	{
	TObjectPropertyCacheHeader* propertyCacheHeader = PropertyCacheDataPtr();

	if (propertyCacheHeader)
		{
		TObjectPropertyCacheProperty* prop = LookupPropertyFast(property);
	
		if (!prop)
			prop = LookupProperty(property);
	
		if (prop)
			{
			Ptr next	= (Ptr) NextProperty(prop);
			Ptr end 	=  PropertyCacheEnd();
	
			//
			// Shuffle all the other blocks UP in the cache over top of the
			// property we just killed... that is, assuming we actually -have-
			// some data to move upwards...
			//
			if (end - (Ptr) next)
				BlockMoveData(next, (Ptr) prop, end - (Ptr) next);	

			//
			// Update the size information (subtract out the size of the property we just killed)
			//	
			propertyCacheHeader->fObjectCacheHeader.fUsed -= ( next - ((Ptr) prop) );
			
			//
			// Flush the "last property" cache of the last property we looked up
			//
			propertyCacheHeader->fLastProperty 		= 0;
			propertyCacheHeader->fLastPropertyPtr 	= nil;
			}
		}
	} // TObjectPropertyCache::RemoveProperty 

//----------------------------------------------------------------------------------------
// TObjectPropertyCache::GetPropertyUInt16: 
//----------------------------------------------------------------------------------------
Boolean TObjectPropertyCache::GetPropertyUInt16(TProperty property, UInt16* dataPtr) const
	{
	Boolean result = false;
	TObjectPropertyCacheHeader* propertyCacheHeader = PropertyCacheDataPtr();

	if (propertyCacheHeader)
		{
		TObjectPropertyCacheProperty* prop = LookupPropertyFast(property);
	
		if (!prop)
			prop = LookupProperty(property);
	
		if (prop)
			{
			*dataPtr = * (UInt16*) ( ((Ptr) prop) + sizeof(TObjectPropertyCacheProperty));
			result = true;
			}
		}
		
	return result;
	} // TObjectPropertyCache::GetPropertyUInt16 


//----------------------------------------------------------------------------------------
// TObjectPropertyCache::GetPropertyUInt32: 
//----------------------------------------------------------------------------------------
Boolean TObjectPropertyCache::GetPropertyUInt32(TProperty property, UInt32* dataPtr) const
	{
	Boolean result = false;
	TObjectPropertyCacheHeader* propertyCacheHeader = PropertyCacheDataPtr();

	if (propertyCacheHeader)
		{
		TObjectPropertyCacheProperty* prop = LookupPropertyFast(property);
	
		if (!prop)
			prop = LookupProperty(property);
	
		if (prop)
			{
			*dataPtr = * (UInt32*) ( ((Ptr) prop) + sizeof(TObjectPropertyCacheProperty));
			result = true;
			}
		}
		
	return result;
	} // TObjectPropertyCache::GetPropertyUInt32 


//----------------------------------------------------------------------------------------
// TObjectPropertyCache::GetPropertyInt64: 
//----------------------------------------------------------------------------------------
Boolean TObjectPropertyCache::GetPropertyInt64(TProperty property, Int64* dataPtr) const
	{
	Boolean result = false;
	TObjectPropertyCacheHeader* propertyCacheHeader = PropertyCacheDataPtr();

	if (propertyCacheHeader)
		{
		TObjectPropertyCacheProperty* prop = LookupPropertyFast(property);
	
		if (!prop)
			prop = LookupProperty(property);
	
		if (prop)
			{
			*dataPtr = * (Int64*) ( ((Ptr) prop) + sizeof(TObjectPropertyCacheProperty));
			result = true;
			}
		}
		
	return result;
	} // TObjectPropertyCache::GetPropertyInt64 

//----------------------------------------------------------------------------------------
// TObjectPropertyCache::GetPropertyTStringPtr: 
//----------------------------------------------------------------------------------------
unsigned char* TObjectPropertyCache::GetPropertyTStringPtr(TProperty property) const
	{
	unsigned char* result = nil;
	TObjectPropertyCacheHeader* propertyCacheHeader = PropertyCacheDataPtr();

	if (propertyCacheHeader)
		{
		TObjectPropertyCacheProperty* prop = LookupPropertyFast(property);
	
		if (!prop)
			prop = LookupProperty(property);
	
		if (prop)
			{
			result = (unsigned char*) &prop->fStringSize;
			}
		}
		
	return result;
	} // TObjectPropertyCache::GetPropertyInt64 


//----------------------------------------------------------------------------------------
// TObjectPropertyCache::GetPropertyTString: 
//----------------------------------------------------------------------------------------
Boolean TObjectPropertyCache::GetPropertyTString(TProperty property, unsigned char* destinationForPString) const
	{
	Boolean result = false;
	TObjectPropertyCacheHeader* propertyCacheHeader = PropertyCacheDataPtr();

	if (propertyCacheHeader)
		{
		TObjectPropertyCacheProperty* prop = LookupPropertyFast(property);
	
		if (!prop)
			prop = LookupProperty(property);
	
		if (prop)
			{
			unsigned char* pstring = (unsigned char*) ( ((Ptr) prop) + sizeof(TObjectPropertyCacheProperty) - sizeof(UInt8));
			BlockMoveData((Ptr)pstring, (Ptr)destinationForPString, pstring[0] + 1);
			result = true;
			}
		}
		
	return result;
	} // TObjectPropertyCache::GetPropertyTString 


//----------------------------------------------------------------------------------------
// TObjectPropertyCache::GetPropertyTStringAndWidth: 
//----------------------------------------------------------------------------------------
Boolean TObjectPropertyCache::GetPropertyTStringAndWidth(TProperty property, unsigned char* destinationForPString, UInt16* widthPtr) const
	{
	Boolean result = false;
	TObjectPropertyCacheHeader* propertyCacheHeader = PropertyCacheDataPtr();

	if (propertyCacheHeader)
		{
		TObjectPropertyCacheProperty* prop = LookupPropertyFast(property);
	
		if (!prop)
			prop = LookupProperty(property);
	
		if (prop)
			{
			unsigned char* pstring = (unsigned char*) ( ((Ptr) prop) + sizeof(TObjectPropertyCacheProperty) - sizeof(UInt8));
			BlockMoveData((Ptr)pstring, (Ptr)destinationForPString, *pstring + 1);
			*widthPtr = prop->fWidth;
			result = true;
			}
		}
		
	return result;
	} // TObjectPropertyCache::GetPropertyTStringAndWidth 

//----------------------------------------------------------------------------------------
// TObjectPropertyCache::Exists: 
//----------------------------------------------------------------------------------------
Boolean TObjectPropertyCache::Exists(TProperty property) const
	{
	Boolean result = false;
	TObjectPropertyCacheHeader* propertyCacheHeader = PropertyCacheDataPtr();

	if (propertyCacheHeader)
		{
		TObjectPropertyCacheProperty* prop = LookupPropertyFast(property);
	
		if (!prop)
			prop = LookupProperty(property);
	
		if (prop)
			result = true;
		}
		
	return result;
	} // TObjectPropertyCache::Exists 


//----------------------------------------------------------------------------------------
// TObjectPropertyCache::IsValid(TProperty) const: 
//----------------------------------------------------------------------------------------
Boolean TObjectPropertyCache::IsValid(TProperty property) const
	{
	Boolean result = false;
	TObjectPropertyCacheHeader* propertyCacheHeader = PropertyCacheDataPtr();

	if (propertyCacheHeader)
		{
		TObjectPropertyCacheProperty* prop = LookupPropertyFast(property);
	
		if (!prop)
			prop = LookupProperty(property);
	
		if (prop)
			result = prop->fIsValid;
		}
	
	return result;
	} // TObjectPropertyCache::IsValid 


//----------------------------------------------------------------------------------------
// TObjectPropertyCache::Width:
//----------------------------------------------------------------------------------------
UInt16 TObjectPropertyCache::Width(TProperty property) const
	{
	UInt16 result = kInvalidCacheWidth;
	TObjectPropertyCacheHeader* propertyCacheHeader = PropertyCacheDataPtr();

	if (propertyCacheHeader)
		{
		TObjectPropertyCacheProperty* prop = LookupPropertyFast(property);
	
		if (!prop)
			prop = LookupProperty(property);
	
		if (prop)
			result = prop->fWidth;
		}
	
	return result;
	} // TObjectPropertyCache::Width 


//----------------------------------------------------------------------------------------
// TObjectPropertyCache::SetPropertyUInt16:
//
// This routine replaces the current property in the cache (if there is one),
// and creates a cache if there wasn't a cache before.  It also resizes the
// cache in order to make the property fit if it wouldn't fit before.
//----------------------------------------------------------------------------------------
void TObjectPropertyCache::SetPropertyUInt16(TProperty property, UInt16 value)
	{
	TObjectPropertyCacheHeader* propertyCacheHeader = PropertyCacheDataPtr();

	if (propertyCacheHeader)
		{
		TObjectPropertyCacheProperty* prop = LookupPropertyFast(property);
	
		if (!prop)
			prop = LookupProperty(property);
	
		if (prop)
			RemoveProperty(property);
		}

	AddProperty(property, (Ptr) &value, sizeof(UInt16));		
	} // TObjectPropertyCache::SetPropertyUInt16 


//----------------------------------------------------------------------------------------
// TObjectPropertyCache::SetPropertyUInt32:
//
// This routine replaces the current property in the cache (if there is one),
// and creates a cache if there wasn't a cache before.  It also resizes the
// cache in order to make the property fit if it wouldn't fit before.
//----------------------------------------------------------------------------------------
void TObjectPropertyCache::SetPropertyUInt32(TProperty property, UInt32 value)
	{
	TObjectPropertyCacheHeader* propertyCacheHeader = PropertyCacheDataPtr();

	if (propertyCacheHeader)
		{
		TObjectPropertyCacheProperty* prop = LookupPropertyFast(property);
	
		if (!prop)
			prop = LookupProperty(property);
	
		if (prop)
			RemoveProperty(property);
		}

	AddProperty(property, (Ptr) &value, sizeof(UInt32));		
	} // TObjectPropertyCache::SetPropertyUInt32 


//----------------------------------------------------------------------------------------
// TObjectPropertyCache::SetPropertyInt64:
//
// This routine replaces the current property in the cache (if there is one),
// and creates a cache if there wasn't a cache before.  It also resizes the
// cache in order to make the property fit if it wouldn't fit before.
//----------------------------------------------------------------------------------------
void TObjectPropertyCache::SetPropertyInt64(TProperty property, Int64 value)
	{
	TObjectPropertyCacheHeader* propertyCacheHeader = PropertyCacheDataPtr();

	if (propertyCacheHeader)
		{
		TObjectPropertyCacheProperty* prop = LookupPropertyFast(property);
	
		if (!prop)
			prop = LookupProperty(property);
	
		if (prop)
			RemoveProperty(property);
		}

	AddProperty(property, (Ptr) &value, sizeof(Int64));		
	} // TObjectPropertyCache::SetPropertyInt64 


//----------------------------------------------------------------------------------------
// TObjectPropertyCache::SetPropertyStruct:
//
// This routine replaces the current property in the cache (if there is one),
// and creates a cache if there wasn't a cache before.  It also resizes the
// cache in order to make the property fit if it wouldn't fit before.
//----------------------------------------------------------------------------------------
void TObjectPropertyCache::SetPropertyStruct(TProperty property, Ptr sourceForStruct, UInt8 structSize)
	{
	TObjectPropertyCacheHeader* propertyCacheHeader = PropertyCacheDataPtr();

	if (propertyCacheHeader)
		{
		TObjectPropertyCacheProperty* prop = LookupPropertyFast(property);
	
		if (!prop)
			prop = LookupProperty(property);
	
		if (prop)
			RemoveProperty(property);
		}

	AddProperty(property, sourceForStruct, structSize);		
	} // TObjectPropertyCache::SetPropertyStruct 

//----------------------------------------------------------------------------------------
// TObjectPropertyCache::SetPropertyTString:
//
// This routine replaces the current property in the cache (if there is one),
// and creates a cache if there wasn't a cache before.  It also resizes the
// cache in order to make the property fit if it wouldn't fit before.
//----------------------------------------------------------------------------------------
void TObjectPropertyCache::SetPropertyTString(TProperty property, unsigned char* str)
	{
	TObjectPropertyCacheHeader* propertyCacheHeader = PropertyCacheDataPtr();

	if (propertyCacheHeader)
		{
		TObjectPropertyCacheProperty* prop = LookupPropertyFast(property);
	
		if (!prop)
			prop = LookupProperty(property);
	
		if (prop)
			RemoveProperty(property);
		}

	AddProperty(property, (Ptr) str + 1, str[0]);		
	} // TObjectPropertyCache::SetPropertyUInt16 

//----------------------------------------------------------------------------------------
// TObjectPropertyCache::SetWidth: 
//----------------------------------------------------------------------------------------
void TObjectPropertyCache::SetWidth(TProperty property, UInt16 width) const
	{
	TObjectPropertyCacheHeader* propertyCacheHeader = PropertyCacheDataPtr();

	if (propertyCacheHeader)
		{
		TObjectPropertyCacheProperty* prop = LookupPropertyFast(property);
	
		if (!prop)
			prop = LookupProperty(property);
	
		if (prop)
			prop->fWidth = width;
		}
	} // TObjectPropertyCache::SetWidth 


//----------------------------------------------------------------------------------------
// TObjectPropertyCache::MakeInvalid: 
//----------------------------------------------------------------------------------------
void TObjectPropertyCache::MakeInvalid(TProperty property) const
	{
	TObjectPropertyCacheHeader* propertyCacheHeader = PropertyCacheDataPtr();

	if (propertyCacheHeader)
		{
		TObjectPropertyCacheProperty* prop = LookupPropertyFast(property);
	
		if (!prop)
			prop = LookupProperty(property);
	
		if (prop)
			prop->fIsValid = false;
		}
	} // TObjectPropertyCache::MakeInvalid 

//----------------------------------------------------------------------------------------
// TObjectPropertyCache::MakeValid: 
//----------------------------------------------------------------------------------------
void TObjectPropertyCache::MakeValid(TProperty property) const
	{
	TObjectPropertyCacheHeader* propertyCacheHeader = PropertyCacheDataPtr();

	if (propertyCacheHeader)
		{
		TObjectPropertyCacheProperty* prop = LookupPropertyFast(property);
	
		if (!prop)
			prop = LookupProperty(property);
	
		if (prop)
			prop->fIsValid = true;
		}
	} // TObjectPropertyCache::MakeInvalid 

//----------------------------------------------------------------------------------------
// TObjectPropertyCache::PropertyCacheStart: 
//
//  inline this
//----------------------------------------------------------------------------------------
Ptr	TObjectPropertyCache::PropertyCacheStart() const
	{
	TObjectPropertyCacheHeader* propertyCacheHeader = PropertyCacheDataPtr();
	
	if (!propertyCacheHeader)
		return nil;
	else
		return ((Ptr) propertyCacheHeader) + sizeof(TObjectPropertyCacheHeader);					
	} // TObjectBlockCache::Start 
	 
//----------------------------------------------------------------------------------------
// TObjectPropertyCache::PropertyCacheEnd: 
//
//  inline this
//----------------------------------------------------------------------------------------
Ptr	TObjectPropertyCache::PropertyCacheEnd() const
	{
	TObjectPropertyCacheHeader* propertyCacheHeader = PropertyCacheDataPtr();
	
	if (!propertyCacheHeader)
		return nil;
	else
		return ((Ptr) propertyCacheHeader) + propertyCacheHeader->fObjectCacheHeader.fUsed;					
	} // TObjectBlockCache::End 

//----------------------------------------------------------------------------------------
// TObjectPropertyCache::PropertyCacheUsed: 
//
//  inline this
//----------------------------------------------------------------------------------------
UInt32 TObjectPropertyCache::PropertyCacheUsed() const
	{
	TObjectPropertyCacheHeader* propertyCacheHeader = PropertyCacheDataPtr();
	
	if (!propertyCacheHeader)
		return 0;
	else
		return propertyCacheHeader->fObjectCacheHeader.fUsed - sizeof(TObjectPropertyCacheHeader);					
	} // TObjectBlockCache::Used 
	
//----------------------------------------------------------------------------------------
// TObjectPropertyCache::NextProperty:
//
// It's actually -quite- important that the compiler do a good job of inlining & optimizing
// what it inlines for this routine. -andy 
//
//  inline this
//----------------------------------------------------------------------------------------
TObjectPropertyCacheProperty* TObjectPropertyCache::NextProperty(TObjectPropertyCacheProperty* thisProperty) const
	{
	TObjectPropertyCacheProperty* nextProperty;
	
	nextProperty = (TObjectPropertyCacheProperty*) (((Ptr) thisProperty) + (UInt32) WordAlign((void*) (sizeof(TObjectPropertyCacheProperty) + thisProperty->fStringSize)));
	
	return nextProperty;
	}

//----------------------------------------------------------------------------------------
// TObjectPropertyCache::LookupPropertyFast: 
//
// Instead of looking through the object cache, see if the "last looked up object"
// matches the property we want now.  With a good optimizing compiler which expands
// multiple levels of inlines, this routine should be executed immediately before
// calling LookupProperty (if this one fails).  The code expansion hit should not
// be that large (again, with a good optimizing compiler). -andy
//
//  inline this
//----------------------------------------------------------------------------------------
TObjectPropertyCacheProperty* TObjectPropertyCache::LookupPropertyFast(TProperty property) const
	{
	TObjectPropertyCacheHeader* propertyCacheHeader = PropertyCacheDataPtr();

	if (propertyCacheHeader->fLastProperty == property)
		{
		TObjectPropertyCacheProperty* propertyPtr = propertyCacheHeader->fLastPropertyPtr;
		
		ASSERT(propertyPtr != nil);
		return propertyPtr;
		}
		
	return nil;
	} // TObjectPropertyCache::LookupPropertyFast 
