/*
	File:		ObjectCache.h

	Contains:	*** put contents here ***

	Written by:	Bruce Horn, Steve Capps, Larry Kenyon, John Meier, scott douglass, Darin Adler,
				Paul Mercer, Bryan Stearns, Dave Owens
				
				Andy Nicholas

	Copyright:	 1988-1995 by Apple Computer, Inc., all rights reserved.

				 2/23/95	ga		
*/

#ifndef ObjectCache_h
#define ObjectCache_h

#ifndef REZ

#include <Types.h>
#include <stddef.h>

//
// Just for define of DEBUG
//
#include "Debug.h"

//
// Bleh:  need the new types.h!
//
typedef unsigned char UInt8;
typedef unsigned short UInt16;
typedef unsigned long UInt32;

//
// Include Int64 because we support caching 64-bit values
//
#include "Int64.h"

//
// All cacheGrowSize values must be a power of 2 minus 1 because the pad value is
// also used as the mask! (ie, 15, 31, 65, 127, 255, etc.)
//
enum
	{
	kCacheGrowSize 				  = 	31
	};

//========================================================================================
// HANDLECACHE (andy)
//
// Or, how we get around the problem of having resizable data attached to all of
// our objects.
//
// The primary cache is the amount of space hanging off the end of our pointer-based
// object.  This cache is fixed-size and can't be grown.  But, if you know the size of
// your cache and know that it will only shrink and not grow, then the primary cache
// is for you!
//
// The secondary cache is a handle hanging off the initial pointer-block.  It is grown
// in 'handleCacheGrowSize' chunks. We pad each object by the amount in the
// handleCacheGrowSize (usually, 16 bytes).  The secondary cache is used to keep
// variable sized data which needs to be attached to an object (TLists, extra properties
// for an object).
//
// We use the Bedrock Memory Manager for our initial object allocation.  Objects and
// their caches are always cleared.  All secondary cache handles are currently dumped
// into the System's heap.  
//
// All handleCacheGrowSize values must be a power of 2 minus 1 because the pad value is
// also used as the mask! (ie, 15, 31, 65, 127, 255, etc.)
//
// If SetHandleCacheSize can't grow the cache, it FAILs.
//
// SetHandleCacheSize should never be able to be called if fHandleCacheGrowSize is
// nil, but we set fHandleCacheGrowSize to kHandleCachePadDefault anyway 
// (defensive programming, or just assume whatever can go wrong, will).
//
// Clients of the Handle Cache should know very little of of the implementation details
// of the Handle Cache -- ie, the physical size and logical size stuff shouldn't be
// available to the world.
//
// We need 3 versions of operator new because the ARM says on page 144 that default
// parameters cannot be passed to overloaded operators.  drat.
//
// ---------------------------------------------------------------------------------------
//
// Upon further reflection, since the size of properties does not change very often
// (it does, sometimes, ie when renaming or the modification date changes), it might be
// possible to have some sort of property management scheme where we relied on the
// speediness of the bedrock memory manager to allocate/de-allocate properties for us.
//
// You would have to be careful about the speed of the allocations because we tend to
// add several properties in a row to an object and then never add any more unless the
// use does something like Get Info on the object.  That's the reason we tried to
// chunkily allocate the size of the handle (because growing the handle is sometimes
// -incredibly- expensive).
//
// My guess at how this should be done is to allocate a second pointer block of a
// small size with the bedrock memory manager, then call "ReAllocate" on the block
// which allows the block to grow, but can change its pointer (which is OK, since
// we will just change the pointer in our object's pointer block.).
//
// This change could potentially save another 8 bytes per object allocated because
// we would not have this "split cache" concept.  Plus, it would allow the addition
// and removal of fields in the future since we wouldn't be using memory just past
// the end of the object.
//
// ---------------------------------------------------------------------------------------
//
// So this is what we did:
//
// Each object has a "cache" associated with it.  This "cache" is located via the
// "ObjectPropertyCache()" pointer.
//
// Pros: Overhead of exactly 4 bytes in the primary object.
//
// Cons: Must call Allocate twice for every object.  This is not necessarily bad because
//       Allocate must just not degrade its performance over time and we'll be happy.
//========================================================================================


//========================================================================================
// PROPERTIES
//
// Properties are normally 16-bit unsigned values.  32-bit properties are a superset
// of 16-bit properties.
//========================================================================================

typedef UInt16 				TProperty;
typedef UInt32				T32BitProperty;

//========================================================================================
// CLASS TObjectCache (andy)
//
// The ObjectCaches allocate a small or large amount of memory and allow it to be
// attached to 1 object. ie, an object cache is "per object."
//
// DO NOT ADD ANY VIRTUALS TO THIS CLASS.  
//
// We do -not- want a vtable pointer to exist in this object.
//
// DO NOT place any methods here which directly allocate space for a cache, rather,
// 		  funnel calls to Allocate from Create() methods in your type of cache which
//		  is derived from TObjectCache.
//========================================================================================

struct TObjectCacheHeader								// this struct is shared
	{													// by the object block cache header and
#ifdef DEBUG
	UInt32						fFirstCookie;
#endif
	UInt32						fPhysicalSize;
	UInt32						fUsed;					// by the object property cache header
	};

class TObjectCache
	{
public:
								TObjectCache() : fCacheData(nil) { }
								~TObjectCache();
	Boolean						Exists()					{ return fCacheData != nil;		}
									
protected:
	Ptr							Allocate(UInt32 cacheSize);

//----------------------------------------------------------------------------------------
// DO NOT allocate any -other- fields for this object because this class is embedded in
// several client classes, and we would like to not cause them to bloat unnecessarily.
//----------------------------------------------------------------------------------------
protected:
	TObjectCacheHeader*			fCacheData;
	};


//========================================================================================
// CLASS TObjectBlockCache (andy)
//
// The object block cache is a "per object" cache to keep large (> 64K) amounts of data
// related to the object.
//
// The client of the object block cache is assumed to want to iterate over the data.
// This data is *NOT* guaranteed to not move!  If you want it to not move, call the
// Lock/Unlock methods.
//
// DO NOT ADD ANY VIRTUALS TO THIS CLASS.  
//
// We do -not- want a vtable pointer to exist in this object.
//
// The Used() method returns the number of bytes actually used inside the block,
// but excludes the size of the ObjectCache Block Header from that value.
//
// IDEA: We could stuff the current tick count into the block and if they grow the
//		 block too often inside of 1 tick, we could start allocating the block in
//		 larger chunks?
//========================================================================================

struct TObjectBlockCacheHeader
	{
	TObjectCacheHeader			fObjectCacheHeader;
	UInt32						fLockCount;						// fyi, this doesn't need to be a UInt32,
																// it's only a UInt32 so that everything in a TList
																// stays 32-bit aligned
	};

class TObjectBlockCache : public TObjectCache
	{
public:
	void						Create(UInt32 cacheSize);
	void						SetUsed(UInt32 used);

	Ptr							Start() const;
	Ptr							End() const;
	UInt32						Used() const;

	void						Lock() const;
	void						Unlock() const;
	Boolean						Locked() const;
	
	UInt16						LockCount() const;
	void						SetLockCount(UInt16) const;

	UInt16						GetUInt16(UInt32 offset) const;
	UInt32						GetUInt32(UInt32 offset) const;
	Int64						GetInt64(UInt32 offset) const;
	Ptr							GetPtr(UInt32 offset) const;

	void						Set(UInt16 value, UInt32 offset);
	void						Set(UInt32 value, UInt32 offset);
	void						Set(Int64 value,  UInt32 offset);

private:	
	TObjectBlockCacheHeader*	BlockCacheDataPtr() const		{ return (TObjectBlockCacheHeader*) fCacheData; }
	};

//
// Formerly, we cached string widths in an unsigned char (UInt8);
// Now they go into a UInt16, so this number could be quite a bit
// longer.  I'm going to leave it at 255, tho, since there's still
// some kind cache code floating around, and it still uses a UInt8
// for the cached width (but is it ever called?).
//
enum
	{
	kInvalidCacheWidth			= 255
	};
	

//========================================================================================
// CLASS TObjectPropertyCache (andy)
//
// The object property cache is a "per object" cache to keep small (less than 64K) amounts
// of data related to the object.  The "64k" limit is imposed only because we don't want
// to use 4 extra bytes per object to make the size and used fields UInt32's.
//
// Accessor methods are provided to get and set "properties" within the object property
// cache.
//
// The API provides for copying shorts, longs, double longs (Int64's), structs, and strings
// into and out of the cache.  Is anything else needed?  Floats?
//
// The implementation details of this cache should remain hidden so that we can change
// it in the future should the need arise.  For now, it just searches for properties
// linearly.  If this turns out to be a problem, we could add a hint for the most
// often accessed property.  This class should not need Lock/Unlock methods because if
// we eventually put the data into handles (which would be stupid, IMHO), -no one-
// should be iterating over this data anyway.
//
// DO NOT ADD ANY VIRTUALS TO THIS CLASS.  
//
// We do -not- want a vtable pointer to exist in this object.
//========================================================================================

struct TObjectPropertyCacheProperty
	{
	UInt32 					fProperty;
	UInt16 					fWidth;
	UInt8					fIsValid		: 1;
	UInt8 					fStringSize;
	};

struct TObjectPropertyCacheHeader
	{
	TObjectCacheHeader				fObjectCacheHeader;				// includes used and physical size
	UInt32							fLastProperty;
	TObjectPropertyCacheProperty*	fLastPropertyPtr;
	};

class TObjectPropertyCache : public TObjectCache
	{

	// --- Methods -----------------------------------------------------------------------
public:
	Boolean							GetPropertyUInt16(TProperty, UInt16*) const;
	Boolean							GetPropertyUInt32(TProperty, UInt32*) const;
	Boolean							GetPropertyInt64( TProperty, Int64*) const;
	Boolean							GetPropertyStruct(TProperty, Ptr destinationForStruct) const;

	unsigned char*					GetPropertyTStringPtr(TProperty) const;
	Boolean							GetPropertyTString(TProperty, unsigned char* destinationForPString) const;
	Boolean							GetPropertyTStringAndWidth(TProperty, unsigned char* destinationForPString, UInt16* widthPtr) const;

	Boolean							Exists(TProperty) const;
	Boolean							IsValid(TProperty) const;
	UInt16							Width(TProperty) const;
		
	void							SetPropertyUInt16(TProperty, UInt16);
	void							SetPropertyUInt32(TProperty, UInt32);
	void							SetPropertyInt64( TProperty, Int64);
	void							SetPropertyStruct(TProperty, Ptr sourceForStruct, UInt8 structSize);

	void							SetPropertyTString(TProperty, unsigned char* sourceForTString);

	void							RemoveProperty(TProperty);

	void							MakeInvalid(TProperty) const;
	void							MakeValid(TProperty) const;
	void							SetWidth(TProperty, UInt16 width) const;
	void							InvalAllWidths() const;						
	
	//
	// These routines are protected because we do -not- want clients of the Property Cache
	// thinking of it in terms that are different from "it's just a box and I put properties
	// into it, and get them out of it." (ie, we don't want to publish any possible way to walk
	// the contents of the property cache other than through this code).
	//
protected:

	TObjectPropertyCacheHeader*		PropertyCacheDataPtr() const	{ return (TObjectPropertyCacheHeader*) fCacheData; }

	Ptr								PropertyCacheStart() const;
	Ptr								PropertyCacheEnd() const;
	UInt32							PropertyCacheUsed() const;

	TObjectPropertyCacheProperty*	LookupProperty(TProperty property) const;						// not inlined
	TObjectPropertyCacheProperty*	LookupPropertyFast(TProperty property) const;					// inlined for necessary speed
	TObjectPropertyCacheProperty* 	NextProperty(TObjectPropertyCacheProperty* thisProperty) const;	// inlined for necessary speed

	void						  	AddProperty(TProperty, Ptr, UInt8);
	};


#endif // REZ
#endif // Object_h

