/*
	File:		MoreAEM.c

	Contains:	C++ wrappers for the AppleEvent Manager

	Written by:	Francis Stanbach, Greg Anderson

	Copyright:	 1992, 1994-1995 by Apple Computer, Inc., all rights reserved.

		<31>	 2/15/95	ga		

*/

#ifdef MWTRACEBACKTABLES
#pragma traceback on
#endif

#include "MoreAEM.h"
#include "Int64.h"

#include <string.h>

#include <AppleEvents.h>
#include <AERegistry.h>
#include <ASRegistry.h>
#include <AEObjects.h>
#include <AEPackObject.h>
#include <Errors.h>
#include <OSUtils.h>
#include <Aliases.h>
#include <TextUtils.h>
#include <Script.h>
#include <GestaltEqu.h>

// #include "FinderRegistry.h"

#include "Exceptions.h"
#include "Debug.h"



#pragma segment AEMStuff

 
MakeTokenDescriptorProcPtr gNullContainerCreationProc = nil;
ProcessDescriptorProcPtr gPreResolveProc = nil;
MergeTokensProcPtr gMergeTokensProc = nil;

short TDescriptor::fCallbackFlags = kAEIDoMinimum;

#define kDayLengthInSecs	(24*60*60)


//================================================================================
// Class TAbstractCompareBehavior
//================================================================================

class TStringCompareBehavior : public TAbstractCompareBehavior
{
	virtual CompareEnumeration CompareDynamicBehavior(const TConstDataReference& object1, const TConstDataReference& object2) const;
	virtual Boolean ContainsDynamicBehavior(const TConstDataReference& object1, const TConstDataReference& object2) const;

	static Boolean IsTextType(DescType theType);
};

//----------------------------------------------------------------------------------------
// TStringCompareBehavior::IsTextType
//----------------------------------------------------------------------------------------
Boolean TStringCompareBehavior::IsTextType(DescType theType)
	{
	return (theType == typeAEText) || (theType == typeIntlText) || (theType == typeStyledText) || (theType == typeChar);
	}

//----------------------------------------------------------------------------------------
// TStringCompareBehavior::CompareDynamicBehavior
//----------------------------------------------------------------------------------------
CompareEnumeration TStringCompareBehavior::CompareDynamicBehavior(const TConstDataReference& object1, const TConstDataReference& object2) const
	{
	CompareEnumeration compareResult = kObjectKeysEqual;
	
	//
	// We can compare any sort of textual objects; anything that isn't
	// text will have to be handled by some other comparitor.
	//
	if(IsTextType(object1.DataType()) && IsTextType(object2.DataType()))
		{
		TConstDataReference object1Data = object1;
		TConstDataReference object2Data = object2;
		TDescriptor tempObject1Desc;
		TDescriptor tempObject2Desc;
		
		//
		// First, coerce the data to be a simple typeChar
		// (International note:  we loose information such
		// as the script code when we do this.  This routine
		// is NOT a good example of how to compare text).
		//
		if(object1.DataType() != typeChar)
			{
			FailErr(AECoercePtr(object1.DataType(), object1.Data(), object1.DataLength(), typeChar, tempObject1Desc));
			tempObject1Desc.Lock();
			object1Data = tempObject1Desc;
			}
		if(object2.DataType() != typeChar)
			{
			FailErr(AECoercePtr(object2.DataType(), object2.Data(), object2.DataLength(), typeChar, tempObject2Desc));
			tempObject2Desc.Lock();
			object2Data = tempObject2Desc;
			}
		
		//
		// Now that we have two strings of typeChar, compare them
		//
		long compareLength = object1Data.DataLength() < object2Data.DataLength() ? object1Data.DataLength() :  object2Data.DataLength();
		const char* data = object1Data.Data();
		const char* compareWith = object2Data.Data();
		
		for(long i=0; i < compareLength; ++i)
			{
			unsigned char compareWithChar = *((unsigned char*)compareWith);
			unsigned char dataChar = *((unsigned char*)data);
			
			if((compareWithChar >= 'a') && (compareWithChar <= 'z'))
				compareWithChar = compareWithChar - 'a' + 'A';
			if((dataChar >= 'a') && (dataChar <= 'z'))
				dataChar = dataChar - 'a' + 'A';
				
			if(compareWithChar > dataChar)
				{
				compareResult = kSecondObjectComesAfter;
				break;
				}
			else if(compareWithChar < dataChar)
				{
				compareResult = kSecondObjectComesBefore;
				break;
				}
			++compareWith;
			++data;
			}
		
		//
		// If the first 'compareLength' bytes of the data stream are equal, then
		// equality is based on the lengths of the strings
		//
		if(compareResult == kObjectKeysEqual)
			{
			if(object2Data.DataLength() > object1Data.DataLength())
				compareResult = kSecondObjectComesAfter;
			if(object2Data.DataLength() < object1Data.DataLength())
				compareResult = kSecondObjectComesBefore;
			}
			
		//
		// Delete our temporary data, if we allocated it.
		//
		tempObject1Desc.Dispose();
		tempObject2Desc.Dispose();
		}
	else
		{
		compareResult = TAbstractCompareBehavior::CompareDynamicBehavior(object1, object2);
		}
	
	return compareResult;
	}

//----------------------------------------------------------------------------------------
// TStringCompareBehavior::ContainsDynamicBehavior
//----------------------------------------------------------------------------------------
Boolean TStringCompareBehavior::ContainsDynamicBehavior(const TConstDataReference& object1, const TConstDataReference& object2) const
	{
	Boolean containsResult = false;
	
	//
	// We can compare any sort of textual objects; anything that isn't
	// text will have to be handled by some other comparitor.
	//
	if(IsTextType(object1.DataType()) && IsTextType(object2.DataType()))
		{
		TConstDataReference object1Data = object1;
		TConstDataReference object2Data = object2;
		TDescriptor tempObject1Desc;
		TDescriptor tempObject2Desc;
		
		//
		// First, coerce the data to be a simple typeChar
		// (International note:  we loose information such
		// as the script code when we do this.  This routine
		// is NOT a good example of how to compare text).
		//
		if(object1.DataType() != typeChar)
			{
			FailErr(AECoercePtr(object1.DataType(), object1.Data(), object1.DataLength(), typeChar, tempObject1Desc));
			object1Data = tempObject1Desc;
			}
		if(object2.DataType() != typeChar)
			{
			FailErr(AECoercePtr(object2.DataType(), object2.Data(), object2.DataLength(), typeChar, tempObject2Desc));
			object2Data = tempObject1Desc;
			}
		
		//
		// Look for object2 inside of object1
		//
		long lengthDifference = object1.DataLength() - object2.DataLength();
		const char* data = object1.Data();
		
		while(lengthDifference >= 0)
			{
			TConstDataReference partialString(object1.DataType(), data, object2.DataLength());
			
			if(partialString.Compare(object2) == kObjectKeysEqual)
				{
				containsResult = true;
				break;
				}
			
			--lengthDifference;
			++data;
			}
					
		//
		// Delete our temporary data, if we allocated it.
		//
		tempObject1Desc.Dispose();
		tempObject2Desc.Dispose();
		}
	else
		{
		containsResult = TAbstractCompareBehavior::ContainsDynamicBehavior(object1, object2);
		}
	
	return containsResult;
	}
	
//----------------------------------------------------------------------------------------
// InstallStringCompareBehavior
//
// Add our cheezy roman-only string compare behavior
//----------------------------------------------------------------------------------------
void InstallStringCompareBehavior()
	{
	TStringCompareBehavior* compareBehavior = new TStringCompareBehavior;
	
	TConstDataReference::AddCompareBehavoir(compareBehavior);
	}
	
//----------------------------------------------------------------------------------------
// InstallNullContainerCreationProc: 
//
// The null container creation proc is used by TDescriptor::Resolve whenever it receives
// a null descriptor to resolve.  This function is used to create a token for the null
// container.
//----------------------------------------------------------------------------------------
void InstallNullContainerCreationProc(MakeTokenDescriptorProcPtr creationProc)
	{
	gNullContainerCreationProc = creationProc;
	} // InstallNullContainerCreationProc 

//----------------------------------------------------------------------------------------
// InstallPreResolveProc: 
//
// The PreResolveProc is called by TDescriptor::Resolve before it actually calls
// AEResolve.  This allows special procs to handle things not understood by AEResolve
// (alias records, for example).
//----------------------------------------------------------------------------------------
void InstallPreResolveProc(ProcessDescriptorProcPtr preResolveProc)
	{
	gPreResolveProc = preResolveProc;
	} // InstallPreResolveProc 

//----------------------------------------------------------------------------------------
// InstallMergeTokensProc: 
//
// The MergeTokensProc is called by TTokenDescriptor::AdoptToken whenever it needs to add
// a token object to any token descriptor that already contains a token.  The merge token
// proc is provided with two TAbstractScriptableObject*'s; it must return a
// TAbstractScriptableObject* that has adopted the two tokens passed to it.
//----------------------------------------------------------------------------------------
void InstallMergeTokensProc(MergeTokensProcPtr mergeTokensProc)
	{
	gMergeTokensProc = mergeTokensProc;
	} // InstallMergeTokensProc 

//----------------------------------------------------------------------------------------
// CreateNullContainerToken: 
//----------------------------------------------------------------------------------------
TTokenDescriptor CreateNullContainerToken()
	{
	TTokenDescriptor nullContainer;
	
	if(gNullContainerCreationProc != nil)
		nullContainer = (*gNullContainerCreationProc)();
	
	return nullContainer;
	} // CreateNullContainerToken 

//========================================================================================
// CLASS TDescriptor
//
// Class TDescriptor is a wrapper class for AEDesc objects.
//========================================================================================

#ifdef NONO

//----------------------------------------------------------------------------------------
// TDescriptor::TDescriptor: 
//----------------------------------------------------------------------------------------
TDescriptor::TDescriptor()
	{
	fDescriptorType = typeNull;
	fDataHandle = nil;
	} // TDescriptor::TDescriptor 

#endif


//----------------------------------------------------------------------------------------
// TDescriptor::Dispose:
//
// Disposed of the descriptor and the data stored inside it.
//----------------------------------------------------------------------------------------
void TDescriptor::Dispose()
	{
	FailErr(AEDisposeDesc(*this));
	} // TDescriptor::Dispose 

//----------------------------------------------------------------------------------------
// TDescriptor::Lock
//----------------------------------------------------------------------------------------
short TDescriptor::Lock() const
	{
	short result = 0;
	
	if(fDataHandle != nil)
		{
		result = HGetState(fDataHandle);
		HLock(fDataHandle);
		}
	
	return result;
	}

//----------------------------------------------------------------------------------------
// TDescriptor::Unlock
//----------------------------------------------------------------------------------------
void TDescriptor::Unlock(short oldState) const
	{
	if(fDataHandle)
		HSetState(fDataHandle, oldState);
	}

//----------------------------------------------------------------------------------------
// TDescriptor::AttemptCoerceInPlace:
//
// Try to coerce the type of the descriptor to something else.  This method will
// NOT fail if the coercion could not be done.
//----------------------------------------------------------------------------------------
OSErr TDescriptor::AttemptCoerceInPlace(DescType typeToCoerceTo, CoerceInPlaceTechnique preserveHandle /*= kPreserveHandle*/)
	{
	OSErr err = noErr;
	
	if(this->DescriptorType() != typeToCoerceTo)
		{
		//
		// This way works, but the data handle of this descriptor
		// is disposed an replaced with a new copy.  This isn't
		// so hot if you have a copy of an AEDesc that you'd like
		// to coerce that someone else is going to dispose later.
		//
		// In short, preserveHandle == kAllowHandleReallocation is
		// faster, but more dangerous.
		//
		if(preserveHandle == kAllowHandleReallocation)
			err = AECoerceDesc(*this,typeToCoerceTo, *this);
		else
			{
			TDescriptor tempDescriptor;
			
			// If AECoerceDesc returns a non-zero result code, then
			// tempDescriptor will be typeNull, and we don't need to
			// worry about disposing of it.
			err = AECoerceDesc(*this,typeToCoerceTo, tempDescriptor);
	
			if(err == noErr)
				{			
				Try
					{
					Handle initialHandle = this->DataHandle();
					Handle coercedHandle = tempDescriptor.DataHandle();
					
					//
					// If doing a coerce-in-place, try to do a SetHandleSize
					// and then a BlockMove so that we end up using the same
					// handle when we exit as we had when we entered.  To do
					// otherwise is to invite double-dispose/memory leak problems
					// if some other peice of code still has a reference to
					// the descriptor before it was coerced
					//
					if((initialHandle != nil) && (coercedHandle != nil))
						{
						Size coercedSize = GetHandleSize(coercedHandle);
						SetHandleSize(initialHandle, coercedSize);
						FailErr(MemError());
						BlockMoveData(*coercedHandle, *initialHandle, coercedSize);
						this->fDescriptorType = tempDescriptor.DescriptorType();
						tempDescriptor.Dispose();
						}
					//
					// There are a couple of rare cases (typeTrue and typeFalse)
					// where the data handle is nil
					//
					else
						{
						//
						// It's undesirable to go from a nil-handle to a non-nil
						// handle, or visa-versa.  Don't do it.  Use 'Coerce'
						// rather than 'CoerceInPlace' if this is necessary.
						//
						if((initialHandle != nil) || (coercedHandle != nil))
							ASSERTPRINT(false, ("Potentially undesirable coerce-in-place"));
						//
						// Try to dispose of the old descriptor (ignoring errors)
						// and then plug in the values from the coerced descriptor
						//
						AEDisposeDesc(*this);
						this->AdoptDesc(tempDescriptor);
						}
					}
				Catch(err)
					{
					tempDescriptor.Dispose();
					}
				}
			}
		}
	
	return err;
	} // TDescriptor::AttemptCoerceInPlace 

//----------------------------------------------------------------------------------------
// TDescriptor::CoerceInPlace:
//
// Require that this object be coerced to the specified data type.  This method will
// fail if the coercion does not work.
//----------------------------------------------------------------------------------------
void TDescriptor::CoerceInPlace(DescType typeToCoerceTo, CoerceInPlaceTechnique preserveHandle /* = kPreserveHandle*/)
	{
	FailErr(this->AttemptCoerceInPlace(typeToCoerceTo, preserveHandle));
	} // TDescriptor::CoerceInPlace 

//----------------------------------------------------------------------------------------
// TDescriptor::AttemptToCoerce:
//----------------------------------------------------------------------------------------
TDescriptor TDescriptor::AttemptToCoerce(DescType typeToCoerceTo, OSErr& err) const
	{
	TDescriptor coercedDescriptor;
	
	err = AECoerceDesc(*this,typeToCoerceTo,coercedDescriptor);
	
	return coercedDescriptor;
	} // TDescriptor::AttemptToCoerce 

//----------------------------------------------------------------------------------------
// TDescriptor::Coerce:
//
// Require that this object be coerced to the specified data type.  This method will
// fail if the coercion does not work.
//----------------------------------------------------------------------------------------
TDescriptor TDescriptor::Coerce(DescType typeToCoerceTo) const
	{
	TDescriptor coercedDescriptor;
	
	FailErr(AECoerceDesc(*this,typeToCoerceTo,coercedDescriptor));
	
	return coercedDescriptor;
	} // TDescriptor::Coerce 
	
//----------------------------------------------------------------------------------------
// TDescriptor::CoerceToStandardType:
//
// This method attempts to coerce its data into some form of standard type
// (e.g. typeChar or typeLongInteger).  It is really annoying that someone can
// invent yet another data type that is coercable to text or as a long integer,
// but we wouldn't have any way to know that such a coercion should be attempted.
// typeMagnitude and typeEnumeration could throw us for a loop too.
//
// In short, I really hate this routine, but I'm not sure what else to do about it.
//----------------------------------------------------------------------------------------
TDescriptor TDescriptor::CoerceToStandardType()
	{
	TDescriptor result;
	
	switch(this->DescriptorType())
		{
		case typeAEText:
		case typeIntlText:
		case typeStyledText:
			{
			result = this->Coerce(typeChar);
			break;
			}
		
		case typeShortInteger:
			{
			result = this->Coerce(typeLongInteger);
			break;
			}
		}
	
	return result;
	} // TDescriptor::CoerceToStandardType 
	
//----------------------------------------------------------------------------------------
// TDescriptor::Clone:
//
// Return an exact copy of this descriptor
//----------------------------------------------------------------------------------------
TDescriptor TDescriptor::Clone() const
	{
	TDescriptor clonedDescriptor;
	
	FailErr(AEDuplicateDesc(*this, clonedDescriptor));
	
	return clonedDescriptor;
	} // TDescriptor::Clone 

//----------------------------------------------------------------------------------------
// TDescriptor::CopyDesc: 
//----------------------------------------------------------------------------------------
void TDescriptor::CopyDesc(const TDescriptor& desc)
	{
	FailErr(AEDuplicateDesc(desc, *this));
	} // TDescriptor::CopyDesc 

//----------------------------------------------------------------------------------------
// TDescriptor::AdoptDesc: 
//----------------------------------------------------------------------------------------
void TDescriptor::AdoptDesc(TDescriptor& desc)
	{
	//
	// It would be very tempting to set 'desc' to the null descriptor now
	// that we have adopted it here.  That would be a bad idea, though, because
	// we might intend this descriptor to actually be just a reference to
	// the adopted desc, and not actually the new 'owner' of the data.
	//
	this->AdoptHandle(desc.DescriptorType(), desc.DataHandle());
	} // TDescriptor::AdoptDesc 

//----------------------------------------------------------------------------------------
// TDescriptor::CreateList:
//
// Create an empty descriptor list or AERecord with factored data.
//
// n.b.	It is better to call MakeList, MakeEmptyList or MakeAERecord rather than
// using TDescriptor::CreateList directly.
//----------------------------------------------------------------------------------------
void TDescriptor::CreateList(Boolean isRecord, Ptr factoringPtr, Size factoredSize)
	{
	FailErr(AECreateList(factoringPtr, factoredSize, isRecord, *this));
	} // TDescriptor::CreateList 

//----------------------------------------------------------------------------------------
// TDescriptor::MakeNull: 
//----------------------------------------------------------------------------------------
void TDescriptor::MakeNull()
	{
	this->AdoptHandle(typeNull, nil);
	} // TDescriptor::MakeNull 

//----------------------------------------------------------------------------------------
// TDescriptor::AdoptHandle:
//
// Assign some data and a type directly to the descriptor
//----------------------------------------------------------------------------------------
void TDescriptor::AdoptHandle(DescType dataType, Handle dataHandle)
	{
	fDescriptorType = dataType;
	fDataHandle = dataHandle;
	} // TDescriptor::AdoptHandle

//----------------------------------------------------------------------------------------
// TDescriptor::CopyData: 
//----------------------------------------------------------------------------------------
void TDescriptor::CopyData(const DescType typeCode, const Ptr data, const Size length)
	{
	FailErr(AECreateDesc(typeCode, data, length, *this));
	} // TDescriptor::CopyData 

//----------------------------------------------------------------------------------------
// TDescriptor::Make
//----------------------------------------------------------------------------------------
void TDescriptor::Make(const TConstDataReference& data)
	{
	this->CopyData(data.DataType(), (const Ptr)data.Data(), data.DataLength());
	}
	
//----------------------------------------------------------------------------------------
// TDescriptor::MakeBoolean: 
//----------------------------------------------------------------------------------------
void TDescriptor::MakeBoolean(const Boolean data)
	{
	this->CopyData(typeBoolean, (Ptr)&data, sizeof(Boolean));
	} // TDescriptor::MakeBoolean 
	
//----------------------------------------------------------------------------------------
// TDescriptor::MakeShort: 
//----------------------------------------------------------------------------------------
void TDescriptor::MakeShort(const short data)
	{
	this->CopyData(typeShortInteger, (Ptr)&data, sizeof(short));
	} // TDescriptor::MakeShort

//----------------------------------------------------------------------------------------
// TDescriptor::MakeLong: 
//----------------------------------------------------------------------------------------
void TDescriptor::MakeLong(const long data)
	{
	this->CopyData(typeLongInteger, (Ptr)&data, sizeof(long));
	} // TDescriptor::MakeLong
	
//----------------------------------------------------------------------------------------
// TDescriptor::MakeInt64: 
//----------------------------------------------------------------------------------------
void TDescriptor::MakeInt64(Int64& data)
	{
	this->CopyData(typeInt64, (Ptr)&data, sizeof(long) + sizeof(long));
	} // TDescriptor::MakeInt64

//----------------------------------------------------------------------------------------
// TDescriptor::MakeUnsignedLong: 
//----------------------------------------------------------------------------------------
void TDescriptor::MakeUnsignedLong(const unsigned long data)
	{
	this->CopyData(typeMagnitude, (Ptr)&data, sizeof(unsigned long));
	} // TDescriptor::MakeUnsignedLong

//----------------------------------------------------------------------------------------
// TDescriptor::MakeEnumeration: 
//----------------------------------------------------------------------------------------
void TDescriptor::MakeEnumeration(const DescType enumeration)
	{
	this->CopyData(typeEnumeration, (Ptr)&enumeration, sizeof(enumeration));
	} // TDescriptor::MakeEnumeration 

//----------------------------------------------------------------------------------------
// TDescriptor::MakeDescType: 
//----------------------------------------------------------------------------------------
void TDescriptor::MakeDescType(const DescType data)
	{
	this->CopyData(typeType, (Ptr)&data, sizeof(DescType));
	} // TDescriptor::MakeDescType 

//----------------------------------------------------------------------------------------
// TDescriptor::MakeKeyword: 
//----------------------------------------------------------------------------------------
void TDescriptor::MakeKeyword(const AEKeyword theKeyword)
	{
	this->CopyData(typeKeyword, (Ptr)&theKeyword, sizeof(AEKeyword));
	} // TDescriptor::MakeKeyword 

//----------------------------------------------------------------------------------------
// TDescriptor::MakeOrdinal: 
//----------------------------------------------------------------------------------------
void TDescriptor::MakeOrdinal(const DescType data)
	{
	this->CopyData(typeAbsoluteOrdinal, (Ptr)&data, sizeof(DescType));
	} // TDescriptor::MakeOrdinal 
	
//----------------------------------------------------------------------------------------
// TDescriptor::MakeTypeOrInteger:
//
// This is useful for debugging; it makes a 'type' if the data
// appears to consist of printable characters (e.g. 'ABCD'), but
// otherwise it makes an integer (e.g. -12).
//----------------------------------------------------------------------------------------
void TDescriptor::MakeTypeOrInteger(const DescType data)
	{
	if(((data >= 'A000') && (data <= 'Zzzz')) || ((data >= 'a000') && (data <= 'zzzz')))
		this->MakeDescType(data);
	else
		this->MakeLong(data);
	} // TDescriptor::MakeTypeOrInteger 

//----------------------------------------------------------------------------------------
// TDescriptor::MakePoint: 
//----------------------------------------------------------------------------------------
void TDescriptor::MakePoint(const Point& thePoint)
	{
	this->CopyData(typeQDPoint, (Ptr)&thePoint, sizeof(Point));
	} // TDescriptor::MakePoint 

//----------------------------------------------------------------------------------------
// TDescriptor::MakePoint: 
//----------------------------------------------------------------------------------------
void TDescriptor::MakePoint(const short h, const short v)
	{
	Point thePoint;
	
	thePoint.v = v;
	thePoint.h = h;
	this->MakePoint(thePoint);
	} // TDescriptor::MakePoint 

//----------------------------------------------------------------------------------------
// TDescriptor::MakeRect: 
//----------------------------------------------------------------------------------------
void TDescriptor::MakeRect(const Rect& theRect)
	{
	this->CopyData(typeQDRectangle, (Ptr)&theRect, sizeof(Rect));
	} // TDescriptor::MakeRect 

//----------------------------------------------------------------------------------------
// TDescriptor::MakeString: 
//----------------------------------------------------------------------------------------
void TDescriptor::MakeString(ConstStr255Param data)
	{
	Size length = ((unsigned char*)(data))[0];
	
	this->CopyData(typeChar, (Ptr)&data[1], length);
	} // TDescriptor::MakeString 

//----------------------------------------------------------------------------------------
// TDescriptor::MakeString: 
//----------------------------------------------------------------------------------------
void TDescriptor::MakeString(const char* data)
	{
	Size length = strlen(data);
	
	this->CopyData(typeChar, (Ptr)&data[0], length);
	} // TDescriptor::MakeString 

//----------------------------------------------------------------------------------------
// TDescriptor::MakeDateTimeRec: 
//----------------------------------------------------------------------------------------
void TDescriptor::MakeDateTimeRec(const DateTimeRec dateTime)
	{
	this->CopyData(typeDateTimeRec, (Ptr)&dateTime, sizeof(DateTimeRec));
	} // TDescriptor::MakeDateTimeRec 

//----------------------------------------------------------------------------------------
// TDescriptor::MakeDateTimeRec: 
//----------------------------------------------------------------------------------------
void TDescriptor::MakeDateTimeRec(const long secsSince1904)
	{
	DateTimeRec dateTime;
	
	//
	// Let the OSUtilities do the hard work of converting seconds
	// since 1904 into a DateTimeRec
	//
	SecondsToDate(secsSince1904,&dateTime);
	this->MakeDateTimeRec(dateTime);
	} // TDescriptor::MakeDateTimeRec 

//----------------------------------------------------------------------------------------
// TDescriptor::MakeLongDateTimeRec: 
//----------------------------------------------------------------------------------------
void TDescriptor::MakeLongDateTimeRec(const LongDateRec dateTime)
	{
	this->CopyData(typeLongDateTimeRec, (Ptr)&dateTime, sizeof(LongDateRec));
	} // TDescriptor::MakeLongDateTimeRec 

//----------------------------------------------------------------------------------------
// TDescriptor::MakeLongDateTimeRec: 
//----------------------------------------------------------------------------------------
void TDescriptor::MakeLongDateTimeRec(const DateTimeRec dateTime)
	{
	LongDateRec longDateTime;
		
	longDateTime.od.eraAlt = 0;
	longDateTime.od.oldDate = dateTime;
	
	this->MakeLongDateTimeRec(longDateTime);
	} // TDescriptor::MakeLongDateTimeRec 

#if false

//----------------------------------------------------------------------------------------
// TDescriptor::MakeLongDateTimeRec: 
//----------------------------------------------------------------------------------------
void TDescriptor::MakeLongDateTimeRec(const long secsSince1904)
	{
	DateTimeRec dateTime;
	
	Secs2Date(secsSince1904,&dateTime);
	this->MakeLongDateTimeRec(dateTime);
	} // TDescriptor::MakeLongDateTimeRec 

#endif

//----------------------------------------------------------------------------------------
// TDescriptor::MakeLongDateTimeRec: 
//----------------------------------------------------------------------------------------
void TDescriptor::MakeLongDateTimeRec(LongDateTime lsecs)
	{
	LongDateRec longDateTime;
	
	LongSecondsToDate(&lsecs, &longDateTime);
	this->MakeLongDateTimeRec(longDateTime);
	} // TDescriptor::MakeLongDateTimeRec 

//----------------------------------------------------------------------------------------
// TDescriptor::MakeLongDateTime: 
//----------------------------------------------------------------------------------------
void TDescriptor::MakeLongDateTime(const LongDateTime lsecs)
	{
	this->CopyData(typeLongDateTime, (Ptr)&lsecs, sizeof(LongDateTime));
	} // TDescriptor::MakeLongDateTime 

//----------------------------------------------------------------------------------------
// TDescriptor::MakeFSS: 
//----------------------------------------------------------------------------------------
void TDescriptor::MakeFSS(const FSSpec& spec)
	{
	this->CopyData(typeFSS, (Ptr)&spec, sizeof(FSSpec));
	} // TDescriptor::MakeFSS 

//----------------------------------------------------------------------------------------
// TDescriptor::AdoptAlias: 
//
// Takes the provided alias handle and keeps it forever.
//----------------------------------------------------------------------------------------
void TDescriptor::AdoptAlias(Handle alias)
	{
	this->AdoptHandle(typeAlias, alias);
	} // TDescriptor::AdoptAlias 
	
//----------------------------------------------------------------------------------------
// TDescriptor::MakeAlias: 
//
// You would think that someone would have spent five minutes to make a coercion
// handler that goes from FSSpecs to Alias records, but nooooooo...
//----------------------------------------------------------------------------------------
void TDescriptor::MakeAlias(FSSpec& spec)
	{
	AliasHandle alias;
	
	FailErr(NewAliasMinimal(&spec, &alias));
	this->AdoptAlias((Handle)alias);
	} // TDescriptor::MakeAlias 

//----------------------------------------------------------------------------------------
// TDescriptor::MakeProcessSerialNumber: 
//----------------------------------------------------------------------------------------
void TDescriptor::MakeProcessSerialNumber(ProcessSerialNumber psn)
	{
	this->CopyData(typeProcessSerialNumber, (Ptr)&psn, sizeof(ProcessSerialNumber));
	} // TDescriptor::MakeProcessSerialNumber 

//----------------------------------------------------------------------------------------
// TDescriptor::MakeObjectSpecifier: 
//----------------------------------------------------------------------------------------
void TDescriptor::MakeObjectSpecifier(DescType desiredClass, TDescriptor container, DescType keyForm, TDescriptor keyData, Boolean disposeInputs)
	{
	FailErr(CreateObjSpecifier(desiredClass,container,keyForm,keyData,disposeInputs,*this));
	} // TDescriptor::MakeObjectSpecifier 

//----------------------------------------------------------------------------------------
// TDescriptor::MakeObjectSpecifierForProperty: 
//----------------------------------------------------------------------------------------
void TDescriptor::MakeObjectSpecifierForProperty(DescType propertyIdentifier, TDescriptor container, Boolean disposeInputs)
	{
	TDescriptor propertyIDDesc;
	
	propertyIDDesc.MakeDescType(propertyIdentifier);
	this->MakeObjectSpecifier(cProperty, container, formPropertyID, propertyIDDesc, disposeInputs);
	propertyIDDesc.Dispose();
	} // TDescriptor::MakeObjectSpecifierForProperty
	
//----------------------------------------------------------------------------------------
// TDescriptor::MakeSpecifierForPropertyOfObjectBeingExamined: 
//----------------------------------------------------------------------------------------
void TDescriptor::MakeSpecifierForPropertyOfObjectBeingExamined(DescType propertyIdentifier)
	{
	TDescriptor objectBeingExamined;
	
	//
	// 'objectBeingExamined' is a special object specifier that only
	// has meaning in the context of a comparison descriptor
	//
	objectBeingExamined.AdoptHandle(typeObjectBeingExamined,nil);
	
	//
	// result is "propertyIdentifier of (object being examined)"
	//
	this->MakeObjectSpecifierForProperty(propertyIdentifier, objectBeingExamined, true);
	} // TDescriptor::MakeSpecifierForPropertyOfObjectBeingExamined 

//----------------------------------------------------------------------------------------
// TDescriptor::MakeCompDescriptor: 
//----------------------------------------------------------------------------------------
void TDescriptor::MakeCompDescriptor(DescType comparisonOperator, TDescriptor comparitor, TDescriptor compareWith, Boolean disposeInputs)
	{
	CreateCompDescriptor(comparisonOperator, comparitor, compareWith, disposeInputs, *this);
	} // TDescriptor::MakeCompDescriptor 

//----------------------------------------------------------------------------------------
// TDescriptor::MakeCompDescriptor: 
//----------------------------------------------------------------------------------------
void TDescriptor::MakeCompDescriptor(DescType comparisonOperator, DescType propertyIdentifier, TDescriptor compareWith, Boolean disposeInputs)
	{
	TDescriptor operand1;

	operand1.MakeSpecifierForPropertyOfObjectBeingExamined(propertyIdentifier);

	//
	// Make a compDescriptor of operand1 compared with (comparisonOperator) compareWith
	//
	CreateCompDescriptor(comparisonOperator, operand1, compareWith, disposeInputs, *this);
	operand1.Dispose();
	} // TDescriptor::MakeCompDescriptor 

//----------------------------------------------------------------------------------------
// TDescriptor::GetBlock: 
//----------------------------------------------------------------------------------------
void TDescriptor::GetBlock(Ptr data, Size length, DescType desiredType) const
	{
	if(this->DataHandleIsNil())
		FailErr(errAECantSupplyType);
	
	if(this->DescriptorType() != desiredType)
		{
		TDescriptor coerceData;
		
		coerceData = this->Coerce(desiredType);
		coerceData.GetBlock(data,length,desiredType);
		coerceData.Dispose();
		}
	else
		{
		//
		// It would be just wrong to have a situation where a
		// descriptor is of the requested type, but not of the
		// requested length.  This method is only usable for
		// fixed-size blocks
		//
		if(GetHandleSize(fDataHandle) != length)
			{
			ASSERTPRINT(false, ("GetBlock encountered a descriptor whose data was the wrong length"));
			FailErr(errAECorruptData);
			}

		BlockMoveData(*fDataHandle, data, length);
		}
	} // TDescriptor::GetBlock 

//----------------------------------------------------------------------------------------
// TDescriptor::GetShort: 
//----------------------------------------------------------------------------------------
short TDescriptor::GetShort() const
	{
	short result = 0;

	this->GetBlock((Ptr)&result, sizeof(short), typeShortInteger);

	return result;
	} // TDescriptor::GetShort 
	
//----------------------------------------------------------------------------------------
// TDescriptor::GetLong: 
//----------------------------------------------------------------------------------------
long TDescriptor::GetLong() const
	{
	long result = 0;

	this->GetBlock((Ptr)&result, sizeof(long), typeLongInteger);

	return result;
	} // TDescriptor::GetLong 
	
//----------------------------------------------------------------------------------------
// TDescriptor::GetInt64: 
//----------------------------------------------------------------------------------------
Int64 TDescriptor::GetInt64() const
	{
	Int64 result = 0;

	this->GetBlock((Ptr)&result, sizeof(Int64), typeInt64);

	return result;
	} // TDescriptor::GetInt64 

//----------------------------------------------------------------------------------------
// TDescriptor::GetBoolean: 
//----------------------------------------------------------------------------------------
Boolean TDescriptor::GetBoolean() const
	{
	Boolean result = 0;

	this->GetBlock((Ptr)&result, sizeof(Boolean), typeBoolean);

	return result;
	} // TDescriptor::GetBoolean 

//----------------------------------------------------------------------------------------
// TDescriptor::GetDescType: 
//----------------------------------------------------------------------------------------
DescType TDescriptor::GetDescType() const
	{
	DescType result = 0;
	
	this->GetBlock((Ptr)&result, sizeof(DescType), typeType);
	
	return result;
	} // TDescriptor::GetDescType 

//----------------------------------------------------------------------------------------
// TDescriptor::GetKeyword: 
//----------------------------------------------------------------------------------------
AEKeyword TDescriptor::GetKeyword() const
	{
	AEKeyword result = 0;
	
	this->GetBlock((Ptr)&result, sizeof(AEKeyword), typeKeyword);
	
	return result;
	} // TDescriptor::GetKeyword 
	
//----------------------------------------------------------------------------------------
// TDescriptor::GetEnumeration: 
//----------------------------------------------------------------------------------------
DescType TDescriptor::GetEnumeration() const
	{
	DescType result = 0;
	
	this->GetBlock((Ptr)&result, sizeof(DescType), typeEnumeration);
	
	return result;
	} // TDescriptor::GetEnumeration 

//----------------------------------------------------------------------------------------
// TDescriptor::GetOrdinal: 
//----------------------------------------------------------------------------------------
DescType TDescriptor::GetOrdinal() const
	{
	DescType result = 0;
	
	this->GetBlock((Ptr)&result, sizeof(DescType), typeAbsoluteOrdinal);
	
	return result;
	} // TDescriptor::GetOrdinal 

//----------------------------------------------------------------------------------------
// TDescriptor::GetPoint: 
//----------------------------------------------------------------------------------------
Point TDescriptor::GetPoint() const
	{
	Point result;

	this->GetBlock((Ptr)&result, sizeof(Point), typeQDPoint);

	return result;
	} // TDescriptor::GetPoint 

//----------------------------------------------------------------------------------------
// TDescriptor::GetRect: 
//----------------------------------------------------------------------------------------
Rect TDescriptor::GetRect() const
	{
	Rect result;

	this->GetBlock((Ptr)&result, sizeof(Rect), typeQDRectangle);
	
	return result;
	} // TDescriptor::GetRect 

//----------------------------------------------------------------------------------------
// TDescriptor::GetString: 
//----------------------------------------------------------------------------------------
void TDescriptor::GetString(Str255& result) const
	{
	this->GetString(result, 255);
	}

//----------------------------------------------------------------------------------------
// TDescriptor::GetString: 
//----------------------------------------------------------------------------------------
void TDescriptor::GetString(unsigned char* result, short maxLength) const
	{
	result[0] = 0;

	if(this->DataHandleIsNil())
		FailErr(errAECantSupplyType);
		
	if(this->DescriptorType() == typeChar)
		{
		Size length = GetHandleSize(fDataHandle);
		if(length > maxLength)
			length = maxLength;
		BlockMoveData(*fDataHandle, (Ptr)&result[1], length);
		result[0] = (unsigned char) length;
		}
	else
		{
		TDescriptor makeDataString;
		
		makeDataString = this->Coerce(typeChar);
		ASSERT(makeDataString.DescriptorType() == typeChar);
		makeDataString.GetString(result);
		makeDataString.Dispose();
		}
	} // TDescriptor::GetString 
	
//----------------------------------------------------------------------------------------
// TDescriptor::GetString: 
//----------------------------------------------------------------------------------------
void TDescriptor::GetString(char* result) const
	{
	result[0] = 0;

	if(this->DataHandleIsNil())
		FailErr(errAECantSupplyType);
		
	if(this->DescriptorType() == typeChar)
		{
		Size length = GetHandleSize(fDataHandle);
		if(length > 255)
			length = 255;
		BlockMoveData(*fDataHandle, (Ptr)&result[0], length);
		result[length] = 0;
		}
	else
		{
		TDescriptor makeDataString;
		
		makeDataString = this->Coerce(typeChar);
		ASSERT(makeDataString.DescriptorType() == typeChar);
		makeDataString.GetString(result);
		makeDataString.Dispose();
		}
	} // TDescriptor::GetString 

//----------------------------------------------------------------------------------------
// TDescriptor::GetDateTimeRec: 
//----------------------------------------------------------------------------------------
DateTimeRec TDescriptor::GetDateTimeRec() const
	{
	DateTimeRec dateTime;
	
	//
	// We'll do our own coercion from typeLongInteger, if necessary.
	// We should really have a coercion handler for this
	//
	if(this->DescriptorType() == typeLongInteger)
		{
		long secsSince1904 = this->GetLong();
		SecondsToDate(secsSince1904,&dateTime);
		}
	else
		{
		//
		// Get the time date rec
		//
		this->GetBlock( (Ptr)&dateTime, sizeof(DateTimeRec), typeDateTimeRec);
		}
	
	return dateTime;
	} // TDescriptor::GetDateTimeRec 

//----------------------------------------------------------------------------------------
// TDescriptor::GetFSS: 
//----------------------------------------------------------------------------------------
void TDescriptor::GetFSS(FSSpec& spec) const
	{
	this->GetBlock((Ptr)&spec, sizeof(FSSpec), typeFSS);
	} // TDescriptor::GetFSS 

//----------------------------------------------------------------------------------------
// ReverseComparisonOperator: 
//
// Sometimes it cannot be avoided:  we must convert from "a compareisonOp b" to
// "b reverseComparisonOp a".  This routine determines what the reverse operation
// should be.
//----------------------------------------------------------------------------------------
DescType ReverseComparisonOperator(DescType comparisonOperator)
	{
	DescType result = comparisonOperator;
	
	switch(comparisonOperator)
		{
		case kAEEquals:
		case kASNotEqual:
			{
			result = comparisonOperator;
			}
			break;
			
		case kAEGreaterThan:
			{
			result = kAELessThan;
			}
			break;
		
		case kAEGreaterThanEquals:
			{
			result = kAELessThanEquals;
			}
			break;
		
		case kAELessThan:
			{
			result = kAEGreaterThan;
			}
			break;
		
		case kAELessThanEquals:
			{
			result = kAEGreaterThanEquals;
			}
			break;

		case kAEBeginsWith:
			{
			result = kAEAtTheBeginningOf;
			}
			break;
			
		case kAEEndsWith:
			{
			result = kAEAtTheEndOf;
			}
			break;
			
		case kAEContains:
			{
			result = kAEIsContainedIn;
			}
			break;

		case kAEAtTheBeginningOf:
			{
			result = kAEBeginsWith;
			}
			break;
			
		case kAEAtTheEndOf:
			{
			result = kAEEndsWith;
			}
			break;
			
		case kAEIsContainedIn:
			{
			result = kAEContains;
			}
			break;
		
		//
		// Script:  We should fail if we don't recognize the comparison operator
		//
		default:
			{
			result = 0;
			}
		}
	
	return result;
	}
	
//----------------------------------------------------------------------------------------
// InterpretCompareResult:
//
// Given two keys and a comparison opperator, this function determines if
// the comparison between key2 and key1 is true or false.  key1 and key2 should
// be either two long integers representing the various values of the two descriptors
// being compared, or key1 may be set to 0 and key2 set to -1 (key2 < key1), 0
// (key1 == key2) or 1 (key2 > key1).
//
// This function is used by GeneralCompare, which is used by TDescriptor::Compare
//----------------------------------------------------------------------------------------
Boolean InterpretCompareResult(DescType comparisonOperator, long key1, long key2 )
	{
	switch(comparisonOperator)
		{
		case kAEEquals:
			{
			return key1 == key2;
			}
		
		case kASNotEqual:
			{
			return key1 != key2;
			}
		
		case kAEGreaterThan:
			{
			return key1 > key2;
			}
		
		case kAEGreaterThanEquals:
			{
			return key1 >= key2;
			}
		
		case kAELessThan:
			{
			return key1 < key2;
			}
		
		case kAELessThanEquals:
			{
			return key1 <= key2;
			}
		
		//
		//  Perhaps we should fail if we don't recognize the comparisonOperator
		//
		default:
			{
			return false;
			}
		}
	} // InterpretCompareResult 

#if 0 

//----------------------------------------------------------------------------------------
// GeneralCompare: 
//
// This is a general comparison routine that works on streams of bytes.  It is assumed
// that the data being compared follows the same ordering rules as strings.  The
// 'caseSensitive' flag should always be true for non-string objects, of course.
//----------------------------------------------------------------------------------------
Boolean GeneralCompare(DescType comparisonOperator, Ptr data, Size length, Ptr compareWith, Size comparisonLength, Boolean caseSensitive)
	{
	Boolean comparisonResult = false;
	long i = 0;
		
	switch(comparisonOperator)
		{
		case kAEEquals:		
		case kASNotEqual:
		case kAEGreaterThan:
		case kAEGreaterThanEquals:
		case kAELessThan:
		case kASLessThanOrEqual:
			{
			short intermediateResult = 0;
			Size max = (length < comparisonLength ? length : comparisonLength);
			
			//
			// We have a little bit of code duplication here
			// to make the code run a little faster
			//
			if(caseSensitive)
				{
				for(i=0; i < max; ++i)
					{
					if(*((unsigned char*)compareWith) > *((unsigned char*)data))
						{
						intermediateResult = 1;
						break;
						}
					else if(*((unsigned char*)compareWith) < *((unsigned char*)data))
						{
						intermediateResult = -1;
						break;
						}
					++compareWith;
					++data;
					}
				}
			else
				{
				for(i=0; i < max; ++i)
					{
					unsigned char compareWithChar = *((unsigned char*)compareWith);
					unsigned char dataChar = *((unsigned char*)data);
					
					if((compareWithChar >= 'a') && (compareWithChar <= 'z'))
						compareWithChar = compareWithChar - 'a' + 'A';
					if((dataChar >= 'a') && (dataChar <= 'z'))
						dataChar = dataChar - 'a' + 'A';
						
					if(compareWithChar > dataChar)
						{
						intermediateResult = 1;
						break;
						}
					else if(compareWithChar < dataChar)
						{
						intermediateResult = -1;
						break;
						}
					++compareWith;
					++data;
					}
				}
			
			if(intermediateResult == 0)
				intermediateResult = (comparisonLength > length) - (comparisonLength < length);
			
			comparisonResult = InterpretCompareResult(comparisonOperator,0,intermediateResult);
			
			break;
			}
		
		case kAEBeginsWith:
			{
			comparisonResult = GeneralCompare(kAEEquals, data, comparisonLength, compareWith, comparisonLength,caseSensitive);
			break;
			}
			
		case kAEEndsWith:
			{
			i = (length - comparisonLength);
			if(i >= 0)
				comparisonResult = GeneralCompare(kAEEquals, data + i, comparisonLength, compareWith, comparisonLength,caseSensitive);
			break;
			}
			
		case kAEContains:
			{
			for(i=0;i<=length - comparisonLength;++i)
				{
				if(GeneralCompare(kAEEquals, data + i, comparisonLength, compareWith, comparisonLength,caseSensitive))
					{
					comparisonResult = true;
					break;
					}
				}
			
			break;
			}
			
		default:
			{
			comparisonResult = false;
			}
		}
	
	return comparisonResult;
	} // GeneralCompare 

#endif

//----------------------------------------------------------------------------------------
// StringContains: 
//----------------------------------------------------------------------------------------
Boolean	StringContains(Str255 thisString, Str255 withString)
	{
	unsigned char thisSize = (unsigned char)thisString[0];
	unsigned char withSize = (unsigned char)withString[0];

	if ((thisSize == 0) || (withSize > thisSize))		// quick check
		return false;

	Boolean b = false;
	unsigned char* a = &thisString[0];
	unsigned char* last = a + thisSize - withSize + 1;

#if 0
	do
		{
		b = IdenticalText(Ptr(++a), Ptr(withString[1]), withSize, withSize, gEnvironment->Intl2());
		}
	while (b && (a < last));
#endif

	return (b == false);
	} // StringContains 

#if 0

//----------------------------------------------------------------------------------------
// TDescriptor::Compare: 
//----------------------------------------------------------------------------------------
Boolean TDescriptor::Compare(DescType comparisonOperator, Str255 withString) const
	{
	Str255 thisString;
	Size thisSize;
	Size withSize;
	Boolean compareResult = false;
	
	//
	// Fetch a copy of 'thisString' (which we will freely destroy later)
	//
	this->GetString(thisString);
	thisSize = (unsigned char)thisString[0];
	withSize = (unsigned char)withString[0];

	switch(comparisonOperator)
		{
		//
		// Check to see if we can find the string anywhere inside
		//
		case kAEContains:
			{
			long	i;

#if 0			
			unsigned char* a = &thisString[0];
			for(i=0;i<thisSize - withSize + 1;++i)
				{
				*a = (unsigned char)withSize;
				compareResult = (IdenticalString((ConstStr255Param) a, (ConstStr255Param) withString, gEnvironment->Intl2()) == 0);
				if(compareResult == true)
					break;
				++a;
				}
#endif
			}
			break;
		
		//
		// Relational tests are easy:  use CompareString
		//
		case kAEGreaterThan:
		case kAEGreaterThanEquals:
		case kAELessThan:
		case kASLessThanOrEqual:
			{
#if 0
			compareResult = InterpretCompareResult(comparisonOperator,
						CompareString(thisString, withString, gEnvironment->Intl2()), 0);
#endif
			}
			break;
		
		//
		// For AEEndsWith, move the end of the string down to
		// the beginning of the string, then fall through to
		// the 'begins with' case
		//
		// International allert:  this could clip a two-byte
		// character in half, but doing so probably won't cause
		// a problem, since we are looking for exact matches.
		//
		case kAEEndsWith:
			{
			if(withSize <= thisSize)
				BlockMoveData((Ptr)&thisString[thisSize - withSize + 1],(Ptr)&thisString[1],withSize);
			}
			// !!! fall through
			
		//
		// For kAEBeginsWith, set the length of the 'thisString'
		// to the length of the 'withString' and do an equality test
		//
		// International allert:  this could clip a two-byte
		// character in half, but doing so probably won't cause
		// a problem, since we are looking for exact matches.
		//
		case kAEBeginsWith:
			{
			thisString[0] = (unsigned char)withSize;
			}
			// !!! fall through...
		
		//
		// kAEEquals and kASNotEqual must come immediately after
		// begins with and ends with
		//
		case kAEEquals:
		case kASNotEqual:
			{
#if 0
			compareResult = (IdenticalString(thisString, withString, gEnvironment->Intl2()) == 0);
			if(comparisonOperator == kASNotEqual)
				compareResult = !compareResult;
#endif
			}
			break;
		
		//
		// Shouldn't ever get here...
		//
		default:
			{
			compareResult = GeneralCompare(comparisonOperator,(Ptr)&thisString[1],thisSize,(Ptr)&withString[1],withSize,false);
			}
			break;
		}
	
	return compareResult;
	} // TDescriptor::Compare 

//----------------------------------------------------------------------------------------
// TDescriptor::CompareRelativeTime: 
//----------------------------------------------------------------------------------------
Boolean TDescriptor::CompareRelativeTime(DescType comparisonOperator, TDescriptor& compareWith) const
	{
	Boolean result = false;
	
	if(comparisonOperator != kAEEquals)
		FailErr(errAEEventNotHandled);

	unsigned long currentTimeInSecs;
	//unsigned long startTime;
	//unsigned long finishTime;
	
	// get the time - yesterday, today, last year, etc.-- then work the relationship
	DateTimeRec currentTimeInDateRec;
	GetDateTime (&currentTimeInSecs);								// get current time in seconds
	GetTime (&currentTimeInDateRec);								// get current time in date record form

	// set up date record for today at midnight
	DateTimeRec todayTimeInDateRec;									// today at midnight
	todayTimeInDateRec = currentTimeInDateRec;				
	todayTimeInDateRec.hour = 0;									// reset hours, minutes, seconds but leave day and month
	todayTimeInDateRec.minute = 0;
	todayTimeInDateRec.second = 0;

	// now convert today at midnight to seconds
	unsigned long todayTimeInSecs;									// today at midnight in seconds
	DateToSeconds (&todayTimeInDateRec, &todayTimeInSecs);
	
	if(compareWith.DescriptorType() == typeRelativeTime)
		{
		result = this->GetLong() >= currentTimeInSecs - compareWith.GetRelativeTime();
		}
	else if(compareWith.DescriptorType() == typeConceptualTime)
		{
		unsigned long weekStartTimeInSecs;
		unsigned long lastWeekStartTimeInSecs;
		switch(compareWith.GetConceptualTime())
			{
			case enumToday:
				// midnight today until this moment
				result = (this->GetLong() >= todayTimeInSecs && this->GetLong() <= currentTimeInSecs);
				break;
				
			case enumYesterday:
				// midnight yesterday to midnight today
				unsigned long yesterdayTimeInSecs;					// yesterday at midnight
				yesterdayTimeInSecs = todayTimeInSecs - (1 * kDayLengthInSecs);
				result = (this->GetLong() >= yesterdayTimeInSecs && this->GetLong() <= todayTimeInSecs);
				break;

			case enumThisWeek:
				// Find time in seconds at start of week
				weekStartTimeInSecs = todayTimeInSecs - ((todayTimeInDateRec.dayOfWeek - 1) * kDayLengthInSecs);
				result = (this->GetLong() >= weekStartTimeInSecs && this->GetLong() <= currentTimeInSecs);
				break;


			case enumLastWeek:
				// Find time in seconds at start of week
				weekStartTimeInSecs = todayTimeInSecs - ((todayTimeInDateRec.dayOfWeek - 1) * kDayLengthInSecs);
				lastWeekStartTimeInSecs = weekStartTimeInSecs - (7 * kDayLengthInSecs);
				result = (this->GetLong() >= lastWeekStartTimeInSecs && this->GetLong() <= weekStartTimeInSecs);
				break;

			case enumThisMonth:
			case enumLastMonth:
			case enumThisYear:
			case enumLastYear:
			
			default:
				FailErr(errAEEventNotHandled);
			}
		}
	else
		FailErr(errAEEventNotHandled);
	
	return result;
	} // TDescriptor::CompareRelativeTime 

#endif

//----------------------------------------------------------------------------------------
// TDescriptor::Compare: 
//
// Compare two descriptors.  ComparisonOperator should be one of the following:
//
// kAEEquals, kASNotEqual, kAEGreaterThan, kAEGreaterThanEquals, kAELessThan,
// kASLessThanOrEqual, kAEBeginsWith, kAEEndsWith, or kAEContains
//----------------------------------------------------------------------------------------
Boolean TDescriptor::Compare(DescType comparisonOperator, TDescriptor& compareWith) const
	{
	Boolean compareResult = false;
	
	//
	// It might be nice to try to do a type coercion here, but for now
	// we only compare descriptors of the same type
	//
	if(this->DescriptorType() != compareWith.DescriptorType())
		FailErr(errAEEventNotHandled);

	return CompareTypedData(comparisonOperator, *this, compareWith);
	} // TDescriptor::Compare

//----------------------------------------------------------------------------------------
// CompareTypedData
//----------------------------------------------------------------------------------------
Boolean CompareTypedData(DescType comparisonOperator, const TConstDataReference& compareThisRef, const TConstDataReference& compareWithRef)
	{
	Boolean compareResult = false;
	
	switch(comparisonOperator)
		{
		case kAEEquals:		
		case kASNotEqual:
		case kAEGreaterThan:
		case kAEGreaterThanEquals:
		case kAELessThan:
		case kASLessThanOrEqual:
			{
			compareResult = InterpretCompareResult(comparisonOperator, compareThisRef.Compare(compareWithRef), 0 );
			break;
			}
		
		//
		// Question:  is this begins-with clipping bogus?
		//		
		case kAEBeginsWith:
			{
			if(compareWithRef.DataLength() > compareThisRef.DataLength())
				compareResult = false;
			else
				{
				TConstDataReference beginningOfThis(compareThisRef.DataType(), compareThisRef.Data(), compareWithRef.DataLength());
				compareResult = InterpretCompareResult(kAEEquals, beginningOfThis.Compare(compareWithRef), 0 );
				}
			break;
			}
			
		//
		// Question:  is this ends-with clipping bogus?
		//		
		case kAEEndsWith:
			{
			long sizeDifference = compareThisRef.DataLength() - compareWithRef.DataLength();
			if(sizeDifference < 0)
				compareResult = false;
			else
				{
				TConstDataReference endOfThis(compareThisRef.DataType(), compareThisRef.Data() + sizeDifference, compareWithRef.DataLength());
				compareResult = InterpretCompareResult(kAEEquals, endOfThis.Compare(compareWithRef), 0 );
				}
			break;
			}

		case kAEContains:
			{
			compareResult = compareThisRef.Contains(compareWithRef);
			break;
			}

		case kAEAtTheBeginningOf:
			{
			compareResult = CompareTypedData(kAEBeginsWith, compareWithRef, compareThisRef);
			break;
			}
			
		case kAEAtTheEndOf:
			{
			compareResult = CompareTypedData(kAEEndsWith, compareWithRef, compareThisRef);
			break;
			}

		case kAEIsContainedIn:
			{
			compareResult = compareWithRef.Contains(compareThisRef);
			break;
			}
		
		default:
			{
			FailErr(errAEEventNotHandled);
			break;
			}
		}
	
	return compareResult;
	} // CompareTypedData

//========================================================================================
// CLASS TDescriptorLoop
//
// The class TDescriptorLoop is used by the macro FOREACHDESCRIPTOR.  Note that the
// descriptor returned by "Next" is a copy that must be disposed of by the programmer.
//
// We could have TDescriptorLoop keep track of the descriptor it created, and have it
// dispose it when done.  This would necessitate an 'orphan' method, though, in case
// someone else wanted to adopt the descriptor.
//========================================================================================

#ifdef notinlineconst
//----------------------------------------------------------------------------------------
// TDescriptorLoop::TDescriptorLoop: 
//----------------------------------------------------------------------------------------
TDescriptorLoop::TDescriptorLoop(const TDescriptor* dl)
	{
	fDescriptorList = dl;
	fIndex = 0;
	fCount = dl->CountItems();	
	} // TDescriptorLoop::TDescriptorLoop 
#endif
//----------------------------------------------------------------------------------------
// TDescriptorLoop::Next: 
//
// If 'key' is not nil, it is filled in with the keyword from the i'th item of the
// list (if the list is a record)
//
// If 'dataHandle' is not nil, it is filled in with the data handle from the i'th
// item of the list.  This is only done to be nice to the FOREACHTOKEN macro.
//----------------------------------------------------------------------------------------
Boolean TDescriptorLoop::Next(TDescriptor& d, AEKeyword* key, Handle* dataHandle)
	{
	Boolean moreToDo = false;
	
	//
	// Set up on the first time through the loop
	//
	if(fIndex == 0)
		{
		fCount = fDescriptorList->CountItems();
		if(d.IsNullDescriptor() == false)
			{
			ASSERTPRINT(false, ("TDescriptorLoop::Next:  'd' was not null on first iteration!"));
			FailErr(errAEEventFailed);
			}
		}

	//
	// Advance to the next item in the list...
	//
	fIndex++;
	
	//
	// Dispose of the 'd' from last time through the loop
	//
	d.Dispose();
	
	//
	// Are there still items in the list?
	//
	if (fIndex <= fCount)
		{
		//
		// At this point, we could check to see if 'fDescriptorList' actually
		// points to something that is _not_ a list; if so, we coule return a
		// _reference_ to the single object, and set a field that indicates that
		// 'd' is a reference, and should not be disposed next time through the
		// loop.  The reason we do not do this is that 'd' will be disposed by
		// the client code if there is a failure, and there is no good way for
		// the client to know if d is a reference or not.  We could remember
		// what 'd' was and clean it up in our destructor, but failures can
		// cause our TDescriptorLoop object to be orphaned without ever being
		// destructed.
		//
		d = fDescriptorList->GetNthDescriptor(fIndex, typeWildCard, key);
		
		moreToDo = true;
		}
	//
	// No more items, clean up
	//
	else
		{
		if(key != nil)
			*key = typeNull;
		d.MakeNull();
		}
	
	//
	// Fill in the 'dataHandle' parameter if it was passed in.
	// (We only do this to make life easy for the FOREACHTOKEN macro)
	//
	if(dataHandle != nil)
		*dataHandle = d.DataHandle();
	
	return moreToDo;
	} // TDescriptorLoop::Next 


//----------------------------------------------------------------------------------------
// TDescriptor::MakeEmptyList:
//
// Make an empty descriptor list.  It is usually not necessary to make an empty list,
// because all of the TDescriptor routines will interpret null descriptors as
// empty lists.  Sometimes it is necessary to return an empty list as the result of
// an AppleEvent handler, though, and in this case a null descriptor is _not_ equivalent.
//
// By default, include no factored data and make an AEList.
//----------------------------------------------------------------------------------------
void TDescriptor::MakeEmptyList()
	{
	this->CreateList(kMakeAEList);
	} // TDescriptor::MakeEmptyList 

//----------------------------------------------------------------------------------------
// TDescriptor::MakeList: 
//
// If this descriptor is null, an empty list is created.  If this descriptor is not
// empty, then a list is created and this descriptor is placed inside it.  If this
// descriptor is already a list, then no action is taken.
//
// DANGER:  Does a coerce-in-place!
//----------------------------------------------------------------------------------------
void TDescriptor::MakeList()
	{
	if(this->IsNullDescriptor())
		{
		this->CreateList();
		}
	else if(this->DescriptorType() != typeAEList)
		{
		this->CoerceInPlace(typeAEList);
		}
	} // TDescriptor::MakeList 
	
//----------------------------------------------------------------------------------------
// TDescriptor::CountItems:
//
// Count the items in a descriptor list. If this descriptor is not of typeAEList, then it
// only has one item, and if the descriptor is null, then it has no items.
//----------------------------------------------------------------------------------------
long TDescriptor::CountItems() const
	{
	long items = 1;
	if(this->IsNullDescriptor())
		items = 0;
	else if((this->DescriptorType() == typeAEList) || (this->DescriptorType() == typeAERecord))
		FailErr(AECountItems(*this, &items));
	return items;
	} // TDescriptor::CountItems 

//----------------------------------------------------------------------------------------
// TDescriptor::GetNthDescriptor:
//
// Get the indexed descriptor from an event.  This method will fail if given an index
// less than one or greater than the count returned by TDescriptor::CountItems.
// The descriptor returned is _always_ a copy that must be disposed of with
// TDescriptor::Dispose.
//----------------------------------------------------------------------------------------
TDescriptor TDescriptor::GetNthDescriptor(long index, DescType desiredType, AEKeyword* key) const
	{
	TDescriptor descriptor;
	AEKeyword ignoreKey;
	
	if(key == nil)
		key = &ignoreKey;
		
	//
	// We can "GetNthDescriptor" on an item that is not of typeAEList
	// if the index is one; we do this by cloning ourself
	// (n.b. AERecords are also lists)
	//
	if((DescriptorType() != typeAEList) && (DescriptorType() != typeAERecord))
		{
		if(index != 1)
			FailErr(-1);		//  index out of range
		
		*key = typeWildCard; 
		descriptor = this->Clone();
		}
	else
		{
		FailErr(AEGetNthDesc(*this, index, desiredType, key, descriptor));
		}
	
	return descriptor;
	} // TDescriptor::GetNthDescriptor 

//----------------------------------------------------------------------------------------
// TDescriptor::AddDescriptor:
//
// Add data to a descriptor list at a specified index, replacing any item already stored
// at that position.
//----------------------------------------------------------------------------------------
void TDescriptor::AddDescriptor(long index, TDescriptor& data)
	{
	//
	// If the list is completely empty, just put the data directly into
	// this descriptor (don't make a list)
	//
	if(this->IsNullDescriptor())
		{
		this->CopyDesc(data);
		}
	else
		{
		// If we're not a list yet, we'd better make ourselves one now.
		if(this->DescriptorType() != typeAEList)
			this->MakeList();
		FailErr(AEPutDesc(*this, index, data));
		}
	} // TDescriptor::AddDescriptor 

//----------------------------------------------------------------------------------------
// TDescriptor::AddDescriptor:
//
// Add data to a descriptor list at the end of the list
//----------------------------------------------------------------------------------------
void TDescriptor::AddDescriptor(TDescriptor& data)
	{
	this->AddDescriptor(0,data);
	} // TDescriptor::AddDescriptor 

//----------------------------------------------------------------------------------------
// TDescriptor::AddData: 
//----------------------------------------------------------------------------------------
void TDescriptor::AddData(long index, DescType descType, Ptr data, Size length)
	{
	//
	// If the list is completely empty, we'd better make an empty list
	// (the type coercion from null to typeAEList probably works,
	// but I have not tested it yet; the following two lines may
	// be superfluous)
	//
	if(this->IsNullDescriptor())
		this->MakeEmptyList();
	
	//
	// If we're not a list yet, we'd better make ourselves one now.
	//
	if(this->DescriptorType() != typeAEList)
		this->CoerceInPlace(typeAEList);
	
	//
	// Put the data into the list
	//
	FailErr(AEPutPtr(*this,index,descType,data,length));
	} // TDescriptor::AddData 

//----------------------------------------------------------------------------------------
// TDescriptor::AddData: 
//----------------------------------------------------------------------------------------
void TDescriptor::AddData(DescType descType, Ptr data, Size length)
	{
	this->AddData(0,descType,data,length);
	} // TDescriptor::AddData 

//----------------------------------------------------------------------------------------
// TDescriptor::AddLong: 
//----------------------------------------------------------------------------------------
void TDescriptor::AddLong(long number)
	{
	this->AddData(typeLongInteger,(Ptr)&number,sizeof(long));
	} // TDescriptor::AddLong 

//----------------------------------------------------------------------------------------
// TDescriptor::AddType: 
//----------------------------------------------------------------------------------------
void TDescriptor::AddType(DescType descType)
	{
	this->AddData(typeType,(Ptr)&descType,sizeof(DescType));
	} // TDescriptor::AddType 

//----------------------------------------------------------------------------------------
// TDescriptor::AppendList:
//
// Append a list onto the end of this one.  The list passed in may be a null descriptor
// or a single item as well as a list.
//----------------------------------------------------------------------------------------
void TDescriptor::AppendList(const TDescriptor& list)
	{
	TDescriptor descriptor;
	OSErr err = noErr;
	
	//
	// Take no action if our input is null
	//
	if(list.IsNullDescriptor() == false)
		{
		//
		// If this descriptor is a null descriptor, then we just want
		// to copy the incoming list.
		//
		// In the past we did not do this check, which resulted in
		// the unfortunate side effect of converting a single-item
		// list into a single (non-list) descriptor whenever said
		// list was 'AppendList'ed into an empty list.  While this
		// may be more efficient, it really hosed us over in the case
		// of GetData on the selection property, which is SUPPOSED
		// to return a single-item list (as opposed to a single
		// descriptor) if there is only one item in the list.  (This
		// is quite different from most other objects, which are
		// supposed to return single descriptors instead of single-
		// item lists whenever there is only one object in the result.)
		//
		if(this->IsNullDescriptor())
			{
			this->CopyDesc(list);
			}
		//
		// Script:  in the past we did not check to see
		// if the list was an AERecord; this was very
		// bad because appending mutliple records together
		// into a list didn't work correctly (we would
		// FOREACHDESCRIPTOR over the record--which works--
		// and add the data without its key!).
		//
		// Actually, we'll be a bit more efficient and
		// test for non-lists instead
		//
		else if(list.DescriptorType() != typeAEList)
			this->AddDescriptor(list);		
		else
			{
			Try
				{
				FOREACHDESCRIPTOR(&list, descriptor)
					{
					this->AddDescriptor(descriptor);
					}
				}
			Catch(err)
				{
				descriptor.Dispose();
				Throw(err);
				}
			}
		}
	} // TDescriptor::AppendList 

//----------------------------------------------------------------------------------------
// TDescriptor::AdoptList: 
//----------------------------------------------------------------------------------------
void TDescriptor::AdoptList(TDescriptor* list)
	{
	if(this->IsNullDescriptor())
		{
		this->AdoptDesc(*list);
		}
	else
		{
		this->AppendList(*list);
		list->Dispose();
		}
	} // TDescriptor::AdoptList 

//----------------------------------------------------------------------------------------
// TDescriptor::MakeAERecord:
//
// CreateList creates both AERecords and AELists.
//----------------------------------------------------------------------------------------
void TDescriptor::MakeAERecord()
	{
	this->CreateList(kMakeAERecord);
	} // TDescriptor::MakeAERecord 
	
//----------------------------------------------------------------------------------------
// TDescriptor::GetDescriptor:
//
// Get the ae descriptor from the event. If the desired type is not specified, it
// defaults to typeWildCard
//----------------------------------------------------------------------------------------
TDescriptor TDescriptor::GetDescriptor(AEKeyword key, DescType desiredType) const
	{
	TDescriptor result;

	//
	// n.b. This object must be typeAERecord or typeAppleEvent, or this
	// call will fail.  Other record-like descriptors (e.g. object specifiers)
	// cannot be passed to AEGetParamDesc.
	//
	FailErr(AEGetParamDesc(*this, key, desiredType, result));
	
	return result;
	} // TDescriptor::GetDescriptor 
	
//----------------------------------------------------------------------------------------
// TDescriptor::GetOptionalParameter:
//
// Like 'GetDescriptor, but returns a null descriptor if the specified parameter could not
// be extracted.
//----------------------------------------------------------------------------------------
TDescriptor TDescriptor::GetOptionalParameter(AEKeyword key, DescType desiredType) const
	{
	TDescriptor result;
	OSErr err = noErr;

	//
	// It's probably not necessary to call MakeNull if there is
	// an error, but we might as well be safe (paranoid?).
	//
	err = AEGetParamDesc(*this, key, desiredType, result);
	if(err != noErr)
		result.MakeNull();
	
	return result;
	} // TDescriptor::GetOptionalParameter 

//----------------------------------------------------------------------------------------
// TDescriptor::GetParameterPtr:
//
// Get the ae descriptor from the record
//----------------------------------------------------------------------------------------
void TDescriptor::GetParameterPtr(AEKeyword key, DescType desiredType, DescType *typeCode, Ptr dataPtr, Size maximumSize, Size *actualSize) const
	{
	//
	// n.b. This object must be typeAERecord or typeAppleEvent, or this
	// call will fail.  Other record-like descriptors (e.g. object specifiers)
	// cannot be passed to AEGetParamPtr.
	//
	FailErr(AEGetParamPtr(*this, key, desiredType, typeCode, dataPtr, maximumSize, actualSize));
	} // TDescriptor::GetParameterPtr 
	
//----------------------------------------------------------------------------------------
// TDescriptor::GetLongParameter:
//
// Get the first longword of data from a parameter coerced to the specified type
// (usually typeLongInteger)
//----------------------------------------------------------------------------------------
long TDescriptor::GetLongParameter(AEKeyword key, DescType desiredType) const
	{
	long actualSize;
	DescType typeCode;
	long result = 0;
	GetParameterPtr(key, desiredType, &typeCode, (Ptr) &result, sizeof(long), &actualSize);
	
	return result;
	} // TDescriptor::GetLongParameter 
	
//----------------------------------------------------------------------------------------
// TDescriptor::GetInt64Parameter:
//----------------------------------------------------------------------------------------
Int64 TDescriptor::GetInt64Parameter(AEKeyword key) const
	{
	long actualSize;
	DescType typeCode;
	Int64 result = 0;
	GetParameterPtr(key, typeInt64, &typeCode, (Ptr) &result, sizeof(Int64), &actualSize);
	
	return result;
	} // TDescriptor::GetInt64Parameter 
	
//----------------------------------------------------------------------------------------
// TDescriptor::GetBooleanParameter:
//----------------------------------------------------------------------------------------
Boolean TDescriptor::GetBooleanParameter(AEKeyword key) const
	{
	long actualSize;
	DescType typeCode;
	Boolean result = 0;
	GetParameterPtr(key, typeBoolean, &typeCode, (Ptr) &result, sizeof(Boolean), &actualSize);
	
	return result;
	} // TDescriptor::GetBooleanParameter 
	
//----------------------------------------------------------------------------------------
// TDescriptor::GetDescTypeParameter:
//----------------------------------------------------------------------------------------
DescType TDescriptor::GetDescTypeParameter(AEKeyword key) const
	{
	return (DescType)this->GetLongParameter(key, typeType);
	} // TDescriptor::GetDescTypeParameter 
	
//----------------------------------------------------------------------------------------
// TDescriptor::GetKeywordParameter:
//----------------------------------------------------------------------------------------
AEKeyword TDescriptor::GetKeywordParameter(AEKeyword key) const
	{
	return (AEKeyword)this->GetLongParameter(key, typeKeyword);
	} // TDescriptor::GetKeywordParameter 

//----------------------------------------------------------------------------------------
// TDescriptor::GetEnumerationParameter:
//----------------------------------------------------------------------------------------
DescType TDescriptor::GetEnumerationParameter(AEKeyword key) const
	{
	return (DescType)this->GetLongParameter(key, typeEnumeration);
	} // TDescriptor::GetEnumerationParameter 

//----------------------------------------------------------------------------------------
// TDescriptor::GetOrdinalParameter:
//----------------------------------------------------------------------------------------
DescType TDescriptor::GetOrdinalParameter(AEKeyword key) const
	{
	return (DescType)this->GetLongParameter(key, typeAbsoluteOrdinal);
	} // TDescriptor::GetOrdinalParameter 

//----------------------------------------------------------------------------------------
// TDescriptor::GetPointParameter:
//----------------------------------------------------------------------------------------
Point TDescriptor::GetPointParameter(AEKeyword key) const
	{
	long actualSize;
	DescType typeCode;
	Point result;
	GetParameterPtr(key, typeQDPoint, &typeCode, (Ptr) &result, sizeof(Point), &actualSize);
	
	return result;
	} // TDescriptor::GetPointParameter 

//----------------------------------------------------------------------------------------
// TDescriptor::GetRectParameter:
//----------------------------------------------------------------------------------------
Rect TDescriptor::GetRectParameter(AEKeyword key) const
	{
	long actualSize;
	DescType typeCode;
	Rect result ;
	GetParameterPtr(key, typeRectangle, &typeCode, (Ptr) &result, sizeof(Rect), &actualSize);
	
	return result;
	} // TDescriptor::GetRectParameter 

//----------------------------------------------------------------------------------------
// TDescriptor::GetObjectSpecifierParameters:
//----------------------------------------------------------------------------------------
void TDescriptor::GetObjectSpecifierParameters(DescType& desiredClass, DescType& keyForm, TDescriptor* keyData, TDescriptor* containerDesc) const
	{
	TDescriptor thisAsRecord = this->Coerce(typeAERecord);
		
	//
	// Extract the desired class, the key form and the key data from
	// the object specifier
	//
	desiredClass		= thisAsRecord.GetDescTypeParameter(keyAEDesiredClass);
	keyForm				= thisAsRecord.GetEnumerationParameter(keyAEKeyForm);
	if(keyData != nil)
		*keyData		= thisAsRecord.GetDescriptor(keyAEKeyData);
	if(containerDesc != nil)
		*containerDesc	= thisAsRecord.GetDescriptor(keyAEContainer);
	
	thisAsRecord.Dispose();
	} // TDescriptor::GetObjectSpecifierParameters
	
//----------------------------------------------------------------------------------------
// TDescriptor::GetDirectObject:
//
// Get the direct object of the event
//----------------------------------------------------------------------------------------
TDescriptor TDescriptor::GetDirectObject() const
	{
	return GetDescriptor(keyDirectObject);
	} // TDescriptor::GetDirectObject 

//----------------------------------------------------------------------------------------
// TDescriptor::GetErrorCode:
//
// Get an error code
//----------------------------------------------------------------------------------------
long TDescriptor::GetErrorCode() const
	{
	return GetLongParameter(keyErrorNumber);
	} // TDescriptor::GetErrorCode 

//----------------------------------------------------------------------------------------
// TDescriptor::PutDescriptor:
//
// Put a TDescriptor into the event
//----------------------------------------------------------------------------------------
void TDescriptor::PutDescriptor(AEKeyword key, TDescriptor data)
	{
	FailErr(AEPutParamDesc(*this, key, data));
	} // TDescriptor::PutDescriptor 
	
//----------------------------------------------------------------------------------------
// TDescriptor::PutParameterPtr:
//
// Create a descriptor from a block of data and insert it into the record
//----------------------------------------------------------------------------------------
void TDescriptor::PutParameterPtr(AEKeyword key, DescType typeCode, Ptr dataPtr, Size dataSize)
	{
	FailErr(AEPutParamPtr(*this, key, typeCode, dataPtr, dataSize));
	} // TDescriptor::PutParameterPtr 

//----------------------------------------------------------------------------------------
// TDescriptor::PutParameterHandle: 
//----------------------------------------------------------------------------------------
void TDescriptor::PutParameterHandle(AEKeyword key, DescType typeCode, Handle dataHandle)
	{
	short		handleState;
	Size		dataSize;
	
	//
	// Feature:  if the dataHandle is nil, nothing is added to the AERecord.
	//
	if(dataHandle != nil)
		{
		handleState = HGetState(dataHandle);
		HLock(dataHandle);
		dataSize = GetHandleSize(dataHandle);
		
		this->PutParameterPtr(key,typeCode,*dataHandle,dataSize);
		
		HSetState(dataHandle,handleState);
		}
	} // TDescriptor::PutParameterHandle 
	
//----------------------------------------------------------------------------------------
// TDescriptor::PutLongParameter:
//
// Put a long parameter (type specified)
//----------------------------------------------------------------------------------------
void TDescriptor::PutLongParameter(AEKeyword key, DescType typeCode, long data)
	{
	this->PutParameterPtr(key, typeCode, (Ptr) &data, sizeof(long));
	} // TDescriptor::PutLongParameter 
	
//----------------------------------------------------------------------------------------
// TDescriptor::PutLongParameter:
//
// Put a long parameter (type is 'typeLongInteger', the most common case)
//----------------------------------------------------------------------------------------
void TDescriptor::PutLongParameter(AEKeyword key, long data)
	{
	this->PutLongParameter(key,typeLongInteger,data);
	} // TDescriptor::PutLongParameter 
	
//----------------------------------------------------------------------------------------
// TDescriptor::PutShortParameter:
//
// Put a short parameter (type specified)
//----------------------------------------------------------------------------------------
void TDescriptor::PutShortParameter(AEKeyword key, DescType typeCode, short data)
	{
	this->PutParameterPtr(key, typeCode, (Ptr) &data, sizeof(short));
	} // TDescriptor::PutShortParameter 
	
//----------------------------------------------------------------------------------------
// TDescriptor::PutShortParameter:
//
// Put a short parameter (type is 'typeShortInteger', the most common case)
//----------------------------------------------------------------------------------------
void TDescriptor::PutShortParameter(AEKeyword key, short data)
	{
	this->PutShortParameter(key,typeShortInteger,data);
	} // TDescriptor::PutShortParameter 


//----------------------------------------------------------------------------------------
// TDescriptor::PutOrdinalParameter:
//----------------------------------------------------------------------------------------
void TDescriptor::PutOrdinalParameter(AEKeyword key, DescType ordinalData)
	{
	this->PutParameterPtr(key, typeAbsoluteOrdinal, (Ptr) &ordinalData, sizeof(DescType));
	} // TDescriptor::GetOrdinalParameter 

//----------------------------------------------------------------------------------------
// TDescriptor::PutResult: 
//
// Put the provided descriptor into the result parameter of this event.  If the data
// is a null descriptor, then put in an empty list instead.
//----------------------------------------------------------------------------------------
void TDescriptor::PutResult(TDescriptor data)
	{
	OSErr err = noErr;
	
	//
	// If this is a null descriptor, it means that the client
	// doesn't care about the reply, so ignore the result.
	//
	if(this->IsNullDescriptor() == false)
		{
		//
		// If there's nothing in the data, then insert an empty list.
		//
		if(data.IsNullDescriptor())
			{
			TDescriptor emptyList;
	
			Try
				{
				emptyList.MakeEmptyList();
				this->PutDescriptor(keyAEResult, emptyList);
				emptyList.Dispose();
				}
			Catch(err)
				{
				emptyList.Dispose();
				Throw(err);
				}
			}
		else
			{
			this->PutDescriptor(keyAEResult,data);
			}
		}
	} // TDescriptor::PutResult 
	
//----------------------------------------------------------------------------------------
// TDescriptor::PutErrorCode:
//
// Put an error code
//----------------------------------------------------------------------------------------
void TDescriptor::PutErrorCode(long theErr)
	{
	this->PutLongParameter(keyErrorNumber,theErr);
	} // TDescriptor::PutErrorCode 

TDescriptor gErrorDescriptor;
Boolean gInstalledErrorCallback = false;

#ifdef POWERPC
	static RoutineDescriptor gGetErrorDescCallbackRD	= BUILD_ROUTINE_DESCRIPTOR(uppOSLGetErrDescProcInfo, TDescriptor::GetErrorDesc);
#endif

//----------------------------------------------------------------------------------------
// TDescriptor::GetErrorDesc:
//
// Tell the Object support library where it can put a descriptor (an object specifier)
// to the object that caused the resolution to fail.
//----------------------------------------------------------------------------------------
pascal OSErr TDescriptor::GetErrorDesc(TDescriptor** errorDescriptor )
	{
	*errorDescriptor = &gErrorDescriptor;
	
	return noErr;
	} // TDescriptor::GetErrorDesc 

//----------------------------------------------------------------------------------------
// TDescriptor::Resolve: 
//----------------------------------------------------------------------------------------
TTokenDescriptor TDescriptor::Resolve(TTransaction* t, TDescriptor* objectThatCausedError /*= nil*/)
	{
	return TDescriptor::Resolve(t, fCallbackFlags, objectThatCausedError);
	} // TDescriptor::Resolve 

//----------------------------------------------------------------------------------------
// TDescriptor::Resolve:
//
// Resolves a list of object specifiers into a list of tokens.  A null descriptor will
// resolve to a token to the null container.
//
// n.b. AEResolve will return an error if you pass it a null descriptor; this is why
//		we include 'magic' checking for null descriptors here.
//----------------------------------------------------------------------------------------
TTokenDescriptor TDescriptor::Resolve(TTransaction* t, short callbackFlags, TDescriptor* objectThatCausedError /*= nil*/)
	{
	TTokenDescriptor resolvedToken;
	TTokenDescriptor intermediate;
	TDescriptor descriptor;
	
	long startTime = TickCount();
	long durration = 0;
	
	//
	// First:  special checking for null descriptors
	//
	if(this->IsNullDescriptor())
		{
		resolvedToken = CreateNullContainerToken();
		}
	//
	// Next, process non-lists
	//
	else if(this->DescriptorType() != typeAEList)
		{
		//
		// If a pre-resolve proc was installed, then call it.
		//
		// This is where we take over from the OSL and resolve
		// object specifiers ourselves...
		//
		if(gPreResolveProc != nil)
			{
			resolvedToken = (*gPreResolveProc)(t, *this, objectThatCausedError);
			}

		//
		// If the pre-resolve proc did not resolve the token,
		// then call AEResolve
		//
		if(resolvedToken.IsNullDescriptor())
			{
			if(gInstalledErrorCallback == false)
				{
#ifdef POWERPC
				AEInstallSpecialHandler(keyAEGetErrDescProc, (UniversalProcPtr)&gGetErrorDescCallbackRD, false);
#else
				AEInstallSpecialHandler(keyAEGetErrDescProc, (UniversalProcPtr)TDescriptor::GetErrorDesc, false);
#endif

				gInstalledErrorCallback = true;
				}
			
			OSErr resolveErr = AEResolve(*this, callbackFlags, resolvedToken);
			if((resolveErr) != (noErr && objectThatCausedError != nil))
				{
				objectThatCausedError->AdoptDesc(gErrorDescriptor);
				gErrorDescriptor.MakeNull();
				}
			else
				gErrorDescriptor.Dispose();
			FailErr(resolveErr);
			}
		}
	//
	// If we have a list, try to resolve each element of the list in turn
	//
	else
		{
		OSErr err = noErr;
		
		Try
			{
			FOREACHDESCRIPTOR(this, descriptor)
				{
				//
				// We do not want to dispose of the intermediate token descriptor
				// because its data is adopted by the 'result' token descriptor.
				//
				intermediate = descriptor.Resolve(t, callbackFlags, objectThatCausedError);
				resolvedToken.AdoptToken(intermediate);
				intermediate.MakeNull();
				}
			}
		Catch(err)
			{
			descriptor.Dispose();
			intermediate.DisposeToken();
			resolvedToken.DisposeToken();
			Throw(err);
			}
		}
	
	durration = TickCount() - startTime;
	
	return resolvedToken;
	} // TDescriptor::Resolve 


//========================================================================================
// CLASS TAEvent
//
// AppleEvents are just like AERecords, but they can contain attributes as well as
// parameters.
//========================================================================================


//----------------------------------------------------------------------------------------
// TAEvent::TAEvent: 
//----------------------------------------------------------------------------------------
TAEvent::TAEvent()
	{

	} // TAEvent::TAEvent 


//----------------------------------------------------------------------------------------
// TAEvent::MakeAppleEvent: 
//----------------------------------------------------------------------------------------
void TAEvent::MakeAppleEvent(AEEventClass eventClass, AEEventID eventID,
						const AEAddressDesc *target, short returnID, long transactionID)
	{
	FailErr(AECreateAppleEvent(eventClass, eventID,
						target, returnID, transactionID, *this));
	} // TAEvent::MakeAppleEvent 

//----------------------------------------------------------------------------------------
// TAEvent::MakeAppleEvent: 
//----------------------------------------------------------------------------------------
void TAEvent::MakeAppleEvent(AEEventClass eventClass, AEEventID eventID,
						const ProcessSerialNumber& psn, short returnID, long transactionID)
	{
	TDescriptor address;
	address.CopyData(typeProcessSerialNumber, (Ptr) &psn, sizeof(ProcessSerialNumber));
	
	this->MakeAppleEvent(eventClass, eventID, (const AEAddressDesc *) &address, returnID, transactionID);
	
	address.Dispose();
	} // TAEvent::MakeAppleEvent 


#define kCheckAEM			0
#define kDontUseSelfSend	1
#define kSelfSendSafe		2

short gCanUseSelfSend = kCheckAEM;

//----------------------------------------------------------------------------------------
// TAEvent::MakeEventAddressedToSelf: 
//----------------------------------------------------------------------------------------
void TAEvent::MakeEventAddressedToSelf(AEEventClass eventClass, AEEventID eventID, short returnID, long transactionID)
	{
	ProcessSerialNumber current;		// create psn for self
	current.highLongOfPSN = 0;
	current.lowLongOfPSN = kCurrentProcess;
	
	//
	// Old versions of the AEM leak memory if send-to-self is used
	//
	if(gCanUseSelfSend == kCheckAEM)
		{
		long aevtSelector = 0;
		OSErr err = Gestalt('evnt', &aevtSelector);
		gCanUseSelfSend = ((aevtSelector & 2) != 0) ? kSelfSendSafe : kDontUseSelfSend;
		}
	
	//
	// Don't make the self-addressed event unless it is safe to do so
	//
	if(gCanUseSelfSend == kSelfSendSafe)
		{
		this->MakeAppleEvent(eventClass, eventID, current, returnID, transactionID);
		}
	else
		FailErr(errAEEventNotHandled);
	} // TAEvent::MakeEventAddressedToSelf 

//----------------------------------------------------------------------------------------
// TAEvent::MakeEventAddressedToSystem: 
//----------------------------------------------------------------------------------------
void TAEvent::MakeEventAddressedToSystem(AEEventClass eventClass, AEEventID eventID, short returnID, long transactionID)
	{
	ProcessSerialNumber systemPSN;		// create psn for the System
	systemPSN.highLongOfPSN = 0;
	systemPSN.lowLongOfPSN = kSystemProcess;
	
	this->MakeAppleEvent(eventClass, eventID, systemPSN, returnID, transactionID);
	} // TAEvent::MakeEventAddressedToSystem 

//----------------------------------------------------------------------------------------
// TAEvent::Send: 
//----------------------------------------------------------------------------------------
void TAEvent::Send( TAEvent* reply,
					AESendMode sendMode,
					AESendPriority sendPriority,
					long timeOutInTicks,
					AEIdleUPP idleProc,
					AEFilterUPP filterProc)
	{
	OSErr err = noErr;
	
	err = AESend(*this, *reply, sendMode, sendPriority, timeOutInTicks, idleProc, filterProc);
	if((err != errAETimeout) || (timeOutInTicks != 0))
		FailErr(err);
	} // TAEvent::Send 

#define kAEDontExecute 0x2000 // for SendMode, when you just want an event recorded, not executed

//----------------------------------------------------------------------------------------
// TAEvent::SendNoExecute: 
//
// This routine sends the event with the 'kAEDontExecute' bit set, so that the
// event will be recorded, but the recipient will not execute it.  The event is
// also sent 'no reply', since it does not make sense to expect a result from
// a command that is not actually executed.
//----------------------------------------------------------------------------------------
void TAEvent::SendNoExecute()
	{
	//
	// The reply isn't actuall filled in due to the kAENoReply flag
	//
	TAEvent reply;
	Send(&reply, kAENoReply+kAEDontExecute);
	} // TAEvent::SendNoExecute 


//----------------------------------------------------------------------------------------
// TAEvent::SuspendTheCurrentEvent: 
//----------------------------------------------------------------------------------------
void TAEvent::SuspendTheCurrentEvent()
	{
	FailErr(AESuspendTheCurrentEvent(*this));
	} // TAEvent::SuspendTheCurrentEvent 

//----------------------------------------------------------------------------------------
// TAEvent::ResumeTheCurrentEvent: 
//----------------------------------------------------------------------------------------
void TAEvent::ResumeTheCurrentEvent(TAEvent* reply, AEEventHandlerUPP dispatcher, long refCon)
	{
	FailErr(AEResumeTheCurrentEvent(*this, *reply, dispatcher, refCon));
	} // TAEvent::ResumeTheCurrentEvent 

//----------------------------------------------------------------------------------------
// TAEvent::SetTheCurrentEvent: 
//----------------------------------------------------------------------------------------
void TAEvent::SetTheCurrentEvent()
	{
	FailErr(AESetTheCurrentEvent(*this));
	} // TAEvent::SetTheCurrentEvent 

//----------------------------------------------------------------------------------------
// TAEvent::ResetTimer: 
//
// Note:	requires that this event be a 'reply' event passed to one of our event
//			handlers by the AppleEvent manager 
//----------------------------------------------------------------------------------------
void TAEvent::ResetTimer()
	{
	FailErr(AEResetTimer(*this));
	} // TAEvent::ResetTimer 

//----------------------------------------------------------------------------------------
// TAEvent::SpecifyThatParameterIsOptional: 
//----------------------------------------------------------------------------------------
void TAEvent::SpecifyThatParameterIsOptional(AEKeyword theOptionalKeyword)
	{
	TDescriptor optionalKeywordsList;
	TDescriptor optionalKeyword;
	OSErr err = noErr;
	
	Try
		{
		//
		// Script:  We need to test to see if keyOptionalKeywordAttr
		// already exists, and if so, then just add one more element to
		// the list that is already inside the event.
		//
		optionalKeywordsList.MakeList();
		optionalKeyword.MakeKeyword(theOptionalKeyword);
		optionalKeywordsList.AddDescriptor(optionalKeyword);
		optionalKeyword.Dispose();
		this->PutAttribute(keyOptionalKeywordAttr, optionalKeywordsList);
		optionalKeywordsList.Dispose();
		}
	Catch(err)
		{
		optionalKeyword.Dispose();
		optionalKeywordsList.Dispose();
		
		Throw(err);
		}
	} // TAEvent::SpecifyThatParameterIsOptional 

//----------------------------------------------------------------------------------------
// TAEvent::GetAttribute: 
//----------------------------------------------------------------------------------------
TDescriptor TAEvent::GetAttribute(AEKeyword key, DescType desiredType /* = typeWildCard */)
	{
	TDescriptor result;
	
	FailErr(AEGetAttributeDesc(*this, key, desiredType, result));
	
	return result;
	} // TAEvent::GetAttribute 

//----------------------------------------------------------------------------------------
// TAEvent::GetOptionalAttribute: 
//----------------------------------------------------------------------------------------
TDescriptor TAEvent::GetOptionalAttribute(AEKeyword key, DescType desiredType /* = typeWildCard */)
	{
	TDescriptor result;
	
	OSErr err = AEGetAttributeDesc(*this, key, desiredType, result);
	if(err != noErr)
		result.MakeNull();
	
	return result;
	} // TAEvent::GetOptionalAttribute 

//----------------------------------------------------------------------------------------
// TAEvent::GetLongAttribute: 
//----------------------------------------------------------------------------------------
long TAEvent::GetLongAttribute(AEKeyword key, DescType desiredType /* = typeLongInteger */)
	{
	DescType typeCode;
	long actualSize;
	long result;
	
	FailErr(AEGetAttributePtr(*this, key, desiredType, &typeCode, (Ptr)&result, sizeof(long), &actualSize));
	
	return result;
	} // TAEvent::GetLongAttribute 

//----------------------------------------------------------------------------------------
// TAEvent::PutAttribute: 
//----------------------------------------------------------------------------------------
void TAEvent::PutAttribute(AEKeyword key, TDescriptor attribute)
	{
	FailErr(AEPutAttributeDesc(*this, key, attribute));
	} // TAEvent::PutAttribute 

//----------------------------------------------------------------------------------------
// PutOptionalDescriptor: 
//----------------------------------------------------------------------------------------
void TAEvent::PutOptionalDescriptor(AEKeyword key, TDescriptor data)
	{
	this->PutDescriptor(key, data);
	this->SpecifyThatParameterIsOptional(key);
	} // TAEvent::PutOptionalDescriptor 

//----------------------------------------------------------------------------------------
// TAEvent::PutLongAttribute: 
//----------------------------------------------------------------------------------------
void TAEvent::PutLongAttribute(AEKeyword key, long attributeValue)
	{
	FailErr(AEPutAttributePtr(*this, key, typeLongInteger, (Ptr)&attributeValue, sizeof(long)));
	} // TAEvent::PutLongAttribute 
	

/*************
		
		// AppleEvents have attributes, but AERecords do not
		// So far we have not had to extract attributes from AppleEvents,
		// so we have not implimented these routines yet.
		
		// get the attribute from the event
		OSErr TAEvent::GetAttribute(AEKeyword key, DescType desiredType, TDescriptor *result)
			{
			return AEGetAttributeDesc(*this, key, desiredType, result);
			}
			
		// get a pointer to the attribute from the event
		OSErr TAEvent::GetAttributePtr(AEKeyword key, DescType desiredType, DescType *typeCode, Ptr dataPtr, Size maximumSize, Size *actualSize)
			{
			return AEGetAttributePtr(*this, key, desiredType, typeCode, dataPtr, maximumSize, actualSize);
			}
			
		// get the size of the attribute
		OSErr TAEvent::SizeOfAttribute(AEKeyword key, DescType desiredType, Size *size)
			{
			return AESizeOfAttribute(*this, key, desiredType, size)
			};
		
		// put the ae attribute from the event
		OSErr TAEvent::PutAttribute(AEKeyword key, DescType desiredType, const TDescriptor *descriptor)
			{
			return AEPutAttributeDesc(*this, key, desiredType, descriptor);
			}
			
		// put a pointer to the attribute from the event
		OSErr TAEvent::PutAttributePtr(AEKeyword key, DescType desiredType, Ptr dataPtr, Size dataSize)
			{
			return AEPutAttributePtr(*this, key, desiredType, dataPtr, dataSize);
			}
		
		// get and set parameters
		
		// get the ae descriptor from the event
		OSErr TAEvent::GetParameterPtr(AEKeyword key, DescType desiredType, DescType *typeCode, Ptr dataPtr, Size maximumSize, Size *actualSize)
			{
			return AEGetParamPtr(*this, key, desiredType, typeCode, maximumSize, actualSize);
			}
	
*************/


//========================================================================================
// CLASS TTokenDescriptor
//
// The class TTokenDescriptor represents descriptors returned by AEResolve
//========================================================================================


//----------------------------------------------------------------------------------------
// TTokenDescriptor::TTokenDescriptor: 
//----------------------------------------------------------------------------------------
TTokenDescriptor::TTokenDescriptor()
	{
	fDescriptorType = typeNull;
	fDataHandle = nil;
	} // TTokenDescriptor::TTokenDescriptor 

//----------------------------------------------------------------------------------------
// TTokenDescriptor::TTokenDescriptor: 
//----------------------------------------------------------------------------------------
TTokenDescriptor::TTokenDescriptor(TDescriptor desc)
	{
	this->fDescriptorType = desc.DescriptorType();
	this->fDataHandle = desc.DataHandle();
	} // TTokenDescriptor::TTokenDescriptor 
	
//----------------------------------------------------------------------------------------
// TTokenDescriptor::DisposeToken:
//
// It is very important that token descriptors be disposed of by AEDisposeToken,
// _not_ AEDisposeDesc.
//----------------------------------------------------------------------------------------
void TTokenDescriptor::DisposeToken()
	{
	AEDisposeToken(*this);  // Should fail on error?  Probably not.

	// FailErr(AEDisposeDesc(*this));
	} // TTokenDescriptor::DisposeToken 


//----------------------------------------------------------------------------------------
// TTokenDescriptor::TokenObject:
//
// Return the object stored inside the token descriptor
//----------------------------------------------------------------------------------------
TAbstractScriptableObject* TTokenDescriptor::TokenObject()
	{
	Handle tokenHandle = this->DataHandle();
	TAbstractScriptableObject* tokenObject = nil;
	
	if(this->DescriptorType() == typeTokenObject)
		{
		tokenObject = (TAbstractScriptableObject*) tokenHandle;
		}
	else if(this->DescriptorType() == typeTokenInHandle)
		{
		//
		// tokenHandle is a handle that contains a TTokenObject*,
		// so it is a pointer to a pointer to a TTokenObject* (wow).
		// This code is never used if your application supports marking.
		//
		tokenObject = ** ((TAbstractScriptableObject***) tokenHandle);
		}
	else if((this->DescriptorType() == typeAEList) && (gMergeTokensProc != nil))
		{
		TDescriptor descriptor;

		//
		// This will only happen if the OSL merges tokens into
		// an AEList for you, which will only happen if your
		// application does not support marking.
		//
		FOREACHDESCRIPTOR(this, descriptor)
			{
			TAbstractScriptableObject* intermediateObject = ((TTokenDescriptor*)&descriptor)->TokenObject();
			if(tokenObject == nil)
				tokenObject = intermediateObject;
			else
				tokenObject = (*gMergeTokensProc)(tokenObject, intermediateObject);
			}
		}
		
	if(tokenObject == nil)
		{
		//
		//  What error code should we fail with here?
		//
		FailErr(errAENoSuchObject);
		}
	
	return tokenObject;
	} // TTokenDescriptor::TokenObject 


//----------------------------------------------------------------------------------------
// TTokenDescriptor::AdoptToken:
//
// IMPORTANT: ADOPTS THE TOKEN PASSED TO IT
//----------------------------------------------------------------------------------------
void TTokenDescriptor::AdoptToken(TTokenDescriptor& tokenDescriptor)
	{
	//
	// Don't adopt the new token if it is empty
	//
	if(tokenDescriptor.IsNullDescriptor() == false)
		{
		//
		// If this descriptor is empty, it is easy to adopt the new token
		//
		if( this->IsNullDescriptor() )
			{
			TDescriptor::AdoptHandle(tokenDescriptor.DescriptorType(), tokenDescriptor.DataHandle() );
			tokenDescriptor.MakeNull();
			}
		//
		// If the token being adopted is a list, then append it to us
		// (we really want to avoid making AELists of AELists).
		//
		else if(tokenDescriptor.DescriptorType() == typeAEList)
			{
			//
			// We don't expect to get here any more
			//
			ASSERTPRINT(false, ("Got an AEList in TTokenDescriptor::AdoptToken!"));
			}
		else
			{
			TAbstractScriptableObject* tokenObject = tokenDescriptor.TokenObject();
			this->AdoptToken(tokenObject);
			
			//
			// n.b.	'Dispose', not 'DisposeToken'.  The token that
			// was formerly contained in tokenDescriptor is now referenced
			// in 'this' token descriptor, so it would be a bad idea to
			// delete it.
			//
			tokenDescriptor.Dispose();
			}
		}
	} // TTokenDescriptor::AdoptToken 

Boolean gCheatAndInlineToken = true;

//----------------------------------------------------------------------------------------
// TTokenDescriptor::AdoptToken: 
//----------------------------------------------------------------------------------------
void TTokenDescriptor::AdoptToken(TAbstractScriptableObject* tokenObject)
	{
	//
	// Don't adopt the new token if it is not a valid object
	//
	if(tokenObject != nil)
		{
		//
		// If this descriptor is empty, then put the token object
		// into this token.  If marking is supported, then the
		// token pointer may be placed into the token directly; if
		// marking is not supported, then we must create a handle
		// to hold the token.
		//
		if(this->IsNullDescriptor() )
			{
			// if((fCallbackFlags & kAEIDoMarking) != 0)
			if(gCheatAndInlineToken)
				this->AdoptHandle(typeTokenObject, (Handle)tokenObject);
			else
				{
				this->CopyData(typeTokenInHandle, (Ptr)&tokenObject, sizeof(Ptr));
				}
			}
		//
		// If there is already something here, then we need
		// to somehow merge the two tokens together.
		// Usually, we want to create a mark token
		// to do the merge.  However, MoreAEM does not depend
		// on files such as MarkToken.h which are outside of
		// the Blue folder.  Therefore, we call a callback proc
		// to fill in the mark token.  The callback proc is
		// installed in InitializeScriptability.
		//
		else
			{
			if(gMergeTokensProc == nil)
				FailErr(errAEEventNotHandled);
				
			//
			// The MergeTokensProc takes two TAbstractScriptableObject*'s
			// and merges them into another TAbstractScriptableObject*
			// (which adopts the two original tokens).
			//
			TAbstractScriptableObject* mergedTokens = (*gMergeTokensProc)(this->TokenObject(), tokenObject);
			this->MakeNull();
			this->AdoptToken(mergedTokens);
			}
		}
	} // TTokenDescriptor::AdoptToken 


#pragma segment CFrontCruft
