/*
	File:		MoreAEM.h

	Contains:	C++ wrappers for the AppleEvent Manager

	Written by:	Bruce Horn, Steve Capps, Larry Kenyon,
				John Meier, scott douglass, Darin Adler,
				Paul Mercer, Bryan Stearns, Dave Owens,
				Francis Stanbach

	Copyright:	 1992, 1994-1995 by Apple Computer, Inc., all rights reserved.

		<23>	  2/9/95	jtr		

*/

#ifndef MoreAEM_h
#define MoreAEM_h

#ifndef __APPLEEVENTS__					// Needed for AEDesc & c.
#include <AppleEvents.h>
#endif

//
// for TConstDataReference
//
#include "AbstractData.h"

#include "Int64.h"

#if GENERATINGPOWERPC
#pragma options align=mac68k
#endif

enum
{
	typeTokenObject			= 'tkob',
	typeTokenInHandle		= 'thnd',
	typeDateTimeRec			= 'dtim',
	typeLongDateTimeRec		= 'ldtr'
};


//
// We need to define these comparison operators
// in order to allow us to reverse the order of
// any two descriptors being compared
//
enum
	{
	kAEAtTheBeginningOf		= 'atbg',
	kAEAtTheEndOf			= 'aten',
	kAEIsContainedIn		= 'isin'
	};
	
// Use with CreateList:
#define kMakeAEList		false
#define kMakeAERecord	true

//
// The class name TTransaction is only used by Resolve
//
class TTransaction;

// Some procedural routines

DescType ReverseComparisonOperator(DescType comparisonOperator);
Boolean InterpretCompareResult(DescType comparisonOperator, long key1, long key2 );
// Boolean GeneralCompare(DescType comparisonOperator, Ptr data, Size length, Ptr compareWith, Size comparisonLength, Boolean caseSensitive);
Boolean CompareTypedData(DescType comparisonOperator, const TConstDataReference& compareThisRef, const TConstDataReference& compareWithRef);


class TDescriptor;
class TTokenDescriptor;

//
// Used with CoerceInPlace
//
enum CoerceInPlaceTechnique
	{
	kPreserveHandle,
	kAllowHandleReallocation
	};

class TDescriptor
	{	
protected:
	//
	// A TDescriptor is an AEDesc, so it must have
	// two and only two fields, and they must be
	// the descriptor type and the data handle.
	//
	DescType			fDescriptorType;
	Handle				fDataHandle;
	
	//
	// These flags are passed to AEResolve when
	// specific flags are not specified.  This
	// allows an application to define kAEIDoWhose
	// and/or kAEIDoMarking once at startup, and
	// subsequentially call AEResolve without
	// any parameters elsewhere in the code.
	//
	static short		fCallbackFlags;
	
public:
	static void			SetCallbackFlags(short callbackFlags)	{ fCallbackFlags = callbackFlags; };
	
	//
	// Construction methods
	//
						TDescriptor() : fDescriptorType(typeNull), fDataHandle(nil) {};
						TDescriptor(DescType type, Handle dataHandle) : fDescriptorType(type), fDataHandle(dataHandle) {};
						TDescriptor(TDescriptor& desc) : fDescriptorType(desc.DescriptorType()), fDataHandle(desc.DataHandle()) {};
						TDescriptor(AEDesc& desc) : fDescriptorType(desc.descriptorType), fDataHandle(desc.dataHandle) {};
						TDescriptor(AEDesc* desc) : fDescriptorType(desc->descriptorType), fDataHandle(desc->dataHandle) {};
	
	//
	// Conversion operators
	//
	operator			const AEDesc*()	const					{ return (const AEDesc*) this; };
	operator			AEDesc*()								{ return (AEDesc*) this; };
	operator			AEDesc()								{ return *((AEDesc*) this); };

	operator			const TConstDataReference() const		{ return fDataHandle ? TConstDataReference(fDescriptorType, *fDataHandle, GetHandleSize(fDataHandle)) : TConstDataReference(fDescriptorType, nil, 0); }

	//
	// Dispose
	//
	void				Dispose();
	
	//
	// Direct access to descriptor information
	//
	DescType			DescriptorType() const					{ return fDescriptorType; };
	Handle				DataHandle()							{ return fDataHandle; };
	Boolean				IsNullDescriptor() const				{ return fDescriptorType == typeNull; };
	Boolean				DataHandleIsNil() const					{ return fDataHandle == nil; };
	
	short				Lock() const;
	void				Unlock(short oldState) const;

	//
	// Content manipulation
	//
	OSErr				AttemptCoerceInPlace(DescType typeToCoerceTo, CoerceInPlaceTechnique preserveHandle = kPreserveHandle);
	void				CoerceInPlace(DescType typeToCoerceTo, CoerceInPlaceTechnique preserveHandle = kPreserveHandle);
	TDescriptor			AttemptToCoerce(DescType typeToCoerceTo, OSErr& err) const;
	TDescriptor			Coerce(DescType typeToCoerceTo) const;
	TDescriptor			CoerceToStandardType();
	TDescriptor			Clone() const;
	void				CopyDesc(const TDescriptor& desc);
	void				AdoptDesc(TDescriptor& desc);
	void				AdoptHandle(DescType dataType, Handle dataHandle);
	void				CopyData(const DescType typeCode, const Ptr data, const Size length);

	void				CreateList(Boolean isRecord = kMakeAEList, Ptr factoringPtr = nil, Size factoredSize = 0);
	void				MakeNull();
	
	//
	// Descriptor creation
	//
	void				Make(const TConstDataReference& data);
	void				MakeBoolean(const Boolean data);
	void				MakeLong(const long data);
	void				MakeShort(const short data);
	void				MakeInt64(Int64& data);
	void				MakePoint(const Point& thePoint);
	void				MakePoint(const short h, const short v);
	void				MakeRect(const Rect& theRect);
	void				MakeUnsignedLong(const unsigned long data);
	void				MakeEnumeration(const DescType enumeration);
	void				MakeDescType(const DescType data);
	void				MakeKeyword(const AEKeyword theKeyword);
	void				MakeOrdinal(const DescType data);
	void				MakeTypeOrInteger(const DescType data);
	void				MakeString(ConstStr255Param data);
	void				MakeString(const char* data);
	void				MakeDateTimeRec(const DateTimeRec dateTime);
	void				MakeDateTimeRec(const long dateTime);
	void				MakeLongDateTimeRec(const LongDateRec dateTime);
	void				MakeLongDateTimeRec(const DateTimeRec dateTime);
//	void				MakeLongDateTimeRec(const long secsSince1904);	// Confuses CFront
	void				MakeLongDateTimeRec(LongDateTime lsecs);
	void				MakeLongDateTime(const LongDateTime lsecs);
//	void				MakeRelativeTime(const long relativeTime);
//	void				MakeConceptualTime(const long relativeTime);
	void				MakeFSS(const FSSpec& spec);
	void				AdoptAlias(Handle alias);
	void				MakeAlias(FSSpec& spec);
	void				MakeProcessSerialNumber(ProcessSerialNumber psn);
	
	void				MakeObjectSpecifier(DescType desiredClass, TDescriptor container, DescType keyForm, TDescriptor keyData, Boolean disposeInputs);
	void				MakeObjectSpecifierForProperty(DescType property, TDescriptor container, Boolean disposeInputs);
	void				MakeSpecifierForPropertyOfObjectBeingExamined(DescType propertyIdentifier);
	void				MakeCompDescriptor(DescType comparisonOperator, TDescriptor comparitor, TDescriptor compareWith, Boolean disposeInputs);
	void				MakeCompDescriptor(DescType comparisonOperator, DescType propertyIdentifier, TDescriptor compareWith, Boolean disposeInputs);
	
	//
	// Data extraction
	//
	void				GetBlock(Ptr data, Size length, DescType desiredType) const;
	short				GetShort() const;
	long				GetLong() const;
	Int64				GetInt64() const;
	Boolean				GetBoolean() const;
	DescType			GetDescType() const;
	AEKeyword			GetKeyword() const;
	DescType			GetEnumeration() const;
	DescType			GetOrdinal() const;
	Point				GetPoint() const;
	Rect				GetRect() const;
	void				GetString(Str255&) const;
	void				GetString(unsigned char* result, short maxLength) const;
	void				GetString(char*) const;
	DateTimeRec			GetDateTimeRec() const;
//	long				GetRelativeTime() const;
//	long				GetConceptualTime() const;
	void				GetFSS(FSSpec& spec) const;

//	Boolean				Compare(DescType comparisonOperator, Str255 withString) const;
	Boolean				Compare(DescType comparisonOperator, TDescriptor& compareWith) const;
//	Boolean				CompareRelativeTime(DescType comparisonOperator, TDescriptor& compareWith) const;
	
	//
	// List methods
	//
	// Formerly, we had a separate class, TDescriptorList.
	// Now, however, there is no difference between a TDescriptor
	// and a TDescriptorList
	//
	void				MakeEmptyList();
	void				MakeList();
	
	long				CountItems() const;
	TDescriptor			GetNthDescriptor(long index, DescType desiredType = typeWildCard, AEKeyword* key = nil) const;

	void				AddDescriptor(long index, TDescriptor& data);
	void				AddDescriptor(TDescriptor& data);

	void				AddData(long index, DescType descType, Ptr data, Size length);
	void				AddData(DescType descType, Ptr data, Size length);
	void				AddLong(long number);
	void				AddType(DescType descType);

	void				AppendList(const TDescriptor& list);
	void				AdoptList(TDescriptor* list);

	//
	// AERecord methods
	//
	// Formerly, we had a separate class, TAERecord.
	// Now, however, AERecords and AEDescriptors are
	// used interchangably enough that there is really
	// no reason to have a separate class.
	//
	void				MakeAERecord();

	//
	// Conversion opperators
	//
	// We can't have multiple conversion operators
	// (AERecord and AEDesc) because AERecord is
	// typedef'ed to be AEDesc.
	//
#ifdef DUPLICATE
	operator const AERecord*() const		{ return (const AERecord*) this; };
	operator AERecord*()					{ return (AERecord*) this; };
	operator AERecord()						{ return *((AERecord*) this); };
#endif

	TDescriptor			GetDescriptor(AEKeyword key, DescType desiredType = typeWildCard) const;	
	TDescriptor			GetOptionalParameter(AEKeyword key, DescType desiredType = typeWildCard) const;	
	void				GetParameterPtr(AEKeyword key, DescType desiredType, DescType *typeCode, Ptr dataPtr, Size maximumSize, Size *actualSize) const;
	long				GetLongParameter(AEKeyword key, DescType desiredType = typeLongInteger) const;	
	Int64				GetInt64Parameter(AEKeyword key) const;
	Boolean				GetBooleanParameter(AEKeyword key) const;
	DescType			GetDescTypeParameter(AEKeyword key) const;
	AEKeyword			GetKeywordParameter(AEKeyword key) const;
	DescType			GetEnumerationParameter(AEKeyword key) const;
	DescType			GetOrdinalParameter(AEKeyword key) const;
	Point				GetPointParameter(AEKeyword key) const;
	Rect				GetRectParameter(AEKeyword key) const;

	void				GetObjectSpecifierParameters(DescType& desiredClass, DescType& keyForm, TDescriptor* keyData = nil, TDescriptor* containerDesc = nil) const;
	
	TDescriptor GetDirectObject() const;
	long GetErrorCode() const;

	void				PutDescriptor(AEKeyword key, TDescriptor data);
	void				PutParameterPtr(AEKeyword key, DescType typeCode, Ptr dataPtr, Size dataSize);
	void				PutParameterHandle(AEKeyword key, DescType typeCode, Handle dataHandle);
	void				PutLongParameter(AEKeyword key, DescType typeCode, long data);	
	void				PutLongParameter(AEKeyword key, long data);	
	void				PutShortParameter(AEKeyword key, DescType typeCode, short data);	
	void				PutShortParameter(AEKeyword key, short data);	
	void				PutOrdinalParameter(AEKeyword key, DescType ordinalData);

	void PutResult(TDescriptor data);
	void PutErrorCode(long theErr);
	
	//
	// Object support library methods
	//
	static pascal OSErr	GetErrorDesc(TDescriptor** errorDescriptor);
	TTokenDescriptor	Resolve(TTransaction*, TDescriptor* objectThatCausedError = nil);	
	TTokenDescriptor	Resolve(TTransaction*, short callbackFlags, TDescriptor* objectThatCausedError = nil);	
	};



//
// A descriptor loop iterates over every item in a descriptor
// list or an AERecord.
//
class TDescriptorLoop
	{
public:
						TDescriptorLoop(const TDescriptor* dl) {fDescriptorList = dl; fIndex = 0; fCount = 0; };
	Boolean				Next(TDescriptor& d, AEKeyword* key = nil, Handle* h = nil);

private:
	const TDescriptor*		fDescriptorList;
	long					fIndex;
	long					fCount;
	};

#define FOREACHDESCRIPTOR(listToIterate, descriptor)			\
	TDescriptorLoop loop(listToIterate);						\
	AEKeyword keyword;											\
	while (loop.Next(descriptor, &keyword, nil))





class TAEvent : public TDescriptor
	{
public:

	TAEvent();
	
	//
	// Conversion opperators
	//
	operator const AppleEvent*() const		{ return (const AppleEvent*) this; };
	operator AppleEvent*()					{ return (AppleEvent*) this; };
	operator AppleEvent()					{ return *((AppleEvent*) this); };
			
	void MakeAppleEvent(AEEventClass eventClass, AEEventID eventID,
				const AEAddressDesc *target, short returnID = kAutoGenerateReturnID, long transactionID = kAnyTransactionID);
	void MakeAppleEvent(AEEventClass eventClass, AEEventID eventID,
				const ProcessSerialNumber& psn, short returnID = kAutoGenerateReturnID, long transactionID = kAnyTransactionID);
	
	void MakeEventAddressedToSelf(AEEventClass eventClass, AEEventID eventID, short returnID = kAutoGenerateReturnID, long transactionID = kAnyTransactionID);
	void MakeEventAddressedToSystem(AEEventClass eventClass, AEEventID eventID, short returnID = kAutoGenerateReturnID, long transactionID = kAnyTransactionID);
	
	void Send(TAEvent* reply, AESendMode sendMode, AESendPriority sendPriority = kAENormalPriority,
			long timeOutInTicks = kAEDefaultTimeout, AEIdleUPP idleProc = nil, AEFilterUPP filterProc = nil);

	void Send(TAEvent* future)	{ this->Send(future, kAEWaitReply, kAENormalPriority, 0); };
	void SendNoReply()			{ TAEvent reply; this->Send(&reply, kAENoReply); }
	void SendNoExecute();
	
	// misc

	void SuspendTheCurrentEvent();
	void ResumeTheCurrentEvent(TAEvent* reply, AEEventHandlerUPP dispatcher = nil, long refCon = 0);
	void SetTheCurrentEvent();

	void ResetTimer();
		
	void SpecifyThatParameterIsOptional(AEKeyword theOptionalKeyword);

	// get and set attributes for the Apple event

	TDescriptor GetAttribute(AEKeyword key, DescType desiredType = typeWildCard);
	TDescriptor GetOptionalAttribute(AEKeyword key, DescType desiredType = typeWildCard);
	long GetLongAttribute(AEKeyword key, DescType desiredType = typeLongInteger);
	long GetReturnID() { return this->GetLongAttribute(keyReturnIDAttr); };
	long GetMessageTimeout() { return this->GetLongAttribute(keyTimeoutAttr); };
	
	void PutAttribute(AEKeyword key, TDescriptor attribute);
	void PutOptionalDescriptor(AEKeyword key, TDescriptor data);
	void PutLongAttribute(AEKeyword key, long attributeValue);

	/*
			
		OSErr GetAttribute(AEKeyword key, DescType desiredType, TDescriptor* descriptor);
		OSErr GetAttributePtr(AEKeyword key, DescType desiredType, DescType *typeCode,
								Ptr dataPtr, Size maximumSize, Size *actualSize);		
		OSErr SizeOfAttribute(AEKeyword key, DescType desiredType, Size *size);	
		OSErr PutAttributePtr(AEKeyword key, DescType desiredType, Ptr dataPtr, Size dataSize);	
		
	*/
	};


//
// MoreAEM doesn't know anything about TAbstractScriptableObject
// other than the fact that a class with that name exists.
//
class TAbstractScriptableObject;

typedef TTokenDescriptor (*MakeTokenDescriptorProcPtr)();
typedef TTokenDescriptor (*ProcessDescriptorProcPtr)(TTransaction*, TDescriptor descToResolve, TDescriptor* objectThatCausedError);
typedef TAbstractScriptableObject* (*MergeTokensProcPtr)(TAbstractScriptableObject* baseToken, TAbstractScriptableObject* mergeToken);

void InstallNullContainerCreationProc(MakeTokenDescriptorProcPtr creationProc);
void InstallPreResolveProc(ProcessDescriptorProcPtr preResolveProc);
void InstallMergeTokensProc(MergeTokensProcPtr mergeTokensProc);

TTokenDescriptor CreateNullContainerToken();



class TTokenDescriptor : public TDescriptor
	{
public:
					TTokenDescriptor();
					TTokenDescriptor(TDescriptor desc);

	// destructor & disposer
	void			DisposeToken();

	TAbstractScriptableObject*	TokenObject();

	void			AdoptToken(TTokenDescriptor& tokenDescriptor);
	void			AdoptToken(TAbstractScriptableObject* tokenObject);
	
	};


//
// Prototypes
//
Boolean	StringContains(Str255 thisString, Str255 withString);

#define FOREACHTOKEN(tokenList, tokenDescriptor)			\
	TDescriptorLoop loop(tokenList);						\
	TAbstractScriptableObject* token;									\
	while (loop.Next(tokenDescriptor, nil, nil) && ((token = tokenDescriptor.TokenObject()) != nil))

#if GENERATINGPOWERPC
#pragma options align=reset
#endif


#endif