/*
	File:		AbstractScriptableObject.h

	Contains:	Abstract base class that supports scriptability in objects

	Written by:	Andy Nicholas, Greg Anderson, Tom Conrad, Chris Bingham, Georgiann Puckett

	Copyright:	 1994-1995 by Apple Computer, Inc., all rights reserved.

		<20>	 9/26/95	pco		

*/

#ifndef __ABSTRACTSCRIPTABLEOBJECT__
#define __ABSTRACTSCRIPTABLEOBJECT__
#define AbstractScriptableObject_h

//
// MoreAEM.h is needed because methods defined in this
// file take TDescriptors as pass-by-value parameters,
// which requires the complete definition of the class
// TDescriptor.
// 
#include "MoreAEM.h"

//
// Object.h is needed because TObject is the base class of
// TAbstractScriptableObject, and for the DeclareClassData macros.
//
#include "Object.h"

//
// TAbstractSearchSpec represents a specification for a
// test to be done in a search (e.g. during whose clause resolution).
// TSearchSpecList is, of course, a list of TAbstractSearchSpecs.
//
class TAbstractSearchSpec;
class TSearchSpecList;

//
// Abstract references are 'abreviated' representations of some
// other scriptable object
//
class TAbstractReference;

//
// An abstract object iterator knows how to iterate over some
// list of TAbstractScriptableObjects.
//
class TAbstractObjectIterator;

//
// A list of TAbstractScriptableObjects
//
class TScriptableObjectList;

//
// An object that represents data to compare; could be literal
// data, or a description of a property of some object.
//
class TComparisonOperand;


#define clScriptableObject		2
#define clDesignator			3
#define clProperty				4
#define clGenericProperty		5
#define clProxyToken			6
#define clMarkToken				7
#define clEveryItem				8
#define clEntireContents		9
// #define clCollectorBehavior		11
#define clSearchSpec			12
#define clLogicalSearchSpec		13
#define clGenericSearchSpec		14
#define clSearchEngine			15

#define kObjectRecordedClass true

//
// There are three types of mark tokens:
//
//		kSingleItemOrCollection - example: the selection--but the selection is not
//									a mark token, so this mark type is never used.
//									Returns either a single item or a collection of
//									items.  Operations on the elements of the mark
//									operate on the items in the collection.
//
//		kSingleItemOrUnion		- example:  access by name.  Returns either a single
//									item or a union set of items.  Operations on
//									the elements of the mark operate on the union of
//									every element of every item in the set
//
//		kAlwaysMakeCollection	- example:  whose clauses.  Always returns a collection
//									of items, even if there is only one item in the
//									collection.  This is done so that the result of
//									a whose clause is always a list, even if only one
//									item is returned.
//
enum TypeOfMarkToken
	{
	kSingleItemOrUnion,
	kSingleItemOrCollection,
	kAlwaysMakeCollection
	};


//
// The magic value kNeverCreateGenericProperty should be placed in
// property tables in any instance where a generic property should
// never be created.  Usually this is not needed, as the correct
// property token should be created in AccessByProperty; however, in
// some cases a property is constructed from the property table
// without calling AccessByProperty (for example, to make a property
// of a proxy token); in that case, this value will prevent the incorrect
// type of property from being created.
//
// The magic value kDontIncludeInPropertiesProperty should be placed
// in the property tables in any instance where a property is included
// for definition purposes, but said property should not be part of
// the "properties" property.
//
// These special values may be ORed together.  The low shortword of
// the additional info is still usable by clients if the special values
// are used.  And with 'kReservedRangeUsableByClients' to mask out
// the special range when using the additional info if this is done.
// If the value stored in the additional info field is less than
// 0x7F000000, then the entire longword is usable by the client.
//
enum
	{
	kReservedRangeForPropertyInfo		= 0x7F000000,			// Values in the property additional info >= this value have special meanings
	kReservedRangeUsableByClients		= 0x0000FFFF,			// When the reserved range is used, clients may define the short word of this property to be whatever they wish
	kNeverCreateGenericBit				= 0x00800000,
	kDontIncludeInPropertiesPropertyBit	= 0x00400000,
	kBestTypeIsSameAsDefaultTypeBit		= 0x00200000,			// Set this bit to make BestType == DefaultType; the BestType field may then be used for something else
	kNeverCreateGenericProperty			= (kReservedRangeForPropertyInfo | kNeverCreateGenericBit),
	kDontIncludeInPropertiesProperty	= (kReservedRangeForPropertyInfo | kDontIncludeInPropertiesPropertyBit),
	kBestTypeIsSameAsDefaultType		= (kReservedRangeForPropertyInfo | kBestTypeIsSameAsDefaultTypeBit),
	
	kLookupAdditionalInfo				= 0xFFFFFFFF			// This value is used internally only; don't consider it part of the API
	};


//========================================================================================
//	Inline functions used by TPropertyDescription, below.
//========================================================================================


//----------------------------------------------------------------------------------------
// NeverCreateGenericProperty
//----------------------------------------------------------------------------------------
inline Boolean NeverCreateGenericProperty(unsigned long x)
	{
	return (x >= kReservedRangeForPropertyInfo) && ((x & kNeverCreateGenericBit) != 0);
	}

//----------------------------------------------------------------------------------------
// DontIncludeInPropertiesProperty
//----------------------------------------------------------------------------------------
inline Boolean DontIncludeInPropertiesProperty(unsigned long x)
	{
	return (x >= kReservedRangeForPropertyInfo) && ((x & kDontIncludeInPropertiesProperty) != 0);
	}

//----------------------------------------------------------------------------------------
// BestTypeIsSameAsDefaultType
//----------------------------------------------------------------------------------------
inline Boolean BestTypeIsSameAsDefaultType(unsigned long x)
	{
	return (x >= kReservedRangeForPropertyInfo) && ((x & kBestTypeIsSameAsDefaultType) != 0);
	}


//========================================================================================
// CLASS TPropertyDescription
//========================================================================================

class TPropertyDescription
	{
public:
	unsigned long			fPropertyIdentifier;
	unsigned long			fAdditionalInfo;
	unsigned long			fDefaultType;
	unsigned long			fBestType;

public:
	unsigned long			PropertyIdentifier() const	{ return fPropertyIdentifier; }
	unsigned long			AdditionalInfo() const		{ return fAdditionalInfo; }
	unsigned long			DefaultType() const			{ return fDefaultType; }
	unsigned long			BestType() const			{ return BestTypeIsSameAsDefaultType(fAdditionalInfo) ? fDefaultType : fBestType; }
	unsigned long			ExtendedInfo() const		{ return BestTypeIsSameAsDefaultType(fAdditionalInfo) ? fBestType : 0; }
	
	Boolean operator==(const TPropertyDescription& rhs)
		{
			return	((this->PropertyIdentifier() == rhs.PropertyIdentifier()) &&
					(this->AdditionalInfo() == rhs.AdditionalInfo()) &&
					(this->DefaultType() == rhs.DefaultType()) &&
					(this->BestType() == rhs.BestType()));
		}
	Boolean operator!=(const TPropertyDescription& rhs) { return !this->operator==(rhs); }
	};

//
// New declare / implement macros:
//
#define DeclareClassData(className, superclass)																		\
				static		TPropertyDescription  	fPropertiesOfClass[];											\
				static		TClassData  			fClassData;														\
				typedef		superclass				Inherited;														\
				virtual 	const TClassData*		ClassData() const;

#ifdef DEBUG

#define ImplementClassData(className, classID)																		\
		const TClassData* className::ClassData() const { return &fClassData; } 		\
		TClassData className::fClassData = {#className, 0, classID, sizeof(className), &className::Inherited::fClassData, {sizeof(className::fPropertiesOfClass) / sizeof(TPropertyDescription), className::fPropertiesOfClass} };

#else

#define ImplementClassData(className, classID)																		\
		const TClassData* className::ClassData() const { return &fClassData; } 		\
		TClassData className::fClassData = {0, classID, sizeof(className), &className::Inherited::fClassData, {sizeof(className::fPropertiesOfClass) / sizeof(TPropertyDescription), className::fPropertiesOfClass} };

#endif


class TTransaction;
class TAbstractCollector;

//========================================================================================
//	CLASS TAbstractScriptableObject
//========================================================================================

class TAbstractScriptableObject : public TObject
	{
public:
	DeclareClassData(TAbstractScriptableObject, TObject);

										TAbstractScriptableObject();
	virtual								~TAbstractScriptableObject();
	void								IAbstractScriptableObject();

	// virtual TAbstractScriptableObject&	operator=(const TAbstractScriptableObject& rhs);
	virtual Boolean						IsDesignator();
	virtual void						DisposeDesignator();
	virtual TAbstractScriptableObject*	CloneDesignator();
	
	// Methods that determine class membership
	
	virtual DescType					ObjectClass(TTransaction* transaction, Boolean recordedClass = false); // make this pure-virutal later = 0;
	virtual Boolean						DerivedFromOSLClass(TTransaction* transaction, DescType objectClass);

	// 'Exists' is almost always true, but proxies may need to override & return 'false' in some instances
	// (for example, a token for an empty selection, or a token for a closed window would both return 'false')
	
	virtual Boolean						Exists(TTransaction* transaction);

	// Accessors:
	
	virtual TAbstractScriptableObject*	ParentObject(TTransaction* transaction);
	virtual TAbstractObjectIterator*	ElementIterator(TTransaction*);
	
	long								CountElements(TTransaction* transaction, DescType classToCount = typeWildCard);
	TAbstractScriptableObject*			AccessByIndex(TTransaction* transaction, DescType desiredClass, long index);
	virtual TAbstractScriptableObject*	AccessByProperty(TTransaction* transaction, DescType propertyIdentifier);

	// These accessors are used less frequently:
	
	TAbstractScriptableObject*			AccessByName(TTransaction* transaction, DescType desiredClass, TDescriptor nameDesc);
	virtual TAbstractScriptableObject*	Access(TTransaction* transaction, DescType desiredClass, DescType keyForm, TDescriptor keyData);
	virtual TAbstractScriptableObject*	AccessByUniqueID(TTransaction* transaction, DescType desiredClass, TDescriptor uniqueID);
	virtual TAbstractScriptableObject*	AccessByOrdinal(TTransaction* transaction, DescType desiredClass, DescType ordinal);
	virtual void						AccessBySearchSpec(TTransaction* transaction, TAbstractCollector* collector, DescType desiredClass, TAbstractSearchSpec* searchSpec);
	
	// Methods relating to the data types accessible to this object
	
	virtual DescType					BestType(TTransaction* transaction, DescType propertyName);
	virtual DescType					DefaultType(TTransaction* transaction, DescType propertyName);
	virtual Boolean						CanReturnDataOfType(TTransaction* transaction, DescType propertyName, DescType desiredType);
	DescType							DetermineDataTypeToReturn(TTransaction* transaction, DescType propertyName, TDescriptor requestedTypes);

	// Methods to override to provide data access to this object
	
			TDescriptor					GetData(TTransaction* transaction, DescType requestedType);
			void						SetData(TTransaction* transaction, TDescriptor keyData);

			TDescriptor					GetProperty(TTransaction* transaction, DescType propertyName, DescType desiredType);
			void						SetProperty(TTransaction* transaction, DescType propertyName, TDescriptor& data);

	virtual TDescriptor					GetProperty(TTransaction* transaction, DescType propertyName, DescType desiredType, unsigned long additionalInfo);
	virtual void						SetProperty(TTransaction* transaction, DescType propertyName, TDescriptor& data, unsigned long additionalInfo);
	
	// Make an object specifier for this object
	
	TDescriptor							BuildObjectSpecifierForProperty(TTransaction* transaction, DescType propertyName);
	virtual TDescriptor					BuildObjectSpecifier(TTransaction* transaction);
	virtual void						MakeKeyDataForSelf(TTransaction* transaction, DescType& keyForm, TDescriptor& keyData);
	virtual TDescriptor					BuildSpecifierForParent(TTransaction* transaction);

	virtual TDescriptor					BuildSimpleSpecifier(TTransaction* transaction);

	// Event handlers
	
	virtual TDescriptor					AECommand(TTransaction* transaction, TAEvent ae, TAEvent reply, long aeCommandID, TAbstractScriptableObject* auxObjects = nil, long auxInfo = 0);
	virtual TAbstractScriptableObject*	CreateNewElement(TTransaction* transaction, TAEvent ae, TAEvent reply, DescType newObjectClass, TDescriptor initialData, TDescriptor initialProperties, Boolean& usedInitialData, Boolean& usedInitialProperties);
		
	// Get and set data and properties of this item
	
	TDescriptor							GetDataGivenListOfTypes(TTransaction* transaction, DescType propertyName, TDescriptor listOfTypes, unsigned long additionalInfo = kLookupAdditionalInfo);
	TDescriptor							GetDataOfType(TTransaction* transaction, DescType propertyName, DescType requestedType, unsigned long additionalInfo = kLookupAdditionalInfo);
	long								GetDataSizeGivenListOfTypes(TTransaction* transaction, DescType propertyName, TDescriptor listOfRequestedTypes);
	long								GetDataSize(TTransaction* transaction, DescType propertyName, DescType requestedType);

	void								SetProperties(TTransaction* transaction, TDescriptor& properties);
	TDescriptor							GetProperties(TTransaction* transaction, DescType desiredType);
	
	// Used by Set Data handler to do intense voodoo magic
	
	TDescriptor							ResolveSpecifierAndGetData(TTransaction* transaction, TDescriptor objectSpecifier);
	TDescriptor							ResolveKeyData(TTransaction* transaction, TDescriptor keyData);

	// Used by Get Data handler, more for convenience
	
	void								CoerceResultToRequestedType(TDescriptor& result, DescType requestedType);

	// Compare this object with some other object

	Boolean								CompareDescriptor(TTransaction* transaction, DescType comparisonOperator, TDescriptor& descOrObject);
	Boolean								Compare(TTransaction* transaction, DescType comparisonOperator, TAbstractScriptableObject* compareWith );
	Boolean								Compare(TTransaction* transaction, DescType comparisonOperator, TDescriptor& descOrObject );
	
	// Compare a property of this object with a descriptor
	
	virtual Boolean						CompareProperty(TTransaction* transaction, DescType propertyIdentifier, DescType comparisonOperator, TDescriptor compareWith);
	
	// Parse whose descriptors
	
	TComparisonOperand*					ParseComparisonOperand(TDescriptor& operand);
	TAbstractSearchSpec*				ParseComparisonOperator(DescType comparisonOperator, TDescriptor& object1, TDescriptor& object2);
	TAbstractSearchSpec*				ParseLogicalDescriptor(DescType logicalOperator, TDescriptor logicalTerms);
	TAbstractSearchSpec*				ParseWhoseTest(TDescriptor whoseDescriptor);
	TAbstractSearchSpec*				ParseWhoseDescriptor(TDescriptor whoseDescriptor, TDescriptor& scopeOfSearch);
		
	// Event dispatching
	
	virtual TAbstractScriptableObject*	GetAECommandDispatcher(TTransaction* t, long commandID, long auxInfo, long& oneDispatchSelector);
	virtual TDescriptor					AECommandDispatch(TTransaction* transaction, TAEvent ae, TAEvent reply, long aeCommandID, long auxInfo = 0);

	// Glue for abstract references (not every scriptable object will have a reference object)
	
	virtual TAbstractScriptableObject*	ReturnReferencedObject();
	virtual TAbstractReference*			CreateReference();
	
	virtual TAbstractScriptableObject*	RepresentativeScriptingObject(TTransaction* transaction);
	virtual Boolean						SendCommandToRepresentative(TTransaction* transaction, long aeCommandID);

	// Methods that determine if this object recognizes a certain property
	
	Boolean								PropertyExists(TTransaction* t, DescType propertyName);
	const TPropertyDescription*			DescriptionOfProperty(TTransaction* t, DescType propertyName);
	long								NumberOfProperties(TTransaction* t);
	const TPropertyDescription*			DescriptionOfNthProperty(TTransaction* t, long propertyNumber);
	
	// Virtual access methods for mark tokens/proxy tokens
	
	virtual TAbstractObjectIterator*	DirectObjectIterator(TTransaction*);
	TScriptableObjectList*				ConvertThisToList(TTransaction* transaction, DescType useRepresentative = 0);
	Boolean								ObjectRepresentedInToken(TTransaction* transaction, TAbstractScriptableObject* objectToTestForMembership);
	virtual Boolean						ObjectsAreTheSame(TTransaction* transaction, TAbstractScriptableObject* objectToTest);
	virtual TAbstractScriptableObject*	AdoptToken(TAbstractScriptableObject* token, TypeOfMarkToken markType);
	virtual void						AddThisToMarkToken(TAbstractScriptableObject*& markToken, TypeOfMarkToken markType);
	};


#endif
