/*
	File:		AbstractSearchSpec.h

	Contains:	xxx put contents here xxx

	Written by:	Andy Nicholas, Greg Anderson, Tom Conrad, Chris Bingham, Georgiann Puckett, John Thompson-Rohrlich

	Copyright:	 1994-1995 by Apple Computer, Inc., all rights reserved.

		<10>	 9/18/95	ga		

*/

#ifndef AbstractSearchSpec_h
#define AbstractSearchSpec_h

//
// MoreAEM.h is needed because methods defined in this
// file take TDescriptors as pass-by-value parameters,
// which requires the complete definition of the class
// TDescriptor.
// 
#include "MoreAEM.h"

//
// Object.h is needed because TObject is the base class of
// TAbstractSearchSpec, and for the DECLAREOBJECT macros.
//
#include "Object.h"

//
// For 'cObject'
//
#include <AERegistry.h>

//
// An object of class TAbstractScriptableObject is passed to TAbstractSearchSpec::Compare
// to determine if the object is within the search set.
//
class TAbstractScriptableObject;
class TScriptableObjectList;

//
// Each specific search specification must be able to create a
// search engine to search over the search space that the specification
// covers.
//
class TAbstractSearchEngine;
class TSearchEngineList;
class TTransaction;

class TAbstractCollector;
class TComparisonOperand;

//========================================================================================
//
//	CLASS TAbstractSearchSpec
//
// An object derived from TAbstractSearchSpec specifies some
// criterion for searching some specific namespace.
//
//========================================================================================

class TAbstractSearchSpec : public TObject
	{
	//
	// --- Methods -------------------------------------------------------------------------------------
	//
public:
	DECLAREOBJECT(TAbstractSearchSpec)

										TAbstractSearchSpec() {};
	
	// Methods relating to the use of the search specification
	
	virtual void						AccessBySearchSpec(TTransaction* t, TAbstractCollector* collector, DescType desiredClass, TAbstractScriptableObject* objectToSearch);
	virtual Boolean						Compare(TTransaction* t, TAbstractScriptableObject* itemToTest) = 0;
	
	// Methods relating to the contents of this search specification
	
	virtual DescType					SpecificationOperator();
	virtual long						NumberOfSubterms();
	virtual TAbstractSearchSpec*		SpecificationForTerm(long term);	
	virtual TDescriptor					BuildWhoseTest();		
	TDescriptor							BuildObjectSpecifier(DescType desiredClass, TDescriptor rootOfSearch);

	// Only used by comparison specifications
	
	virtual TComparisonOperand*			GetRightHandComparisonOperand();
	virtual TComparisonOperand*			GetLeftHandComparisonOperand();

	// Used to build up specifications from scratch
	
	virtual TAbstractSearchSpec*		AddTermToSpecification(TAbstractSearchSpec* newSpecification);
		
	// Methods relating to the actual search

	virtual TAbstractSearchEngine*		CreateSearchEngine();
	virtual void						CreateSearchEngines(TSearchEngineList* searchEngines, TAbstractScriptableObject* requestor);
	virtual Boolean						ShouldSearch (TAbstractScriptableObject* searchDisk);
	};



//
// List.h is needed to define TSearchSpecList, TSearchSpecLoop
// and FOREACHSEARCHSPEC.
//
#ifndef List_h
#include "List.h"
#endif

DECLAREOBJECTLIST(TSearchSpecList, TAbstractSearchSpec)
DECLARELOOP(TSearchSpecLoop, TSearchSpecList, TAbstractSearchSpec)

#define FOREACHSEARCHSPEC(listvar) FOREACH(TSearchSpecLoop, listvar, TAbstractSearchSpec, specification)



//========================================================================================
//
//	CLASS TLogicalSearchSpec
//
// A logical search specification contains a list of abstract search
// specifications, and either ANDs or ORs them together.
//
//========================================================================================

class TLogicalSearchSpec : public TAbstractSearchSpec
	{
protected:
	TSearchSpecList*					fListOfLogicalTerms;
	DescType							fLogicalOperator;
	
public:
	DECLAREOBJECT(TLogicalSearchSpec)

										TLogicalSearchSpec() : fListOfLogicalTerms(nil), fLogicalOperator(typeNull) {};
										TLogicalSearchSpec(DescType logicalOperator, TSearchSpecList* list) : fListOfLogicalTerms(list), fLogicalOperator(logicalOperator) {};
	virtual								~TLogicalSearchSpec();
		
	// Methods relating to the contents of this search specification
	
	virtual DescType					SpecificationOperator();
	virtual long						NumberOfSubterms();
	virtual TAbstractSearchSpec*		SpecificationForTerm(long term);
	
	// Methods relating to the use of the search specification
	
	virtual Boolean						Compare(TTransaction* t, TAbstractScriptableObject* itemToTest);

	// Methods relating to creating object specifiers for this search specification
	
	virtual TDescriptor					BuildWhoseTest();		
	virtual TAbstractSearchSpec*		AddTermToSpecification(TAbstractSearchSpec* newSpecification);
	
	};

#endif
