#include "DatabaseDocument.h"
#include "DBElement.h"
#include "DBProperty.h"
#include "Transaction.h"
#include "ReferenceTemplates.h"

#include "Application.h"
#include "ScriptableDBDocument.h"
#include "HFSBackingStore.h"
#include "ThreadContext.h"

//
// For MarkTokenMergeProc
//
#include "MarkToken.h"

#include "MoreAEM.h"
#include "Futures.h"
#include "ResolveObjectSpecifier.h"
#include "Accessor.h"
#include "CoreSuite.h"
#include "Exceptions.h"
#include "PseudoRandom.h"

#include <Events.h>
#include <Threads.h>

class TApplication;

void MainLoop();
void InitAll();

TTokenDescriptor MakeNullContainer();

//
// Sleep time set to 10 seconds, but might as well be sleep forever, since
// we don't have any idle tasks
//
#define kDBSleep 60 * 10
Boolean gApplicationRunning = false;
TApplication* gApplication = nil;

pascal OSErr QuitApplicationEvent(TAEvent& ae, TAEvent& reply, long refCon);

#if USESROUTINEDESCRIPTORS
	static RoutineDescriptor gQuitApplicationHandlerRD	= BUILD_ROUTINE_DESCRIPTOR(uppAEEventHandlerProcInfo, QuitApplicationEvent);
#endif

SysEnvRec		gThisMacintosh;
Boolean			gHasAppleEvents;


void InstallStringCompareBehavior();


//--------------------------------------------------------------------------------
// main
//--------------------------------------------------------------------------------
main()
{
	OSErr err = noErr;
	
	//
	// Make the universe come alive
	//
	InitAll();
	
	//
	// Use a different set of random numbers
	// every time
	//
	SetRandomSeed(TickCount());
	
	Try
	{
		gApplication = new TApplication;
		
		TDescriptor::SetCallbackFlags(kAEIDoWhose | kAEIDoMarking);
		InstallStringCompareBehavior();
		InstallNullContainerCreationProc(MakeNullContainer);
		InstallPreResolveProc(ResolveObjectSpecifier);
		InstallMergeTokensProc(MarkTokenMergeProc);
		
		TAccessor::InstallAEHandlers();
		TCoreSuite::InstallAEHandlers();
			
		//
		// Live here until a failure
		//
		gApplicationRunning = true;
		while(gApplicationRunning)
			MainLoop();

	}
	Catch(err)
	{
		if(err != userCanceledErr)
			DebugStr("\pCaught an error in testmain");
	}
	
	//
	// Before we bail out, we should save all of our documents
	//
	gApplication->TearDownApplication();
	
	return 0;
}

//----------------------------------------------------------------------------------------
// MainLoop: 
//----------------------------------------------------------------------------------------
void MainLoop()
{
	OSErr err = noErr;

	//
	// Before any threads are created, make sure that the main thread
	// gets swap in and swap out context switch messages
	//
	ThreadCreateNotifyHandler(kApplicationThreadID, 0);

	Try
	{
#if USESROUTINEDESCRIPTORS
		AEInstallEventHandler(kCoreEventClass, kAEQuitApplication, &gQuitApplicationHandlerRD, 0, false);
#else
		AEInstallEventHandler(kCoreEventClass, kAEQuitApplication, (AEEventHandlerProcPtr) &QuitApplicationEvent, 0, false);
#endif

		unsigned long lastWNECallTick = 0;
		while(gApplicationRunning == true)
		{
			//
			// Active threads includes the main thread.  If there
			// are no active threads, then call WNE all the time.
			// If there are active threads, let them run for
			// at least 10 ticks before calling WNE.
			//
			// If we had a real scheduler, then the scheduler should
			// ask the threads how busy they are; if there's a lot
			// of work to be done, then the throttle should be high
			// (10 or 20 ticks).  If there are threads running but they
			// aren't doing anything too important, then the throttle
			// could be much lower.  (The throttle could just be the
			// max of what any thread said it wanted.)
			//
			Boolean threadsActive = TThreadContext::ActiveThreads() > 1;
			unsigned long throttleWNETicks = threadsActive ? 10 : 0;
			
			if((TickCount() - lastWNECallTick) > throttleWNETicks)
			{
				//
				// Similarly, if there are threads active, we don't want
				// to sleep for any longer than necessary (don't use zero,
				// because that value is special); note that it's unlikely
				// that WNE will actually come back in one tick.
				//
				// If there are no active threads, then sleep for a "long"
				// time.
				//
				// Again, if we had a scheduler it should ask each thread
				// how long it thought the sleep time should be, and take
				// the minimum returned by any thread.  For the most part,
				// threads that want a high throttle also want a low sleep
				// time.  However, an idle (or mostly inactive) thread might
				// want a low throttle for the time being, but might also
				// desire a low sleep time because it needs to do periodic
				// tasks every N ticks, or it's about to wake up, or...
				//
				long sleepTimeForWNE = threadsActive ? 1 : kDBSleep;
				EventRecord event;
							
				//
				// We require System 7, so we'll always have WaitNextEvent.
				// Remember the tickcount that we called WaitNextEvent
				// at _after_ WNE returns (since we never know how long
				// our clients will hog the CPU for)
				//
				WaitNextEvent(everyEvent, &event, kDBSleep, nil);
				lastWNECallTick = TickCount();
				
				switch(event.what)
				{
					//
					// We live for high level events (and not much else)
					//
					case kHighLevelEvent:
					{
						//
						// We've installed the Futures package,
						// so any call to AEProcessAppleEvent
						// will spawn a thread to process the event
						//
						AEProcessAppleEvent(&event);
						break;
					}
				}
			}
			
			//
			// Idle our application and allow other threads to run.
			//
			gApplication->Idle();
			YieldToAnyThread();
		}
	}
	Catch(err)
	{
		//
		// For now, only pass along user canceled errors
		// (n.b. Errors are not thrown through AEProcessAppleEvent,
		// so it is unlikely that we will ever get here)
		//
		if(err == userCanceledErr)
			{
			gApplicationRunning = false;
			Throw(err);
			}
		
		DebugStr("\pOh no, a main-loop failure!");
	}
} // MainLoop

//----------------------------------------------------------------------------------------
// QuitApplicationEvent: 
//----------------------------------------------------------------------------------------
pascal OSErr QuitApplicationEvent(TAEvent& /*ae*/, TAEvent& /*reply*/, long /*refCon*/)
{
	//
	// You should never call ExitToShell from an AppleEvent handler
	//
	gApplicationRunning = false;
	
	return noErr;
} // QuitApplicationEvent

extern pascal OSErr ThreadCreateHandler(ThreadEntryProcPtr threadEntry, void *threadParam, long handlerRefCon, ThreadID *threadMade);

//----------------------------------------------------------------------------------------
// InitAll
// 
// Initialize all of the relivant Macintosh managers.  We probably
// don't need to do _all_ of this for db+.
//----------------------------------------------------------------------------------------
void InitAll()
{
	OSErr	theErr;
	long	heapSpace;
	Ptr		appLimit;
	THz		appBase;
	long	gestaltResult;
	short	callsToMoreMasters = 10;
	
	appBase = ApplicZone();
	appLimit = GetApplLimit();
	heapSpace = FreeMem();
	MaxApplZone();
	appLimit = GetApplLimit();
	heapSpace = FreeMem();
	while( callsToMoreMasters-- )
		MoreMasters();
	
	InitGraf(&qd.thePort);
	InitFonts();
	FlushEvents(everyEvent, 0);
	InitWindows();
	InitMenus();
	TEInit();
	InitDialogs(0);
	
	//
	// Get the SysEnvirons record
	//
	SysEnvirons( 1, &gThisMacintosh );

	//
	// Check to see if AppleEvents are available (we are going to REQUIRE them)
	//
	theErr = Gestalt( gestaltAppleEventsAttr, &gestaltResult );
	gHasAppleEvents = ( (theErr == noErr) && ((gestaltResult & (1L << gestaltAppleEventsPresent)) != 0) );

	//
	// Initialize the Object Support Library
	// and install the futures package.
	//
	AEObjectInit();
	InitFutures(ThreadCreateHandler);
} // InitAll


//----------------------------------------------------------------------------------------
// MakeNullContainer: 
//----------------------------------------------------------------------------------------
TTokenDescriptor MakeNullContainer()
{
	TTokenDescriptor nullContainerDesc;
	
	nullContainerDesc.AdoptToken(gApplication);
	
	return nullContainerDesc;
} // MakeNullContainer 

#if 0

//----------------------------------------------------------------------------------------
// Global operator new
//----------------------------------------------------------------------------------------
void* operator new(size_t size)
{
}

//----------------------------------------------------------------------------------------
// Global operator delete
//----------------------------------------------------------------------------------------
void operator delete(void* object, size_t size)
{

}

#endif
